/*
 * this code is auto generated. do not edit. change the sources in tools/introspection/ instead.
 *
 * the parse tree:
 *
 * typedef
 *   struct 
 *     struct 
 *       struct 
 *         enum 
 *           DT_LIQUIFY_PATH_INVALIDATED : invalidated
 *           DT_LIQUIFY_PATH_MOVE_TO_V1 : move
 *           DT_LIQUIFY_PATH_LINE_TO_V1 : line
 *           DT_LIQUIFY_PATH_CURVE_TO_V1 : curve
 *         type 
 *           default: 0
 * 
 *         enum 
 *           DT_LIQUIFY_NODE_TYPE_CUSP : cusp
 *           DT_LIQUIFY_NODE_TYPE_SMOOTH : smooth
 *           DT_LIQUIFY_NODE_TYPE_SYMMETRICAL : symmetrical
 *           DT_LIQUIFY_NODE_TYPE_AUTOSMOOTH : autosmooth
 *           DT_LIQUIFY_NODE_TYPE_LAST
 *         node_type 
 *           default: 0
 * 
 *         enum 
 *           DT_LIQUIFY_LAYER_BACKGROUND
 *           DT_LIQUIFY_LAYER_RADIUS
 *           DT_LIQUIFY_LAYER_HARDNESS1
 *           DT_LIQUIFY_LAYER_HARDNESS2
 *           DT_LIQUIFY_LAYER_WARPS
 *           DT_LIQUIFY_LAYER_PATH
 *           DT_LIQUIFY_LAYER_CTRLPOINT1_HANDLE
 *           DT_LIQUIFY_LAYER_CTRLPOINT2_HANDLE
 *           DT_LIQUIFY_LAYER_RADIUSPOINT_HANDLE
 *           DT_LIQUIFY_LAYER_HARDNESSPOINT1_HANDLE
 *           DT_LIQUIFY_LAYER_HARDNESSPOINT2_HANDLE
 *           DT_LIQUIFY_LAYER_STRENGTHPOINT_HANDLE
 *           DT_LIQUIFY_LAYER_CENTERPOINT
 *           DT_LIQUIFY_LAYER_CTRLPOINT1
 *           DT_LIQUIFY_LAYER_CTRLPOINT2
 *           DT_LIQUIFY_LAYER_RADIUSPOINT
 *           DT_LIQUIFY_LAYER_HARDNESSPOINT1
 *           DT_LIQUIFY_LAYER_HARDNESSPOINT2
 *           DT_LIQUIFY_LAYER_STRENGTHPOINT
 *           DT_LIQUIFY_LAYER_LAST
 *         selected 
 *           default: 0
 * 
 *         enum 
 *           DT_LIQUIFY_LAYER_BACKGROUND
 *           DT_LIQUIFY_LAYER_RADIUS
 *           DT_LIQUIFY_LAYER_HARDNESS1
 *           DT_LIQUIFY_LAYER_HARDNESS2
 *           DT_LIQUIFY_LAYER_WARPS
 *           DT_LIQUIFY_LAYER_PATH
 *           DT_LIQUIFY_LAYER_CTRLPOINT1_HANDLE
 *           DT_LIQUIFY_LAYER_CTRLPOINT2_HANDLE
 *           DT_LIQUIFY_LAYER_RADIUSPOINT_HANDLE
 *           DT_LIQUIFY_LAYER_HARDNESSPOINT1_HANDLE
 *           DT_LIQUIFY_LAYER_HARDNESSPOINT2_HANDLE
 *           DT_LIQUIFY_LAYER_STRENGTHPOINT_HANDLE
 *           DT_LIQUIFY_LAYER_CENTERPOINT
 *           DT_LIQUIFY_LAYER_CTRLPOINT1
 *           DT_LIQUIFY_LAYER_CTRLPOINT2
 *           DT_LIQUIFY_LAYER_RADIUSPOINT
 *           DT_LIQUIFY_LAYER_HARDNESSPOINT1
 *           DT_LIQUIFY_LAYER_HARDNESSPOINT2
 *           DT_LIQUIFY_LAYER_STRENGTHPOINT
 *           DT_LIQUIFY_LAYER_LAST
 *         hovered 
 *           default: 0
 * 
 *         int8_t
 *         prev 
 *           default: 0
 *           range: G_MININT8 .. G_MAXINT8
 * 
 *         int8_t
 *         idx 
 *           default: 0
 *           range: G_MININT8 .. G_MAXINT8
 * 
 *         int8_t
 *         next 
 *           default: 0
 *           range: G_MININT8 .. G_MAXINT8
 * 
 *       header 
 * 
 *       struct 
 *         float complex
 *         point 
 *           default: 0.0 + 0.0 * _Complex_I
 *           range: -G_MAXFLOAT + -G_MAXFLOAT * _Complex_I .. G_MAXFLOAT + G_MAXFLOAT * _Complex_I
 * 
 *         float complex
 *         strength 
 *           default: 0.0 + 0.0 * _Complex_I
 *           range: -G_MAXFLOAT + -G_MAXFLOAT * _Complex_I .. G_MAXFLOAT + G_MAXFLOAT * _Complex_I
 * 
 *         float complex
 *         radius 
 *           default: 0.0 + 0.0 * _Complex_I
 *           range: -G_MAXFLOAT + -G_MAXFLOAT * _Complex_I .. G_MAXFLOAT + G_MAXFLOAT * _Complex_I
 * 
 *         float
 *         control1 
 *           default: 0.0
 *           range: -G_MAXFLOAT .. G_MAXFLOAT
 * 
 *         float
 *         control2 
 *           default: 0.0
 *           range: -G_MAXFLOAT .. G_MAXFLOAT
 * 
 *         enum 
 *           DT_LIQUIFY_WARP_TYPE_LINEAR : linear
 *           DT_LIQUIFY_WARP_TYPE_RADIAL_GROW : radial grow
 *           DT_LIQUIFY_WARP_TYPE_RADIAL_SHRINK : radial shrink
 *           DT_LIQUIFY_WARP_TYPE_LAST
 *         type 
 *           default: 0
 * 
 *         enum 
 *           DT_LIQUIFY_STATUS_NONE : none
 *           DT_LIQUIFY_STATUS_NEW : new
 *           DT_LIQUIFY_STATUS_INTERPOLATED : interpolated
 *           DT_LIQUIFY_STATUS_PREVIEW : preview
 *           DT_LIQUIFY_STATUS_LAST
 *         status 
 *           default: 0
 * 
 *       warp 
 * 
 *       struct 
 *         float complex
 *         ctrl1 
 *           default: 0.0 + 0.0 * _Complex_I
 *           range: -G_MAXFLOAT + -G_MAXFLOAT * _Complex_I .. G_MAXFLOAT + G_MAXFLOAT * _Complex_I
 * 
 *         float complex
 *         ctrl2 
 *           default: 0.0 + 0.0 * _Complex_I
 *           range: -G_MAXFLOAT + -G_MAXFLOAT * _Complex_I .. G_MAXFLOAT + G_MAXFLOAT * _Complex_I
 * 
 *       node 
 * 
 *     nodes [MAX_NODES]
 * 
 * dt_iop_liquify_params_t
 *
 */

#include "/build/darktable/src/darktable-5.6.0/src/iop/liquify.c"

#ifdef __cplusplus
extern "C"
{
#endif

#include <string.h>
#include "common/introspection.h"

static dt_introspection_field_t introspection_linear[22+1]
#if !( defined(__cplusplus) && !defined(__clang__) && defined(__GNUC__) && ((__GNUC__*100 + __GNUC_MINOR__) <= 406) )
= {
{
  .Enum = {
    { DT_INTROSPECTION_TYPE_ENUM, (char*)"", (char*)"nodes[0].header.type", (char*)"type", (char*)"", sizeof(((dt_iop_liquify_params_t*)NULL)->nodes[0].header.type), G_STRUCT_OFFSET(dt_iop_liquify_params_t, nodes[0].header.type), NULL },
    /*entries*/ 4, /*values*/ NULL, /*Default*/ 0
  }
},
{
  .Enum = {
    { DT_INTROSPECTION_TYPE_ENUM, (char*)"", (char*)"nodes[0].header.node_type", (char*)"node_type", (char*)"", sizeof(((dt_iop_liquify_params_t*)NULL)->nodes[0].header.node_type), G_STRUCT_OFFSET(dt_iop_liquify_params_t, nodes[0].header.node_type), NULL },
    /*entries*/ 5, /*values*/ NULL, /*Default*/ 0
  }
},
{
  .Enum = {
    { DT_INTROSPECTION_TYPE_ENUM, (char*)"", (char*)"nodes[0].header.selected", (char*)"selected", (char*)"", sizeof(((dt_iop_liquify_params_t*)NULL)->nodes[0].header.selected), G_STRUCT_OFFSET(dt_iop_liquify_params_t, nodes[0].header.selected), NULL },
    /*entries*/ 20, /*values*/ NULL, /*Default*/ 0
  }
},
{
  .Enum = {
    { DT_INTROSPECTION_TYPE_ENUM, (char*)"", (char*)"nodes[0].header.hovered", (char*)"hovered", (char*)"", sizeof(((dt_iop_liquify_params_t*)NULL)->nodes[0].header.hovered), G_STRUCT_OFFSET(dt_iop_liquify_params_t, nodes[0].header.hovered), NULL },
    /*entries*/ 20, /*values*/ NULL, /*Default*/ 0
  }
},
{
  .Int8 = {
    { DT_INTROSPECTION_TYPE_INT8, (char*)"int8", (char*)"nodes[0].header.prev", (char*)"prev", (char*)"", sizeof(((dt_iop_liquify_params_t*)NULL)->nodes[0].header.prev), G_STRUCT_OFFSET(dt_iop_liquify_params_t, nodes[0].header.prev), NULL },
    /*Min*/ G_MININT8, /*Max*/ G_MAXINT8, /*Default*/ 0
  }
},
{
  .Int8 = {
    { DT_INTROSPECTION_TYPE_INT8, (char*)"int8", (char*)"nodes[0].header.idx", (char*)"idx", (char*)"", sizeof(((dt_iop_liquify_params_t*)NULL)->nodes[0].header.idx), G_STRUCT_OFFSET(dt_iop_liquify_params_t, nodes[0].header.idx), NULL },
    /*Min*/ G_MININT8, /*Max*/ G_MAXINT8, /*Default*/ 0
  }
},
{
  .Int8 = {
    { DT_INTROSPECTION_TYPE_INT8, (char*)"int8", (char*)"nodes[0].header.next", (char*)"next", (char*)"", sizeof(((dt_iop_liquify_params_t*)NULL)->nodes[0].header.next), G_STRUCT_OFFSET(dt_iop_liquify_params_t, nodes[0].header.next), NULL },
    /*Min*/ G_MININT8, /*Max*/ G_MAXINT8, /*Default*/ 0
  }
},
{
  .Struct = {
    { DT_INTROSPECTION_TYPE_STRUCT, (char*)"", (char*)"nodes[0].header", (char*)"header", (char*)"", sizeof(((dt_iop_liquify_params_t*)NULL)->nodes[0].header), G_STRUCT_OFFSET(dt_iop_liquify_params_t, nodes[0].header), NULL },
    /*entries*/ 7, /*fields*/ NULL
  }
},
{
  .FloatComplex = {
    { DT_INTROSPECTION_TYPE_FLOATCOMPLEX, (char*)"floatcomplex", (char*)"nodes[0].warp.point", (char*)"point", (char*)"", sizeof(((dt_iop_liquify_params_t*)NULL)->nodes[0].warp.point), G_STRUCT_OFFSET(dt_iop_liquify_params_t, nodes[0].warp.point), NULL },
    /*Min*/ -G_MAXFLOAT + -G_MAXFLOAT * _Complex_I, /*Max*/ G_MAXFLOAT + G_MAXFLOAT * _Complex_I, /*Default*/ 0.0 + 0.0 * _Complex_I
  }
},
{
  .FloatComplex = {
    { DT_INTROSPECTION_TYPE_FLOATCOMPLEX, (char*)"floatcomplex", (char*)"nodes[0].warp.strength", (char*)"strength", (char*)"", sizeof(((dt_iop_liquify_params_t*)NULL)->nodes[0].warp.strength), G_STRUCT_OFFSET(dt_iop_liquify_params_t, nodes[0].warp.strength), NULL },
    /*Min*/ -G_MAXFLOAT + -G_MAXFLOAT * _Complex_I, /*Max*/ G_MAXFLOAT + G_MAXFLOAT * _Complex_I, /*Default*/ 0.0 + 0.0 * _Complex_I
  }
},
{
  .FloatComplex = {
    { DT_INTROSPECTION_TYPE_FLOATCOMPLEX, (char*)"floatcomplex", (char*)"nodes[0].warp.radius", (char*)"radius", (char*)"", sizeof(((dt_iop_liquify_params_t*)NULL)->nodes[0].warp.radius), G_STRUCT_OFFSET(dt_iop_liquify_params_t, nodes[0].warp.radius), NULL },
    /*Min*/ -G_MAXFLOAT + -G_MAXFLOAT * _Complex_I, /*Max*/ G_MAXFLOAT + G_MAXFLOAT * _Complex_I, /*Default*/ 0.0 + 0.0 * _Complex_I
  }
},
{
  .Float = {
    { DT_INTROSPECTION_TYPE_FLOAT, (char*)"float", (char*)"nodes[0].warp.control1", (char*)"control1", (char*)"", sizeof(((dt_iop_liquify_params_t*)NULL)->nodes[0].warp.control1), G_STRUCT_OFFSET(dt_iop_liquify_params_t, nodes[0].warp.control1), NULL },
    /*Min*/ -G_MAXFLOAT, /*Max*/ G_MAXFLOAT, /*Default*/ 0.0
  }
},
{
  .Float = {
    { DT_INTROSPECTION_TYPE_FLOAT, (char*)"float", (char*)"nodes[0].warp.control2", (char*)"control2", (char*)"", sizeof(((dt_iop_liquify_params_t*)NULL)->nodes[0].warp.control2), G_STRUCT_OFFSET(dt_iop_liquify_params_t, nodes[0].warp.control2), NULL },
    /*Min*/ -G_MAXFLOAT, /*Max*/ G_MAXFLOAT, /*Default*/ 0.0
  }
},
{
  .Enum = {
    { DT_INTROSPECTION_TYPE_ENUM, (char*)"", (char*)"nodes[0].warp.type", (char*)"type", (char*)"", sizeof(((dt_iop_liquify_params_t*)NULL)->nodes[0].warp.type), G_STRUCT_OFFSET(dt_iop_liquify_params_t, nodes[0].warp.type), NULL },
    /*entries*/ 4, /*values*/ NULL, /*Default*/ 0
  }
},
{
  .Enum = {
    { DT_INTROSPECTION_TYPE_ENUM, (char*)"", (char*)"nodes[0].warp.status", (char*)"status", (char*)"", sizeof(((dt_iop_liquify_params_t*)NULL)->nodes[0].warp.status), G_STRUCT_OFFSET(dt_iop_liquify_params_t, nodes[0].warp.status), NULL },
    /*entries*/ 5, /*values*/ NULL, /*Default*/ 0
  }
},
{
  .Struct = {
    { DT_INTROSPECTION_TYPE_STRUCT, (char*)"", (char*)"nodes[0].warp", (char*)"warp", (char*)"", sizeof(((dt_iop_liquify_params_t*)NULL)->nodes[0].warp), G_STRUCT_OFFSET(dt_iop_liquify_params_t, nodes[0].warp), NULL },
    /*entries*/ 7, /*fields*/ NULL
  }
},
{
  .FloatComplex = {
    { DT_INTROSPECTION_TYPE_FLOATCOMPLEX, (char*)"floatcomplex", (char*)"nodes[0].node.ctrl1", (char*)"ctrl1", (char*)"", sizeof(((dt_iop_liquify_params_t*)NULL)->nodes[0].node.ctrl1), G_STRUCT_OFFSET(dt_iop_liquify_params_t, nodes[0].node.ctrl1), NULL },
    /*Min*/ -G_MAXFLOAT + -G_MAXFLOAT * _Complex_I, /*Max*/ G_MAXFLOAT + G_MAXFLOAT * _Complex_I, /*Default*/ 0.0 + 0.0 * _Complex_I
  }
},
{
  .FloatComplex = {
    { DT_INTROSPECTION_TYPE_FLOATCOMPLEX, (char*)"floatcomplex", (char*)"nodes[0].node.ctrl2", (char*)"ctrl2", (char*)"", sizeof(((dt_iop_liquify_params_t*)NULL)->nodes[0].node.ctrl2), G_STRUCT_OFFSET(dt_iop_liquify_params_t, nodes[0].node.ctrl2), NULL },
    /*Min*/ -G_MAXFLOAT + -G_MAXFLOAT * _Complex_I, /*Max*/ G_MAXFLOAT + G_MAXFLOAT * _Complex_I, /*Default*/ 0.0 + 0.0 * _Complex_I
  }
},
{
  .Struct = {
    { DT_INTROSPECTION_TYPE_STRUCT, (char*)"", (char*)"nodes[0].node", (char*)"node", (char*)"", sizeof(((dt_iop_liquify_params_t*)NULL)->nodes[0].node), G_STRUCT_OFFSET(dt_iop_liquify_params_t, nodes[0].node), NULL },
    /*entries*/ 2, /*fields*/ NULL
  }
},
{
  .Struct = {
    { DT_INTROSPECTION_TYPE_STRUCT, (char*)"", (char*)"nodes[0]", (char*)"nodes[0]", (char*)"", sizeof(((dt_iop_liquify_params_t*)NULL)->nodes[0]), G_STRUCT_OFFSET(dt_iop_liquify_params_t, nodes), NULL },
    /*entries*/ 3, /*fields*/ NULL
  }
},
{
  .Array = {
    { DT_INTROSPECTION_TYPE_ARRAY, (char*)"[]", (char*)"nodes", (char*)"nodes", (char*)"", sizeof(((dt_iop_liquify_params_t*)NULL)->nodes), G_STRUCT_OFFSET(dt_iop_liquify_params_t, nodes), NULL },
    /*count*/ G_N_ELEMENTS(((dt_iop_liquify_params_t*)NULL)->nodes), /*type*/ DT_INTROSPECTION_TYPE_STRUCT, /*field*/ &introspection_linear[19]
  }
},
{
  .Struct = {
    { DT_INTROSPECTION_TYPE_STRUCT, (char*)"dt_iop_liquify_params_t", (char*)"", (char*)"", (char*)"", sizeof(dt_iop_liquify_params_t), 0, NULL },
    /*entries*/ 1, /*fields*/ NULL
  }
},
  { .header = {DT_INTROSPECTION_TYPE_NONE, NULL, NULL, NULL, NULL, 0, 0, NULL} }
}
#endif
;

static dt_introspection_t introspection = {
  8,
  1,
  "dt_iop_liquify_params_t",
  sizeof(dt_iop_liquify_params_t),
  &introspection_linear[22-1],
  sizeof(dt_iop_module_t),
  G_STRUCT_OFFSET(dt_iop_module_t, default_params)
};

dt_introspection_field_t* get_introspection_linear()
{
  return introspection_linear;
}

dt_introspection_t* get_introspection()
{
  return &introspection;
}

int introspection_init(struct dt_iop_module_so_t *self, int api_version)
{
  // here we check that the generated code matches the api at compile time and also at runtime
  if(introspection.api_version != DT_INTROSPECTION_VERSION || api_version != DT_INTROSPECTION_VERSION)
    return 1;

#if defined(__cplusplus) && !defined(__clang__) && defined(__GNUC__) && ((__GNUC__*100 + __GNUC_MINOR__) <= 406)
  introspection_linear[0].Enum = {
    { DT_INTROSPECTION_TYPE_ENUM, (char*)"", (char*)"nodes[0].header.type", (char*)"type", (char*)"", sizeof(((dt_iop_liquify_params_t*)NULL)->nodes[0].header.type), G_STRUCT_OFFSET(dt_iop_liquify_params_t, nodes[0].header.type), NULL },
    /*entries*/ 4, /*values*/ NULL, /*Default*/ 0
  };
  introspection_linear[1].Enum = {
    { DT_INTROSPECTION_TYPE_ENUM, (char*)"", (char*)"nodes[0].header.node_type", (char*)"node_type", (char*)"", sizeof(((dt_iop_liquify_params_t*)NULL)->nodes[0].header.node_type), G_STRUCT_OFFSET(dt_iop_liquify_params_t, nodes[0].header.node_type), NULL },
    /*entries*/ 5, /*values*/ NULL, /*Default*/ 0
  };
  introspection_linear[2].Enum = {
    { DT_INTROSPECTION_TYPE_ENUM, (char*)"", (char*)"nodes[0].header.selected", (char*)"selected", (char*)"", sizeof(((dt_iop_liquify_params_t*)NULL)->nodes[0].header.selected), G_STRUCT_OFFSET(dt_iop_liquify_params_t, nodes[0].header.selected), NULL },
    /*entries*/ 20, /*values*/ NULL, /*Default*/ 0
  };
  introspection_linear[3].Enum = {
    { DT_INTROSPECTION_TYPE_ENUM, (char*)"", (char*)"nodes[0].header.hovered", (char*)"hovered", (char*)"", sizeof(((dt_iop_liquify_params_t*)NULL)->nodes[0].header.hovered), G_STRUCT_OFFSET(dt_iop_liquify_params_t, nodes[0].header.hovered), NULL },
    /*entries*/ 20, /*values*/ NULL, /*Default*/ 0
  };
  introspection_linear[4].Int8 = {
    { DT_INTROSPECTION_TYPE_INT8, (char*)"int8", (char*)"nodes[0].header.prev", (char*)"prev", (char*)"", sizeof(((dt_iop_liquify_params_t*)NULL)->nodes[0].header.prev), G_STRUCT_OFFSET(dt_iop_liquify_params_t, nodes[0].header.prev), NULL },
    /*Min*/ G_MININT8, /*Max*/ G_MAXINT8, /*Default*/ 0
  };
  introspection_linear[5].Int8 = {
    { DT_INTROSPECTION_TYPE_INT8, (char*)"int8", (char*)"nodes[0].header.idx", (char*)"idx", (char*)"", sizeof(((dt_iop_liquify_params_t*)NULL)->nodes[0].header.idx), G_STRUCT_OFFSET(dt_iop_liquify_params_t, nodes[0].header.idx), NULL },
    /*Min*/ G_MININT8, /*Max*/ G_MAXINT8, /*Default*/ 0
  };
  introspection_linear[6].Int8 = {
    { DT_INTROSPECTION_TYPE_INT8, (char*)"int8", (char*)"nodes[0].header.next", (char*)"next", (char*)"", sizeof(((dt_iop_liquify_params_t*)NULL)->nodes[0].header.next), G_STRUCT_OFFSET(dt_iop_liquify_params_t, nodes[0].header.next), NULL },
    /*Min*/ G_MININT8, /*Max*/ G_MAXINT8, /*Default*/ 0
  };
  introspection_linear[7].Struct = {
    { DT_INTROSPECTION_TYPE_STRUCT, (char*)"", (char*)"nodes[0].header", (char*)"header", (char*)"", sizeof(((dt_iop_liquify_params_t*)NULL)->nodes[0].header), G_STRUCT_OFFSET(dt_iop_liquify_params_t, nodes[0].header), NULL },
    /*entries*/ 7, /*fields*/ NULL
  };
  introspection_linear[8].FloatComplex = {
    { DT_INTROSPECTION_TYPE_FLOATCOMPLEX, (char*)"floatcomplex", (char*)"nodes[0].warp.point", (char*)"point", (char*)"", sizeof(((dt_iop_liquify_params_t*)NULL)->nodes[0].warp.point), G_STRUCT_OFFSET(dt_iop_liquify_params_t, nodes[0].warp.point), NULL },
    /*Min*/ -G_MAXFLOAT + -G_MAXFLOAT * _Complex_I, /*Max*/ G_MAXFLOAT + G_MAXFLOAT * _Complex_I, /*Default*/ 0.0 + 0.0 * _Complex_I
  };
  introspection_linear[9].FloatComplex = {
    { DT_INTROSPECTION_TYPE_FLOATCOMPLEX, (char*)"floatcomplex", (char*)"nodes[0].warp.strength", (char*)"strength", (char*)"", sizeof(((dt_iop_liquify_params_t*)NULL)->nodes[0].warp.strength), G_STRUCT_OFFSET(dt_iop_liquify_params_t, nodes[0].warp.strength), NULL },
    /*Min*/ -G_MAXFLOAT + -G_MAXFLOAT * _Complex_I, /*Max*/ G_MAXFLOAT + G_MAXFLOAT * _Complex_I, /*Default*/ 0.0 + 0.0 * _Complex_I
  };
  introspection_linear[10].FloatComplex = {
    { DT_INTROSPECTION_TYPE_FLOATCOMPLEX, (char*)"floatcomplex", (char*)"nodes[0].warp.radius", (char*)"radius", (char*)"", sizeof(((dt_iop_liquify_params_t*)NULL)->nodes[0].warp.radius), G_STRUCT_OFFSET(dt_iop_liquify_params_t, nodes[0].warp.radius), NULL },
    /*Min*/ -G_MAXFLOAT + -G_MAXFLOAT * _Complex_I, /*Max*/ G_MAXFLOAT + G_MAXFLOAT * _Complex_I, /*Default*/ 0.0 + 0.0 * _Complex_I
  };
  introspection_linear[11].Float = {
    { DT_INTROSPECTION_TYPE_FLOAT, (char*)"float", (char*)"nodes[0].warp.control1", (char*)"control1", (char*)"", sizeof(((dt_iop_liquify_params_t*)NULL)->nodes[0].warp.control1), G_STRUCT_OFFSET(dt_iop_liquify_params_t, nodes[0].warp.control1), NULL },
    /*Min*/ -G_MAXFLOAT, /*Max*/ G_MAXFLOAT, /*Default*/ 0.0
  };
  introspection_linear[12].Float = {
    { DT_INTROSPECTION_TYPE_FLOAT, (char*)"float", (char*)"nodes[0].warp.control2", (char*)"control2", (char*)"", sizeof(((dt_iop_liquify_params_t*)NULL)->nodes[0].warp.control2), G_STRUCT_OFFSET(dt_iop_liquify_params_t, nodes[0].warp.control2), NULL },
    /*Min*/ -G_MAXFLOAT, /*Max*/ G_MAXFLOAT, /*Default*/ 0.0
  };
  introspection_linear[13].Enum = {
    { DT_INTROSPECTION_TYPE_ENUM, (char*)"", (char*)"nodes[0].warp.type", (char*)"type", (char*)"", sizeof(((dt_iop_liquify_params_t*)NULL)->nodes[0].warp.type), G_STRUCT_OFFSET(dt_iop_liquify_params_t, nodes[0].warp.type), NULL },
    /*entries*/ 4, /*values*/ NULL, /*Default*/ 0
  };
  introspection_linear[14].Enum = {
    { DT_INTROSPECTION_TYPE_ENUM, (char*)"", (char*)"nodes[0].warp.status", (char*)"status", (char*)"", sizeof(((dt_iop_liquify_params_t*)NULL)->nodes[0].warp.status), G_STRUCT_OFFSET(dt_iop_liquify_params_t, nodes[0].warp.status), NULL },
    /*entries*/ 5, /*values*/ NULL, /*Default*/ 0
  };
  introspection_linear[15].Struct = {
    { DT_INTROSPECTION_TYPE_STRUCT, (char*)"", (char*)"nodes[0].warp", (char*)"warp", (char*)"", sizeof(((dt_iop_liquify_params_t*)NULL)->nodes[0].warp), G_STRUCT_OFFSET(dt_iop_liquify_params_t, nodes[0].warp), NULL },
    /*entries*/ 7, /*fields*/ NULL
  };
  introspection_linear[16].FloatComplex = {
    { DT_INTROSPECTION_TYPE_FLOATCOMPLEX, (char*)"floatcomplex", (char*)"nodes[0].node.ctrl1", (char*)"ctrl1", (char*)"", sizeof(((dt_iop_liquify_params_t*)NULL)->nodes[0].node.ctrl1), G_STRUCT_OFFSET(dt_iop_liquify_params_t, nodes[0].node.ctrl1), NULL },
    /*Min*/ -G_MAXFLOAT + -G_MAXFLOAT * _Complex_I, /*Max*/ G_MAXFLOAT + G_MAXFLOAT * _Complex_I, /*Default*/ 0.0 + 0.0 * _Complex_I
  };
  introspection_linear[17].FloatComplex = {
    { DT_INTROSPECTION_TYPE_FLOATCOMPLEX, (char*)"floatcomplex", (char*)"nodes[0].node.ctrl2", (char*)"ctrl2", (char*)"", sizeof(((dt_iop_liquify_params_t*)NULL)->nodes[0].node.ctrl2), G_STRUCT_OFFSET(dt_iop_liquify_params_t, nodes[0].node.ctrl2), NULL },
    /*Min*/ -G_MAXFLOAT + -G_MAXFLOAT * _Complex_I, /*Max*/ G_MAXFLOAT + G_MAXFLOAT * _Complex_I, /*Default*/ 0.0 + 0.0 * _Complex_I
  };
  introspection_linear[18].Struct = {
    { DT_INTROSPECTION_TYPE_STRUCT, (char*)"", (char*)"nodes[0].node", (char*)"node", (char*)"", sizeof(((dt_iop_liquify_params_t*)NULL)->nodes[0].node), G_STRUCT_OFFSET(dt_iop_liquify_params_t, nodes[0].node), NULL },
    /*entries*/ 2, /*fields*/ NULL
  };
  introspection_linear[19].Struct = {
    { DT_INTROSPECTION_TYPE_STRUCT, (char*)"", (char*)"nodes[0]", (char*)"nodes[0]", (char*)"", sizeof(((dt_iop_liquify_params_t*)NULL)->nodes[0]), G_STRUCT_OFFSET(dt_iop_liquify_params_t, nodes), NULL },
    /*entries*/ 3, /*fields*/ NULL
  };
  introspection_linear[20].Array = {
    { DT_INTROSPECTION_TYPE_ARRAY, (char*)"[]", (char*)"nodes", (char*)"nodes", (char*)"", sizeof(((dt_iop_liquify_params_t*)NULL)->nodes), G_STRUCT_OFFSET(dt_iop_liquify_params_t, nodes), NULL },
    /*count*/ G_N_ELEMENTS(((dt_iop_liquify_params_t*)NULL)->nodes), /*type*/ DT_INTROSPECTION_TYPE_STRUCT, /*field*/ &introspection_linear[19]
  };
  introspection_linear[21].Struct = {
    { DT_INTROSPECTION_TYPE_STRUCT, (char*)"dt_iop_liquify_params_t", (char*)"", (char*)"", (char*)"", sizeof(dt_iop_liquify_params_t), 0, NULL },
    /*entries*/ 1, /*fields*/ NULL
  };
  introspection_linear[22].header = {DT_INTROSPECTION_TYPE_NONE, NULL, NULL, NULL, NULL, 0, 0, NULL};
#endif

  for(int i = 0; i <= 22; i++)
    introspection_linear[i].header.so = self;

  static dt_introspection_type_enum_tuple_t f0[] = { 
    { "DT_LIQUIFY_PATH_INVALIDATED", DT_LIQUIFY_PATH_INVALIDATED, N_((char*)"invalidated") },
    { "DT_LIQUIFY_PATH_MOVE_TO_V1", DT_LIQUIFY_PATH_MOVE_TO_V1, N_((char*)"move") },
    { "DT_LIQUIFY_PATH_LINE_TO_V1", DT_LIQUIFY_PATH_LINE_TO_V1, N_((char*)"line") },
    { "DT_LIQUIFY_PATH_CURVE_TO_V1", DT_LIQUIFY_PATH_CURVE_TO_V1, N_((char*)"curve") },
    { NULL, 0 },
  };
  static dt_introspection_field_t *f7[] = {
    &introspection_linear[0],
    &introspection_linear[1],
    &introspection_linear[2],
    &introspection_linear[3],
    &introspection_linear[4],
    &introspection_linear[5],
    &introspection_linear[6],
    NULL
  };
  static dt_introspection_field_t *f15[] = {
    &introspection_linear[8],
    &introspection_linear[9],
    &introspection_linear[10],
    &introspection_linear[11],
    &introspection_linear[12],
    &introspection_linear[13],
    &introspection_linear[14],
    NULL
  };
  static dt_introspection_field_t *f18[] = {
    &introspection_linear[16],
    &introspection_linear[17],
    NULL
  };
  static dt_introspection_field_t *f19[] = {
    &introspection_linear[7],
    &introspection_linear[15],
    &introspection_linear[18],
    NULL
  };
  static dt_introspection_field_t *f21[] = {
    &introspection_linear[20],
    NULL
  };

  introspection_linear[0].Enum.values = f0;
  introspection_linear[1].Enum.values = f0;
  introspection_linear[2].Enum.values = f0;
  introspection_linear[3].Enum.values = f0;
  introspection_linear[7].Struct.fields = f7;
  introspection_linear[13].Enum.values = f0;
  introspection_linear[14].Enum.values = f0;
  introspection_linear[15].Struct.fields = f15;
  introspection_linear[18].Struct.fields = f18;
  introspection_linear[19].Struct.fields = f19;
  introspection_linear[21].Struct.fields = f21;

  return 0;
}

void * get_p(const void * param, const char * name)
{
  dt_iop_liquify_params_t * p = (dt_iop_liquify_params_t*)param;
  if(!strcmp(name, "nodes[0].header.type"))
    return &(p->nodes[0].header.type);
  else if(!strcmp(name, "nodes[0].header.node_type"))
    return &(p->nodes[0].header.node_type);
  else if(!strcmp(name, "nodes[0].header.selected"))
    return &(p->nodes[0].header.selected);
  else if(!strcmp(name, "nodes[0].header.hovered"))
    return &(p->nodes[0].header.hovered);
  else if(!strcmp(name, "nodes[0].header.prev"))
    return &(p->nodes[0].header.prev);
  else if(!strcmp(name, "nodes[0].header.idx"))
    return &(p->nodes[0].header.idx);
  else if(!strcmp(name, "nodes[0].header.next"))
    return &(p->nodes[0].header.next);
  else if(!strcmp(name, "nodes[0].header"))
    return &(p->nodes[0].header);
  else if(!strcmp(name, "nodes[0].warp.point"))
    return &(p->nodes[0].warp.point);
  else if(!strcmp(name, "nodes[0].warp.strength"))
    return &(p->nodes[0].warp.strength);
  else if(!strcmp(name, "nodes[0].warp.radius"))
    return &(p->nodes[0].warp.radius);
  else if(!strcmp(name, "nodes[0].warp.control1"))
    return &(p->nodes[0].warp.control1);
  else if(!strcmp(name, "nodes[0].warp.control2"))
    return &(p->nodes[0].warp.control2);
  else if(!strcmp(name, "nodes[0].warp.type"))
    return &(p->nodes[0].warp.type);
  else if(!strcmp(name, "nodes[0].warp.status"))
    return &(p->nodes[0].warp.status);
  else if(!strcmp(name, "nodes[0].warp"))
    return &(p->nodes[0].warp);
  else if(!strcmp(name, "nodes[0].node.ctrl1"))
    return &(p->nodes[0].node.ctrl1);
  else if(!strcmp(name, "nodes[0].node.ctrl2"))
    return &(p->nodes[0].node.ctrl2);
  else if(!strcmp(name, "nodes[0].node"))
    return &(p->nodes[0].node);
  else if(!strcmp(name, "nodes[0]"))
    return &(p->nodes[0]);
  else if(!strcmp(name, "nodes"))
    return &(p->nodes);
  else
    return NULL;
}

dt_introspection_field_t * get_f(const char * name)
{
  if(!g_ascii_strcasecmp(name, "nodes[0].header.type"))
    return &(introspection_linear[0]);
  else if(!g_ascii_strcasecmp(name, "nodes[0].header.node_type"))
    return &(introspection_linear[1]);
  else if(!g_ascii_strcasecmp(name, "nodes[0].header.selected"))
    return &(introspection_linear[2]);
  else if(!g_ascii_strcasecmp(name, "nodes[0].header.hovered"))
    return &(introspection_linear[3]);
  else if(!g_ascii_strcasecmp(name, "nodes[0].header.prev"))
    return &(introspection_linear[4]);
  else if(!g_ascii_strcasecmp(name, "nodes[0].header.idx"))
    return &(introspection_linear[5]);
  else if(!g_ascii_strcasecmp(name, "nodes[0].header.next"))
    return &(introspection_linear[6]);
  else if(!g_ascii_strcasecmp(name, "nodes[0].header"))
    return &(introspection_linear[7]);
  else if(!g_ascii_strcasecmp(name, "nodes[0].warp.point"))
    return &(introspection_linear[8]);
  else if(!g_ascii_strcasecmp(name, "nodes[0].warp.strength"))
    return &(introspection_linear[9]);
  else if(!g_ascii_strcasecmp(name, "nodes[0].warp.radius"))
    return &(introspection_linear[10]);
  else if(!g_ascii_strcasecmp(name, "nodes[0].warp.control1"))
    return &(introspection_linear[11]);
  else if(!g_ascii_strcasecmp(name, "nodes[0].warp.control2"))
    return &(introspection_linear[12]);
  else if(!g_ascii_strcasecmp(name, "nodes[0].warp.type"))
    return &(introspection_linear[13]);
  else if(!g_ascii_strcasecmp(name, "nodes[0].warp.status"))
    return &(introspection_linear[14]);
  else if(!g_ascii_strcasecmp(name, "nodes[0].warp"))
    return &(introspection_linear[15]);
  else if(!g_ascii_strcasecmp(name, "nodes[0].node.ctrl1"))
    return &(introspection_linear[16]);
  else if(!g_ascii_strcasecmp(name, "nodes[0].node.ctrl2"))
    return &(introspection_linear[17]);
  else if(!g_ascii_strcasecmp(name, "nodes[0].node"))
    return &(introspection_linear[18]);
  else if(!g_ascii_strcasecmp(name, "nodes[0]"))
    return &(introspection_linear[19]);
  else if(!g_ascii_strcasecmp(name, "nodes"))
    return &(introspection_linear[20]);
  else
    return NULL;
}

#ifdef __cplusplus
}
#endif

