/*
 * this code is auto generated. do not edit. change the sources in tools/introspection/ instead.
 *
 * the parse tree:
 *
 * typedef
 *   struct dt_iop_levels_params_t
 *     enum dt_iop_levels_mode_t
 *       LEVELS_MODE_MANUAL : manual
 *       LEVELS_MODE_AUTOMATIC : automatic
 *     mode 
 *       default: LEVELS_MODE_MANUAL
 * 
 *     float
 *     black 
 *       default: 0.0
 *       range: 0.0 .. 100.0
 * 
 *     float
 *     gray 
 *       default: 50.0
 *       range: 0.0 .. 100.0
 * 
 *     float
 *     white 
 *       default: 100.0
 *       range: 0.0 .. 100.0
 * 
 *     float
 *     levels [3]
 *       default: 0.0
 *       range: -G_MAXFLOAT .. G_MAXFLOAT
 * 
 * dt_iop_levels_params_t
 *
 */

#include "/build/darktable/src/darktable-5.6.0/src/iop/levels.c"

#ifdef __cplusplus
extern "C"
{
#endif

#include <string.h>
#include "common/introspection.h"

static dt_introspection_field_t introspection_linear[7+1]
#if !( defined(__cplusplus) && !defined(__clang__) && defined(__GNUC__) && ((__GNUC__*100 + __GNUC_MINOR__) <= 406) )
= {
{
  .Enum = {
    { DT_INTROSPECTION_TYPE_ENUM, (char*)"dt_iop_levels_mode_t", (char*)"mode", (char*)"mode", (char*)"", sizeof(((dt_iop_levels_params_t*)NULL)->mode), G_STRUCT_OFFSET(dt_iop_levels_params_t, mode), NULL },
    /*entries*/ 2, /*values*/ NULL, /*Default*/ LEVELS_MODE_MANUAL
  }
},
{
  .Float = {
    { DT_INTROSPECTION_TYPE_FLOAT, (char*)"float", (char*)"black", (char*)"black", (char*)"", sizeof(((dt_iop_levels_params_t*)NULL)->black), G_STRUCT_OFFSET(dt_iop_levels_params_t, black), NULL },
    /*Min*/ 0.0, /*Max*/ 100.0, /*Default*/ 0.0
  }
},
{
  .Float = {
    { DT_INTROSPECTION_TYPE_FLOAT, (char*)"float", (char*)"gray", (char*)"gray", (char*)"", sizeof(((dt_iop_levels_params_t*)NULL)->gray), G_STRUCT_OFFSET(dt_iop_levels_params_t, gray), NULL },
    /*Min*/ 0.0, /*Max*/ 100.0, /*Default*/ 50.0
  }
},
{
  .Float = {
    { DT_INTROSPECTION_TYPE_FLOAT, (char*)"float", (char*)"white", (char*)"white", (char*)"", sizeof(((dt_iop_levels_params_t*)NULL)->white), G_STRUCT_OFFSET(dt_iop_levels_params_t, white), NULL },
    /*Min*/ 0.0, /*Max*/ 100.0, /*Default*/ 100.0
  }
},
{
  .Float = {
    { DT_INTROSPECTION_TYPE_FLOAT, (char*)"float", (char*)"levels[0]", (char*)"levels[0]", (char*)"", sizeof(((dt_iop_levels_params_t*)NULL)->levels[0]), G_STRUCT_OFFSET(dt_iop_levels_params_t, levels), NULL },
    /*Min*/ -G_MAXFLOAT, /*Max*/ G_MAXFLOAT, /*Default*/ 0.0
  }
},
{
  .Array = {
    { DT_INTROSPECTION_TYPE_ARRAY, (char*)"float[]", (char*)"levels", (char*)"levels", (char*)"", sizeof(((dt_iop_levels_params_t*)NULL)->levels), G_STRUCT_OFFSET(dt_iop_levels_params_t, levels), NULL },
    /*count*/ G_N_ELEMENTS(((dt_iop_levels_params_t*)NULL)->levels), /*type*/ DT_INTROSPECTION_TYPE_FLOAT, /*field*/ &introspection_linear[4]
  }
},
{
  .Struct = {
    { DT_INTROSPECTION_TYPE_STRUCT, (char*)"dt_iop_levels_params_t", (char*)"", (char*)"", (char*)"", sizeof(dt_iop_levels_params_t), 0, NULL },
    /*entries*/ 5, /*fields*/ NULL
  }
},
  { .header = {DT_INTROSPECTION_TYPE_NONE, NULL, NULL, NULL, NULL, 0, 0, NULL} }
}
#endif
;

static dt_introspection_t introspection = {
  8,
  2,
  "dt_iop_levels_params_t",
  sizeof(dt_iop_levels_params_t),
  &introspection_linear[7-1],
  sizeof(dt_iop_module_t),
  G_STRUCT_OFFSET(dt_iop_module_t, default_params)
};

dt_introspection_field_t* get_introspection_linear()
{
  return introspection_linear;
}

dt_introspection_t* get_introspection()
{
  return &introspection;
}

int introspection_init(struct dt_iop_module_so_t *self, int api_version)
{
  // here we check that the generated code matches the api at compile time and also at runtime
  if(introspection.api_version != DT_INTROSPECTION_VERSION || api_version != DT_INTROSPECTION_VERSION)
    return 1;

#if defined(__cplusplus) && !defined(__clang__) && defined(__GNUC__) && ((__GNUC__*100 + __GNUC_MINOR__) <= 406)
  introspection_linear[0].Enum = {
    { DT_INTROSPECTION_TYPE_ENUM, (char*)"dt_iop_levels_mode_t", (char*)"mode", (char*)"mode", (char*)"", sizeof(((dt_iop_levels_params_t*)NULL)->mode), G_STRUCT_OFFSET(dt_iop_levels_params_t, mode), NULL },
    /*entries*/ 2, /*values*/ NULL, /*Default*/ LEVELS_MODE_MANUAL
  };
  introspection_linear[1].Float = {
    { DT_INTROSPECTION_TYPE_FLOAT, (char*)"float", (char*)"black", (char*)"black", (char*)"", sizeof(((dt_iop_levels_params_t*)NULL)->black), G_STRUCT_OFFSET(dt_iop_levels_params_t, black), NULL },
    /*Min*/ 0.0, /*Max*/ 100.0, /*Default*/ 0.0
  };
  introspection_linear[2].Float = {
    { DT_INTROSPECTION_TYPE_FLOAT, (char*)"float", (char*)"gray", (char*)"gray", (char*)"", sizeof(((dt_iop_levels_params_t*)NULL)->gray), G_STRUCT_OFFSET(dt_iop_levels_params_t, gray), NULL },
    /*Min*/ 0.0, /*Max*/ 100.0, /*Default*/ 50.0
  };
  introspection_linear[3].Float = {
    { DT_INTROSPECTION_TYPE_FLOAT, (char*)"float", (char*)"white", (char*)"white", (char*)"", sizeof(((dt_iop_levels_params_t*)NULL)->white), G_STRUCT_OFFSET(dt_iop_levels_params_t, white), NULL },
    /*Min*/ 0.0, /*Max*/ 100.0, /*Default*/ 100.0
  };
  introspection_linear[4].Float = {
    { DT_INTROSPECTION_TYPE_FLOAT, (char*)"float", (char*)"levels[0]", (char*)"levels[0]", (char*)"", sizeof(((dt_iop_levels_params_t*)NULL)->levels[0]), G_STRUCT_OFFSET(dt_iop_levels_params_t, levels), NULL },
    /*Min*/ -G_MAXFLOAT, /*Max*/ G_MAXFLOAT, /*Default*/ 0.0
  };
  introspection_linear[5].Array = {
    { DT_INTROSPECTION_TYPE_ARRAY, (char*)"float[]", (char*)"levels", (char*)"levels", (char*)"", sizeof(((dt_iop_levels_params_t*)NULL)->levels), G_STRUCT_OFFSET(dt_iop_levels_params_t, levels), NULL },
    /*count*/ G_N_ELEMENTS(((dt_iop_levels_params_t*)NULL)->levels), /*type*/ DT_INTROSPECTION_TYPE_FLOAT, /*field*/ &introspection_linear[4]
  };
  introspection_linear[6].Struct = {
    { DT_INTROSPECTION_TYPE_STRUCT, (char*)"dt_iop_levels_params_t", (char*)"", (char*)"", (char*)"", sizeof(dt_iop_levels_params_t), 0, NULL },
    /*entries*/ 5, /*fields*/ NULL
  };
  introspection_linear[7].header = {DT_INTROSPECTION_TYPE_NONE, NULL, NULL, NULL, NULL, 0, 0, NULL};
#endif

  for(int i = 0; i <= 7; i++)
    introspection_linear[i].header.so = self;

  static dt_introspection_type_enum_tuple_t f0[] = { 
    { "LEVELS_MODE_MANUAL", LEVELS_MODE_MANUAL, N_((char*)"manual") },
    { "LEVELS_MODE_AUTOMATIC", LEVELS_MODE_AUTOMATIC, N_((char*)"automatic") },
    { NULL, 0 },
  };
  static dt_introspection_field_t *f6[] = {
    &introspection_linear[0],
    &introspection_linear[1],
    &introspection_linear[2],
    &introspection_linear[3],
    &introspection_linear[5],
    NULL
  };

  introspection_linear[0].Enum.values = f0;
  introspection_linear[6].Struct.fields = f6;

  return 0;
}

void * get_p(const void * param, const char * name)
{
  dt_iop_levels_params_t * p = (dt_iop_levels_params_t*)param;
  if(!strcmp(name, "mode"))
    return &(p->mode);
  else if(!strcmp(name, "black"))
    return &(p->black);
  else if(!strcmp(name, "gray"))
    return &(p->gray);
  else if(!strcmp(name, "white"))
    return &(p->white);
  else if(!strcmp(name, "levels[0]"))
    return &(p->levels[0]);
  else if(!strcmp(name, "levels"))
    return &(p->levels);
  else
    return NULL;
}

dt_introspection_field_t * get_f(const char * name)
{
  if(!g_ascii_strcasecmp(name, "mode"))
    return &(introspection_linear[0]);
  else if(!g_ascii_strcasecmp(name, "black"))
    return &(introspection_linear[1]);
  else if(!g_ascii_strcasecmp(name, "gray"))
    return &(introspection_linear[2]);
  else if(!g_ascii_strcasecmp(name, "white"))
    return &(introspection_linear[3]);
  else if(!g_ascii_strcasecmp(name, "levels[0]"))
    return &(introspection_linear[4]);
  else if(!g_ascii_strcasecmp(name, "levels"))
    return &(introspection_linear[5]);
  else
    return NULL;
}

#ifdef __cplusplus
}
#endif

