/*
 * this code is auto generated. do not edit. change the sources in tools/introspection/ instead.
 *
 * the parse tree:
 *
 * typedef
 *   struct dt_iop_channelmixer_rgb_params_t
 *     float
 *     red [CHANNEL_SIZE]
 *       default: 0.0
 *       range: COLOR_MIN .. COLOR_MAX
 * 
 *     float
 *     green [CHANNEL_SIZE]
 *       default: 0.0
 *       range: COLOR_MIN .. COLOR_MAX
 * 
 *     float
 *     blue [CHANNEL_SIZE]
 *       default: 0.0
 *       range: COLOR_MIN .. COLOR_MAX
 * 
 *     float
 *     saturation [CHANNEL_SIZE]
 *       default: 0.0
 *       range: -2.0 .. 2.0
 * 
 *     float
 *     lightness [CHANNEL_SIZE]
 *       default: 0.0
 *       range: -2.0 .. 2.0
 * 
 *     float
 *     grey [CHANNEL_SIZE]
 *       default: 0.0
 *       range: -2.0 .. 2.0
 * 
 *     boolean
 *     normalize_R 
 *       description: normalize channels
 *       default: FALSE
 * 
 *     boolean
 *     normalize_G 
 *       description: normalize channels
 *       default: FALSE
 * 
 *     boolean
 *     normalize_B 
 *       description: normalize channels
 *       default: FALSE
 * 
 *     boolean
 *     normalize_sat 
 *       description: normalize channels
 *       default: FALSE
 * 
 *     boolean
 *     normalize_light 
 *       description: normalize channels
 *       default: FALSE
 * 
 *     boolean
 *     normalize_grey 
 *       description: normalize channels
 *       default: FALSE
 * 
 *     enum dt_illuminant_t
 *       DT_ILLUMINANT_PIPE : same as pipeline (D50)
 *       DT_ILLUMINANT_A : A (incandescent)
 *       DT_ILLUMINANT_D : D (daylight)
 *       DT_ILLUMINANT_E : E (equi-energy)
 *       DT_ILLUMINANT_F : F (fluorescent)
 *       DT_ILLUMINANT_LED : LED (LED light)
 *       DT_ILLUMINANT_BB : Planckian (black body)
 *       DT_ILLUMINANT_CUSTOM : custom
 *       DT_ILLUMINANT_CAMERA : as shot in camera
 *       DT_ILLUMINANT_LAST
 *       DT_ILLUMINANT_DETECT_SURFACES
 *       DT_ILLUMINANT_DETECT_EDGES
 *     illuminant 
 *       default: DT_ILLUMINANT_D
 * 
 *     enum dt_illuminant_fluo_t
 *       DT_ILLUMINANT_FLUO_F1 : F1 (Daylight 6430 K) – medium CRI
 *       DT_ILLUMINANT_FLUO_F2 : F2 (Cool White 4230 K) – medium CRI
 *       DT_ILLUMINANT_FLUO_F3 : F3 (White 3450 K) – medium CRI
 *       DT_ILLUMINANT_FLUO_F4 : F4 (Warm White 2940 K) – medium CRI
 *       DT_ILLUMINANT_FLUO_F5 : F5 (Daylight 6350 K) – medium CRI
 *       DT_ILLUMINANT_FLUO_F6 : F6 (Lite White 4150 K) – medium CRI
 *       DT_ILLUMINANT_FLUO_F7 : F7 (D65 simulator 6500 K) – high CRI
 *       DT_ILLUMINANT_FLUO_F8 : F8 (D50 simulator 5000 K) – high CRI
 *       DT_ILLUMINANT_FLUO_F9 : F9 (Cool White Deluxe 4150 K) – high CRI
 *       DT_ILLUMINANT_FLUO_F10 : F10 (Tuned RGB 5000 K) – low CRI
 *       DT_ILLUMINANT_FLUO_F11 : F11 (Tuned RGB 4000 K) – low CRI
 *       DT_ILLUMINANT_FLUO_F12 : F12 (Tuned RGB 3000 K) – low CRI
 *       DT_ILLUMINANT_FLUO_LAST
 *     illum_fluo 
 *       description: F source
 *       default: DT_ILLUMINANT_FLUO_F3
 * 
 *     enum dt_illuminant_led_t
 *       DT_ILLUMINANT_LED_B1 : B1 (Blue 2733 K)
 *       DT_ILLUMINANT_LED_B2 : B2 (Blue 2998 K)
 *       DT_ILLUMINANT_LED_B3 : B3 (Blue 4103 K)
 *       DT_ILLUMINANT_LED_B4 : B4 (Blue 5109 K)
 *       DT_ILLUMINANT_LED_B5 : B5 (Blue 6598 K)
 *       DT_ILLUMINANT_LED_BH1 : BH1 (Blue-Red hybrid 2851 K)
 *       DT_ILLUMINANT_LED_RGB1 : RGB1 (RGB 2840 K)
 *       DT_ILLUMINANT_LED_V1 : V1 (Violet 2724 K)
 *       DT_ILLUMINANT_LED_V2 : V2 (Violet 4070 K)
 *       DT_ILLUMINANT_LED_LAST
 *     illum_led 
 *       description: LED source
 *       default: DT_ILLUMINANT_LED_B5
 * 
 *     enum dt_adaptation_t
 *       DT_ADAPTATION_LINEAR_BRADFORD : linear Bradford (ICC v4)
 *       DT_ADAPTATION_CAT16 : CAT16 (CIECAM16)
 *       DT_ADAPTATION_FULL_BRADFORD : non-linear Bradford
 *       DT_ADAPTATION_XYZ : XYZ
 *       DT_ADAPTATION_RGB : none (bypass)
 *       DT_ADAPTATION_LAST
 *     adaptation 
 *       default: DT_ADAPTATION_CAT16
 * 
 *     float
 *     x 
 *       default: 0.333
 *       range: -G_MAXFLOAT .. G_MAXFLOAT
 * 
 *     float
 *     y 
 *       default: 0.333
 *       range: -G_MAXFLOAT .. G_MAXFLOAT
 * 
 *     float
 *     temperature 
 *       default: 5003.
 *       range: TEMP_MIN .. TEMP_MAX
 * 
 *     float
 *     gamut 
 *       description: gamut compression
 *       default: 1.0
 *       range: 0.0 .. 12.0
 * 
 *     boolean
 *     clip 
 *       description: clip negative RGB from gamut
 *       default: TRUE
 * 
 *     enum dt_iop_channelmixer_rgb_version_t
 *       CHANNELMIXERRGB_V_1 : version 1 (2020)
 *       CHANNELMIXERRGB_V_2 : version 2 (2021)
 *       CHANNELMIXERRGB_V_3 : version 3 (Apr 2021)
 *     version 
 *       description: saturation algorithm
 *       default: CHANNELMIXERRGB_V_3
 * 
 * dt_iop_channelmixer_rgb_params_t
 *
 */

#include "/build/darktable/src/darktable-5.6.0/src/iop/channelmixerrgb.c"

#ifdef __cplusplus
extern "C"
{
#endif

#include <string.h>
#include "common/introspection.h"

static dt_introspection_field_t introspection_linear[29+1]
#if !( defined(__cplusplus) && !defined(__clang__) && defined(__GNUC__) && ((__GNUC__*100 + __GNUC_MINOR__) <= 406) )
= {
{
  .Float = {
    { DT_INTROSPECTION_TYPE_FLOAT, (char*)"float", (char*)"red[0]", (char*)"red[0]", (char*)"", sizeof(((dt_iop_channelmixer_rgb_params_t*)NULL)->red[0]), G_STRUCT_OFFSET(dt_iop_channelmixer_rgb_params_t, red), NULL },
    /*Min*/ COLOR_MIN, /*Max*/ COLOR_MAX, /*Default*/ 0.0
  }
},
{
  .Array = {
    { DT_INTROSPECTION_TYPE_ARRAY, (char*)"float[]", (char*)"red", (char*)"red", (char*)"", sizeof(((dt_iop_channelmixer_rgb_params_t*)NULL)->red), G_STRUCT_OFFSET(dt_iop_channelmixer_rgb_params_t, red), NULL },
    /*count*/ G_N_ELEMENTS(((dt_iop_channelmixer_rgb_params_t*)NULL)->red), /*type*/ DT_INTROSPECTION_TYPE_FLOAT, /*field*/ &introspection_linear[0]
  }
},
{
  .Float = {
    { DT_INTROSPECTION_TYPE_FLOAT, (char*)"float", (char*)"green[0]", (char*)"green[0]", (char*)"", sizeof(((dt_iop_channelmixer_rgb_params_t*)NULL)->green[0]), G_STRUCT_OFFSET(dt_iop_channelmixer_rgb_params_t, green), NULL },
    /*Min*/ COLOR_MIN, /*Max*/ COLOR_MAX, /*Default*/ 0.0
  }
},
{
  .Array = {
    { DT_INTROSPECTION_TYPE_ARRAY, (char*)"float[]", (char*)"green", (char*)"green", (char*)"", sizeof(((dt_iop_channelmixer_rgb_params_t*)NULL)->green), G_STRUCT_OFFSET(dt_iop_channelmixer_rgb_params_t, green), NULL },
    /*count*/ G_N_ELEMENTS(((dt_iop_channelmixer_rgb_params_t*)NULL)->green), /*type*/ DT_INTROSPECTION_TYPE_FLOAT, /*field*/ &introspection_linear[2]
  }
},
{
  .Float = {
    { DT_INTROSPECTION_TYPE_FLOAT, (char*)"float", (char*)"blue[0]", (char*)"blue[0]", (char*)"", sizeof(((dt_iop_channelmixer_rgb_params_t*)NULL)->blue[0]), G_STRUCT_OFFSET(dt_iop_channelmixer_rgb_params_t, blue), NULL },
    /*Min*/ COLOR_MIN, /*Max*/ COLOR_MAX, /*Default*/ 0.0
  }
},
{
  .Array = {
    { DT_INTROSPECTION_TYPE_ARRAY, (char*)"float[]", (char*)"blue", (char*)"blue", (char*)"", sizeof(((dt_iop_channelmixer_rgb_params_t*)NULL)->blue), G_STRUCT_OFFSET(dt_iop_channelmixer_rgb_params_t, blue), NULL },
    /*count*/ G_N_ELEMENTS(((dt_iop_channelmixer_rgb_params_t*)NULL)->blue), /*type*/ DT_INTROSPECTION_TYPE_FLOAT, /*field*/ &introspection_linear[4]
  }
},
{
  .Float = {
    { DT_INTROSPECTION_TYPE_FLOAT, (char*)"float", (char*)"saturation[0]", (char*)"saturation[0]", (char*)"", sizeof(((dt_iop_channelmixer_rgb_params_t*)NULL)->saturation[0]), G_STRUCT_OFFSET(dt_iop_channelmixer_rgb_params_t, saturation), NULL },
    /*Min*/ -2.0, /*Max*/ 2.0, /*Default*/ 0.0
  }
},
{
  .Array = {
    { DT_INTROSPECTION_TYPE_ARRAY, (char*)"float[]", (char*)"saturation", (char*)"saturation", (char*)"", sizeof(((dt_iop_channelmixer_rgb_params_t*)NULL)->saturation), G_STRUCT_OFFSET(dt_iop_channelmixer_rgb_params_t, saturation), NULL },
    /*count*/ G_N_ELEMENTS(((dt_iop_channelmixer_rgb_params_t*)NULL)->saturation), /*type*/ DT_INTROSPECTION_TYPE_FLOAT, /*field*/ &introspection_linear[6]
  }
},
{
  .Float = {
    { DT_INTROSPECTION_TYPE_FLOAT, (char*)"float", (char*)"lightness[0]", (char*)"lightness[0]", (char*)"", sizeof(((dt_iop_channelmixer_rgb_params_t*)NULL)->lightness[0]), G_STRUCT_OFFSET(dt_iop_channelmixer_rgb_params_t, lightness), NULL },
    /*Min*/ -2.0, /*Max*/ 2.0, /*Default*/ 0.0
  }
},
{
  .Array = {
    { DT_INTROSPECTION_TYPE_ARRAY, (char*)"float[]", (char*)"lightness", (char*)"lightness", (char*)"", sizeof(((dt_iop_channelmixer_rgb_params_t*)NULL)->lightness), G_STRUCT_OFFSET(dt_iop_channelmixer_rgb_params_t, lightness), NULL },
    /*count*/ G_N_ELEMENTS(((dt_iop_channelmixer_rgb_params_t*)NULL)->lightness), /*type*/ DT_INTROSPECTION_TYPE_FLOAT, /*field*/ &introspection_linear[8]
  }
},
{
  .Float = {
    { DT_INTROSPECTION_TYPE_FLOAT, (char*)"float", (char*)"grey[0]", (char*)"grey[0]", (char*)"", sizeof(((dt_iop_channelmixer_rgb_params_t*)NULL)->grey[0]), G_STRUCT_OFFSET(dt_iop_channelmixer_rgb_params_t, grey), NULL },
    /*Min*/ -2.0, /*Max*/ 2.0, /*Default*/ 0.0
  }
},
{
  .Array = {
    { DT_INTROSPECTION_TYPE_ARRAY, (char*)"float[]", (char*)"grey", (char*)"grey", (char*)"", sizeof(((dt_iop_channelmixer_rgb_params_t*)NULL)->grey), G_STRUCT_OFFSET(dt_iop_channelmixer_rgb_params_t, grey), NULL },
    /*count*/ G_N_ELEMENTS(((dt_iop_channelmixer_rgb_params_t*)NULL)->grey), /*type*/ DT_INTROSPECTION_TYPE_FLOAT, /*field*/ &introspection_linear[10]
  }
},
{
  .Bool = {
    { DT_INTROSPECTION_TYPE_BOOL, (char*)"bool", (char*)"normalize_R", (char*)"normalize_R", N_((char*)"normalize channels"), sizeof(((dt_iop_channelmixer_rgb_params_t*)NULL)->normalize_R), G_STRUCT_OFFSET(dt_iop_channelmixer_rgb_params_t, normalize_R), NULL },
    /*Default*/ FALSE
  }
},
{
  .Bool = {
    { DT_INTROSPECTION_TYPE_BOOL, (char*)"bool", (char*)"normalize_G", (char*)"normalize_G", N_((char*)"normalize channels"), sizeof(((dt_iop_channelmixer_rgb_params_t*)NULL)->normalize_G), G_STRUCT_OFFSET(dt_iop_channelmixer_rgb_params_t, normalize_G), NULL },
    /*Default*/ FALSE
  }
},
{
  .Bool = {
    { DT_INTROSPECTION_TYPE_BOOL, (char*)"bool", (char*)"normalize_B", (char*)"normalize_B", N_((char*)"normalize channels"), sizeof(((dt_iop_channelmixer_rgb_params_t*)NULL)->normalize_B), G_STRUCT_OFFSET(dt_iop_channelmixer_rgb_params_t, normalize_B), NULL },
    /*Default*/ FALSE
  }
},
{
  .Bool = {
    { DT_INTROSPECTION_TYPE_BOOL, (char*)"bool", (char*)"normalize_sat", (char*)"normalize_sat", N_((char*)"normalize channels"), sizeof(((dt_iop_channelmixer_rgb_params_t*)NULL)->normalize_sat), G_STRUCT_OFFSET(dt_iop_channelmixer_rgb_params_t, normalize_sat), NULL },
    /*Default*/ FALSE
  }
},
{
  .Bool = {
    { DT_INTROSPECTION_TYPE_BOOL, (char*)"bool", (char*)"normalize_light", (char*)"normalize_light", N_((char*)"normalize channels"), sizeof(((dt_iop_channelmixer_rgb_params_t*)NULL)->normalize_light), G_STRUCT_OFFSET(dt_iop_channelmixer_rgb_params_t, normalize_light), NULL },
    /*Default*/ FALSE
  }
},
{
  .Bool = {
    { DT_INTROSPECTION_TYPE_BOOL, (char*)"bool", (char*)"normalize_grey", (char*)"normalize_grey", N_((char*)"normalize channels"), sizeof(((dt_iop_channelmixer_rgb_params_t*)NULL)->normalize_grey), G_STRUCT_OFFSET(dt_iop_channelmixer_rgb_params_t, normalize_grey), NULL },
    /*Default*/ FALSE
  }
},
{
  .Enum = {
    { DT_INTROSPECTION_TYPE_ENUM, (char*)"dt_illuminant_t", (char*)"illuminant", (char*)"illuminant", (char*)"", sizeof(((dt_iop_channelmixer_rgb_params_t*)NULL)->illuminant), G_STRUCT_OFFSET(dt_iop_channelmixer_rgb_params_t, illuminant), NULL },
    /*entries*/ 12, /*values*/ NULL, /*Default*/ DT_ILLUMINANT_D
  }
},
{
  .Enum = {
    { DT_INTROSPECTION_TYPE_ENUM, (char*)"dt_illuminant_fluo_t", (char*)"illum_fluo", (char*)"illum_fluo", N_((char*)"F source"), sizeof(((dt_iop_channelmixer_rgb_params_t*)NULL)->illum_fluo), G_STRUCT_OFFSET(dt_iop_channelmixer_rgb_params_t, illum_fluo), NULL },
    /*entries*/ 13, /*values*/ NULL, /*Default*/ DT_ILLUMINANT_FLUO_F3
  }
},
{
  .Enum = {
    { DT_INTROSPECTION_TYPE_ENUM, (char*)"dt_illuminant_led_t", (char*)"illum_led", (char*)"illum_led", N_((char*)"LED source"), sizeof(((dt_iop_channelmixer_rgb_params_t*)NULL)->illum_led), G_STRUCT_OFFSET(dt_iop_channelmixer_rgb_params_t, illum_led), NULL },
    /*entries*/ 10, /*values*/ NULL, /*Default*/ DT_ILLUMINANT_LED_B5
  }
},
{
  .Enum = {
    { DT_INTROSPECTION_TYPE_ENUM, (char*)"dt_adaptation_t", (char*)"adaptation", (char*)"adaptation", (char*)"", sizeof(((dt_iop_channelmixer_rgb_params_t*)NULL)->adaptation), G_STRUCT_OFFSET(dt_iop_channelmixer_rgb_params_t, adaptation), NULL },
    /*entries*/ 6, /*values*/ NULL, /*Default*/ DT_ADAPTATION_CAT16
  }
},
{
  .Float = {
    { DT_INTROSPECTION_TYPE_FLOAT, (char*)"float", (char*)"x", (char*)"x", (char*)"", sizeof(((dt_iop_channelmixer_rgb_params_t*)NULL)->x), G_STRUCT_OFFSET(dt_iop_channelmixer_rgb_params_t, x), NULL },
    /*Min*/ -G_MAXFLOAT, /*Max*/ G_MAXFLOAT, /*Default*/ 0.333
  }
},
{
  .Float = {
    { DT_INTROSPECTION_TYPE_FLOAT, (char*)"float", (char*)"y", (char*)"y", (char*)"", sizeof(((dt_iop_channelmixer_rgb_params_t*)NULL)->y), G_STRUCT_OFFSET(dt_iop_channelmixer_rgb_params_t, y), NULL },
    /*Min*/ -G_MAXFLOAT, /*Max*/ G_MAXFLOAT, /*Default*/ 0.333
  }
},
{
  .Float = {
    { DT_INTROSPECTION_TYPE_FLOAT, (char*)"float", (char*)"temperature", (char*)"temperature", (char*)"", sizeof(((dt_iop_channelmixer_rgb_params_t*)NULL)->temperature), G_STRUCT_OFFSET(dt_iop_channelmixer_rgb_params_t, temperature), NULL },
    /*Min*/ TEMP_MIN, /*Max*/ TEMP_MAX, /*Default*/ 5003.
  }
},
{
  .Float = {
    { DT_INTROSPECTION_TYPE_FLOAT, (char*)"float", (char*)"gamut", (char*)"gamut", N_((char*)"gamut compression"), sizeof(((dt_iop_channelmixer_rgb_params_t*)NULL)->gamut), G_STRUCT_OFFSET(dt_iop_channelmixer_rgb_params_t, gamut), NULL },
    /*Min*/ 0.0, /*Max*/ 12.0, /*Default*/ 1.0
  }
},
{
  .Bool = {
    { DT_INTROSPECTION_TYPE_BOOL, (char*)"bool", (char*)"clip", (char*)"clip", N_((char*)"clip negative RGB from gamut"), sizeof(((dt_iop_channelmixer_rgb_params_t*)NULL)->clip), G_STRUCT_OFFSET(dt_iop_channelmixer_rgb_params_t, clip), NULL },
    /*Default*/ TRUE
  }
},
{
  .Enum = {
    { DT_INTROSPECTION_TYPE_ENUM, (char*)"dt_iop_channelmixer_rgb_version_t", (char*)"version", (char*)"version", N_((char*)"saturation algorithm"), sizeof(((dt_iop_channelmixer_rgb_params_t*)NULL)->version), G_STRUCT_OFFSET(dt_iop_channelmixer_rgb_params_t, version), NULL },
    /*entries*/ 3, /*values*/ NULL, /*Default*/ CHANNELMIXERRGB_V_3
  }
},
{
  .Struct = {
    { DT_INTROSPECTION_TYPE_STRUCT, (char*)"dt_iop_channelmixer_rgb_params_t", (char*)"", (char*)"", (char*)"", sizeof(dt_iop_channelmixer_rgb_params_t), 0, NULL },
    /*entries*/ 22, /*fields*/ NULL
  }
},
  { .header = {DT_INTROSPECTION_TYPE_NONE, NULL, NULL, NULL, NULL, 0, 0, NULL} }
}
#endif
;

static dt_introspection_t introspection = {
  8,
  3,
  "dt_iop_channelmixer_rgb_params_t",
  sizeof(dt_iop_channelmixer_rgb_params_t),
  &introspection_linear[29-1],
  sizeof(dt_iop_module_t),
  G_STRUCT_OFFSET(dt_iop_module_t, default_params)
};

dt_introspection_field_t* get_introspection_linear()
{
  return introspection_linear;
}

dt_introspection_t* get_introspection()
{
  return &introspection;
}

int introspection_init(struct dt_iop_module_so_t *self, int api_version)
{
  // here we check that the generated code matches the api at compile time and also at runtime
  if(introspection.api_version != DT_INTROSPECTION_VERSION || api_version != DT_INTROSPECTION_VERSION)
    return 1;

#if defined(__cplusplus) && !defined(__clang__) && defined(__GNUC__) && ((__GNUC__*100 + __GNUC_MINOR__) <= 406)
  introspection_linear[0].Float = {
    { DT_INTROSPECTION_TYPE_FLOAT, (char*)"float", (char*)"red[0]", (char*)"red[0]", (char*)"", sizeof(((dt_iop_channelmixer_rgb_params_t*)NULL)->red[0]), G_STRUCT_OFFSET(dt_iop_channelmixer_rgb_params_t, red), NULL },
    /*Min*/ COLOR_MIN, /*Max*/ COLOR_MAX, /*Default*/ 0.0
  };
  introspection_linear[1].Array = {
    { DT_INTROSPECTION_TYPE_ARRAY, (char*)"float[]", (char*)"red", (char*)"red", (char*)"", sizeof(((dt_iop_channelmixer_rgb_params_t*)NULL)->red), G_STRUCT_OFFSET(dt_iop_channelmixer_rgb_params_t, red), NULL },
    /*count*/ G_N_ELEMENTS(((dt_iop_channelmixer_rgb_params_t*)NULL)->red), /*type*/ DT_INTROSPECTION_TYPE_FLOAT, /*field*/ &introspection_linear[0]
  };
  introspection_linear[2].Float = {
    { DT_INTROSPECTION_TYPE_FLOAT, (char*)"float", (char*)"green[0]", (char*)"green[0]", (char*)"", sizeof(((dt_iop_channelmixer_rgb_params_t*)NULL)->green[0]), G_STRUCT_OFFSET(dt_iop_channelmixer_rgb_params_t, green), NULL },
    /*Min*/ COLOR_MIN, /*Max*/ COLOR_MAX, /*Default*/ 0.0
  };
  introspection_linear[3].Array = {
    { DT_INTROSPECTION_TYPE_ARRAY, (char*)"float[]", (char*)"green", (char*)"green", (char*)"", sizeof(((dt_iop_channelmixer_rgb_params_t*)NULL)->green), G_STRUCT_OFFSET(dt_iop_channelmixer_rgb_params_t, green), NULL },
    /*count*/ G_N_ELEMENTS(((dt_iop_channelmixer_rgb_params_t*)NULL)->green), /*type*/ DT_INTROSPECTION_TYPE_FLOAT, /*field*/ &introspection_linear[2]
  };
  introspection_linear[4].Float = {
    { DT_INTROSPECTION_TYPE_FLOAT, (char*)"float", (char*)"blue[0]", (char*)"blue[0]", (char*)"", sizeof(((dt_iop_channelmixer_rgb_params_t*)NULL)->blue[0]), G_STRUCT_OFFSET(dt_iop_channelmixer_rgb_params_t, blue), NULL },
    /*Min*/ COLOR_MIN, /*Max*/ COLOR_MAX, /*Default*/ 0.0
  };
  introspection_linear[5].Array = {
    { DT_INTROSPECTION_TYPE_ARRAY, (char*)"float[]", (char*)"blue", (char*)"blue", (char*)"", sizeof(((dt_iop_channelmixer_rgb_params_t*)NULL)->blue), G_STRUCT_OFFSET(dt_iop_channelmixer_rgb_params_t, blue), NULL },
    /*count*/ G_N_ELEMENTS(((dt_iop_channelmixer_rgb_params_t*)NULL)->blue), /*type*/ DT_INTROSPECTION_TYPE_FLOAT, /*field*/ &introspection_linear[4]
  };
  introspection_linear[6].Float = {
    { DT_INTROSPECTION_TYPE_FLOAT, (char*)"float", (char*)"saturation[0]", (char*)"saturation[0]", (char*)"", sizeof(((dt_iop_channelmixer_rgb_params_t*)NULL)->saturation[0]), G_STRUCT_OFFSET(dt_iop_channelmixer_rgb_params_t, saturation), NULL },
    /*Min*/ -2.0, /*Max*/ 2.0, /*Default*/ 0.0
  };
  introspection_linear[7].Array = {
    { DT_INTROSPECTION_TYPE_ARRAY, (char*)"float[]", (char*)"saturation", (char*)"saturation", (char*)"", sizeof(((dt_iop_channelmixer_rgb_params_t*)NULL)->saturation), G_STRUCT_OFFSET(dt_iop_channelmixer_rgb_params_t, saturation), NULL },
    /*count*/ G_N_ELEMENTS(((dt_iop_channelmixer_rgb_params_t*)NULL)->saturation), /*type*/ DT_INTROSPECTION_TYPE_FLOAT, /*field*/ &introspection_linear[6]
  };
  introspection_linear[8].Float = {
    { DT_INTROSPECTION_TYPE_FLOAT, (char*)"float", (char*)"lightness[0]", (char*)"lightness[0]", (char*)"", sizeof(((dt_iop_channelmixer_rgb_params_t*)NULL)->lightness[0]), G_STRUCT_OFFSET(dt_iop_channelmixer_rgb_params_t, lightness), NULL },
    /*Min*/ -2.0, /*Max*/ 2.0, /*Default*/ 0.0
  };
  introspection_linear[9].Array = {
    { DT_INTROSPECTION_TYPE_ARRAY, (char*)"float[]", (char*)"lightness", (char*)"lightness", (char*)"", sizeof(((dt_iop_channelmixer_rgb_params_t*)NULL)->lightness), G_STRUCT_OFFSET(dt_iop_channelmixer_rgb_params_t, lightness), NULL },
    /*count*/ G_N_ELEMENTS(((dt_iop_channelmixer_rgb_params_t*)NULL)->lightness), /*type*/ DT_INTROSPECTION_TYPE_FLOAT, /*field*/ &introspection_linear[8]
  };
  introspection_linear[10].Float = {
    { DT_INTROSPECTION_TYPE_FLOAT, (char*)"float", (char*)"grey[0]", (char*)"grey[0]", (char*)"", sizeof(((dt_iop_channelmixer_rgb_params_t*)NULL)->grey[0]), G_STRUCT_OFFSET(dt_iop_channelmixer_rgb_params_t, grey), NULL },
    /*Min*/ -2.0, /*Max*/ 2.0, /*Default*/ 0.0
  };
  introspection_linear[11].Array = {
    { DT_INTROSPECTION_TYPE_ARRAY, (char*)"float[]", (char*)"grey", (char*)"grey", (char*)"", sizeof(((dt_iop_channelmixer_rgb_params_t*)NULL)->grey), G_STRUCT_OFFSET(dt_iop_channelmixer_rgb_params_t, grey), NULL },
    /*count*/ G_N_ELEMENTS(((dt_iop_channelmixer_rgb_params_t*)NULL)->grey), /*type*/ DT_INTROSPECTION_TYPE_FLOAT, /*field*/ &introspection_linear[10]
  };
  introspection_linear[12].Bool = {
    { DT_INTROSPECTION_TYPE_BOOL, (char*)"bool", (char*)"normalize_R", (char*)"normalize_R", N_((char*)"normalize channels"), sizeof(((dt_iop_channelmixer_rgb_params_t*)NULL)->normalize_R), G_STRUCT_OFFSET(dt_iop_channelmixer_rgb_params_t, normalize_R), NULL },
    /*Default*/ FALSE
  };
  introspection_linear[13].Bool = {
    { DT_INTROSPECTION_TYPE_BOOL, (char*)"bool", (char*)"normalize_G", (char*)"normalize_G", N_((char*)"normalize channels"), sizeof(((dt_iop_channelmixer_rgb_params_t*)NULL)->normalize_G), G_STRUCT_OFFSET(dt_iop_channelmixer_rgb_params_t, normalize_G), NULL },
    /*Default*/ FALSE
  };
  introspection_linear[14].Bool = {
    { DT_INTROSPECTION_TYPE_BOOL, (char*)"bool", (char*)"normalize_B", (char*)"normalize_B", N_((char*)"normalize channels"), sizeof(((dt_iop_channelmixer_rgb_params_t*)NULL)->normalize_B), G_STRUCT_OFFSET(dt_iop_channelmixer_rgb_params_t, normalize_B), NULL },
    /*Default*/ FALSE
  };
  introspection_linear[15].Bool = {
    { DT_INTROSPECTION_TYPE_BOOL, (char*)"bool", (char*)"normalize_sat", (char*)"normalize_sat", N_((char*)"normalize channels"), sizeof(((dt_iop_channelmixer_rgb_params_t*)NULL)->normalize_sat), G_STRUCT_OFFSET(dt_iop_channelmixer_rgb_params_t, normalize_sat), NULL },
    /*Default*/ FALSE
  };
  introspection_linear[16].Bool = {
    { DT_INTROSPECTION_TYPE_BOOL, (char*)"bool", (char*)"normalize_light", (char*)"normalize_light", N_((char*)"normalize channels"), sizeof(((dt_iop_channelmixer_rgb_params_t*)NULL)->normalize_light), G_STRUCT_OFFSET(dt_iop_channelmixer_rgb_params_t, normalize_light), NULL },
    /*Default*/ FALSE
  };
  introspection_linear[17].Bool = {
    { DT_INTROSPECTION_TYPE_BOOL, (char*)"bool", (char*)"normalize_grey", (char*)"normalize_grey", N_((char*)"normalize channels"), sizeof(((dt_iop_channelmixer_rgb_params_t*)NULL)->normalize_grey), G_STRUCT_OFFSET(dt_iop_channelmixer_rgb_params_t, normalize_grey), NULL },
    /*Default*/ FALSE
  };
  introspection_linear[18].Enum = {
    { DT_INTROSPECTION_TYPE_ENUM, (char*)"dt_illuminant_t", (char*)"illuminant", (char*)"illuminant", (char*)"", sizeof(((dt_iop_channelmixer_rgb_params_t*)NULL)->illuminant), G_STRUCT_OFFSET(dt_iop_channelmixer_rgb_params_t, illuminant), NULL },
    /*entries*/ 12, /*values*/ NULL, /*Default*/ DT_ILLUMINANT_D
  };
  introspection_linear[19].Enum = {
    { DT_INTROSPECTION_TYPE_ENUM, (char*)"dt_illuminant_fluo_t", (char*)"illum_fluo", (char*)"illum_fluo", N_((char*)"F source"), sizeof(((dt_iop_channelmixer_rgb_params_t*)NULL)->illum_fluo), G_STRUCT_OFFSET(dt_iop_channelmixer_rgb_params_t, illum_fluo), NULL },
    /*entries*/ 13, /*values*/ NULL, /*Default*/ DT_ILLUMINANT_FLUO_F3
  };
  introspection_linear[20].Enum = {
    { DT_INTROSPECTION_TYPE_ENUM, (char*)"dt_illuminant_led_t", (char*)"illum_led", (char*)"illum_led", N_((char*)"LED source"), sizeof(((dt_iop_channelmixer_rgb_params_t*)NULL)->illum_led), G_STRUCT_OFFSET(dt_iop_channelmixer_rgb_params_t, illum_led), NULL },
    /*entries*/ 10, /*values*/ NULL, /*Default*/ DT_ILLUMINANT_LED_B5
  };
  introspection_linear[21].Enum = {
    { DT_INTROSPECTION_TYPE_ENUM, (char*)"dt_adaptation_t", (char*)"adaptation", (char*)"adaptation", (char*)"", sizeof(((dt_iop_channelmixer_rgb_params_t*)NULL)->adaptation), G_STRUCT_OFFSET(dt_iop_channelmixer_rgb_params_t, adaptation), NULL },
    /*entries*/ 6, /*values*/ NULL, /*Default*/ DT_ADAPTATION_CAT16
  };
  introspection_linear[22].Float = {
    { DT_INTROSPECTION_TYPE_FLOAT, (char*)"float", (char*)"x", (char*)"x", (char*)"", sizeof(((dt_iop_channelmixer_rgb_params_t*)NULL)->x), G_STRUCT_OFFSET(dt_iop_channelmixer_rgb_params_t, x), NULL },
    /*Min*/ -G_MAXFLOAT, /*Max*/ G_MAXFLOAT, /*Default*/ 0.333
  };
  introspection_linear[23].Float = {
    { DT_INTROSPECTION_TYPE_FLOAT, (char*)"float", (char*)"y", (char*)"y", (char*)"", sizeof(((dt_iop_channelmixer_rgb_params_t*)NULL)->y), G_STRUCT_OFFSET(dt_iop_channelmixer_rgb_params_t, y), NULL },
    /*Min*/ -G_MAXFLOAT, /*Max*/ G_MAXFLOAT, /*Default*/ 0.333
  };
  introspection_linear[24].Float = {
    { DT_INTROSPECTION_TYPE_FLOAT, (char*)"float", (char*)"temperature", (char*)"temperature", (char*)"", sizeof(((dt_iop_channelmixer_rgb_params_t*)NULL)->temperature), G_STRUCT_OFFSET(dt_iop_channelmixer_rgb_params_t, temperature), NULL },
    /*Min*/ TEMP_MIN, /*Max*/ TEMP_MAX, /*Default*/ 5003.
  };
  introspection_linear[25].Float = {
    { DT_INTROSPECTION_TYPE_FLOAT, (char*)"float", (char*)"gamut", (char*)"gamut", N_((char*)"gamut compression"), sizeof(((dt_iop_channelmixer_rgb_params_t*)NULL)->gamut), G_STRUCT_OFFSET(dt_iop_channelmixer_rgb_params_t, gamut), NULL },
    /*Min*/ 0.0, /*Max*/ 12.0, /*Default*/ 1.0
  };
  introspection_linear[26].Bool = {
    { DT_INTROSPECTION_TYPE_BOOL, (char*)"bool", (char*)"clip", (char*)"clip", N_((char*)"clip negative RGB from gamut"), sizeof(((dt_iop_channelmixer_rgb_params_t*)NULL)->clip), G_STRUCT_OFFSET(dt_iop_channelmixer_rgb_params_t, clip), NULL },
    /*Default*/ TRUE
  };
  introspection_linear[27].Enum = {
    { DT_INTROSPECTION_TYPE_ENUM, (char*)"dt_iop_channelmixer_rgb_version_t", (char*)"version", (char*)"version", N_((char*)"saturation algorithm"), sizeof(((dt_iop_channelmixer_rgb_params_t*)NULL)->version), G_STRUCT_OFFSET(dt_iop_channelmixer_rgb_params_t, version), NULL },
    /*entries*/ 3, /*values*/ NULL, /*Default*/ CHANNELMIXERRGB_V_3
  };
  introspection_linear[28].Struct = {
    { DT_INTROSPECTION_TYPE_STRUCT, (char*)"dt_iop_channelmixer_rgb_params_t", (char*)"", (char*)"", (char*)"", sizeof(dt_iop_channelmixer_rgb_params_t), 0, NULL },
    /*entries*/ 22, /*fields*/ NULL
  };
  introspection_linear[29].header = {DT_INTROSPECTION_TYPE_NONE, NULL, NULL, NULL, NULL, 0, 0, NULL};
#endif

  for(int i = 0; i <= 29; i++)
    introspection_linear[i].header.so = self;

  static dt_introspection_type_enum_tuple_t f18[] = { 
    { "DT_ILLUMINANT_PIPE", DT_ILLUMINANT_PIPE, N_((char*)"same as pipeline (D50)") },
    { "DT_ILLUMINANT_A", DT_ILLUMINANT_A, N_((char*)"A (incandescent)") },
    { "DT_ILLUMINANT_D", DT_ILLUMINANT_D, N_((char*)"D (daylight)") },
    { "DT_ILLUMINANT_E", DT_ILLUMINANT_E, N_((char*)"E (equi-energy)") },
    { "DT_ILLUMINANT_F", DT_ILLUMINANT_F, N_((char*)"F (fluorescent)") },
    { "DT_ILLUMINANT_LED", DT_ILLUMINANT_LED, N_((char*)"LED (LED light)") },
    { "DT_ILLUMINANT_BB", DT_ILLUMINANT_BB, N_((char*)"Planckian (black body)") },
    { "DT_ILLUMINANT_CUSTOM", DT_ILLUMINANT_CUSTOM, N_((char*)"custom") },
    { "DT_ILLUMINANT_CAMERA", DT_ILLUMINANT_CAMERA, N_((char*)"as shot in camera") },
    { "DT_ILLUMINANT_LAST", DT_ILLUMINANT_LAST, (char*)"" },
    { "DT_ILLUMINANT_DETECT_SURFACES", DT_ILLUMINANT_DETECT_SURFACES, (char*)"" },
    { "DT_ILLUMINANT_DETECT_EDGES", DT_ILLUMINANT_DETECT_EDGES, (char*)"" },
    { NULL, 0 },
  };
  static dt_introspection_type_enum_tuple_t f19[] = { 
    { "DT_ILLUMINANT_FLUO_F1", DT_ILLUMINANT_FLUO_F1, N_((char*)"F1 (Daylight 6430 K) – medium CRI") },
    { "DT_ILLUMINANT_FLUO_F2", DT_ILLUMINANT_FLUO_F2, N_((char*)"F2 (Cool White 4230 K) – medium CRI") },
    { "DT_ILLUMINANT_FLUO_F3", DT_ILLUMINANT_FLUO_F3, N_((char*)"F3 (White 3450 K) – medium CRI") },
    { "DT_ILLUMINANT_FLUO_F4", DT_ILLUMINANT_FLUO_F4, N_((char*)"F4 (Warm White 2940 K) – medium CRI") },
    { "DT_ILLUMINANT_FLUO_F5", DT_ILLUMINANT_FLUO_F5, N_((char*)"F5 (Daylight 6350 K) – medium CRI") },
    { "DT_ILLUMINANT_FLUO_F6", DT_ILLUMINANT_FLUO_F6, N_((char*)"F6 (Lite White 4150 K) – medium CRI") },
    { "DT_ILLUMINANT_FLUO_F7", DT_ILLUMINANT_FLUO_F7, N_((char*)"F7 (D65 simulator 6500 K) – high CRI") },
    { "DT_ILLUMINANT_FLUO_F8", DT_ILLUMINANT_FLUO_F8, N_((char*)"F8 (D50 simulator 5000 K) – high CRI") },
    { "DT_ILLUMINANT_FLUO_F9", DT_ILLUMINANT_FLUO_F9, N_((char*)"F9 (Cool White Deluxe 4150 K) – high CRI") },
    { "DT_ILLUMINANT_FLUO_F10", DT_ILLUMINANT_FLUO_F10, N_((char*)"F10 (Tuned RGB 5000 K) – low CRI") },
    { "DT_ILLUMINANT_FLUO_F11", DT_ILLUMINANT_FLUO_F11, N_((char*)"F11 (Tuned RGB 4000 K) – low CRI") },
    { "DT_ILLUMINANT_FLUO_F12", DT_ILLUMINANT_FLUO_F12, N_((char*)"F12 (Tuned RGB 3000 K) – low CRI") },
    { "DT_ILLUMINANT_FLUO_LAST", DT_ILLUMINANT_FLUO_LAST, (char*)"" },
    { NULL, 0 },
  };
  static dt_introspection_type_enum_tuple_t f20[] = { 
    { "DT_ILLUMINANT_LED_B1", DT_ILLUMINANT_LED_B1, N_((char*)"B1 (Blue 2733 K)") },
    { "DT_ILLUMINANT_LED_B2", DT_ILLUMINANT_LED_B2, N_((char*)"B2 (Blue 2998 K)") },
    { "DT_ILLUMINANT_LED_B3", DT_ILLUMINANT_LED_B3, N_((char*)"B3 (Blue 4103 K)") },
    { "DT_ILLUMINANT_LED_B4", DT_ILLUMINANT_LED_B4, N_((char*)"B4 (Blue 5109 K)") },
    { "DT_ILLUMINANT_LED_B5", DT_ILLUMINANT_LED_B5, N_((char*)"B5 (Blue 6598 K)") },
    { "DT_ILLUMINANT_LED_BH1", DT_ILLUMINANT_LED_BH1, N_((char*)"BH1 (Blue-Red hybrid 2851 K)") },
    { "DT_ILLUMINANT_LED_RGB1", DT_ILLUMINANT_LED_RGB1, N_((char*)"RGB1 (RGB 2840 K)") },
    { "DT_ILLUMINANT_LED_V1", DT_ILLUMINANT_LED_V1, N_((char*)"V1 (Violet 2724 K)") },
    { "DT_ILLUMINANT_LED_V2", DT_ILLUMINANT_LED_V2, N_((char*)"V2 (Violet 4070 K)") },
    { "DT_ILLUMINANT_LED_LAST", DT_ILLUMINANT_LED_LAST, (char*)"" },
    { NULL, 0 },
  };
  static dt_introspection_type_enum_tuple_t f21[] = { 
    { "DT_ADAPTATION_LINEAR_BRADFORD", DT_ADAPTATION_LINEAR_BRADFORD, N_((char*)"linear Bradford (ICC v4)") },
    { "DT_ADAPTATION_CAT16", DT_ADAPTATION_CAT16, N_((char*)"CAT16 (CIECAM16)") },
    { "DT_ADAPTATION_FULL_BRADFORD", DT_ADAPTATION_FULL_BRADFORD, N_((char*)"non-linear Bradford") },
    { "DT_ADAPTATION_XYZ", DT_ADAPTATION_XYZ, N_((char*)"XYZ") },
    { "DT_ADAPTATION_RGB", DT_ADAPTATION_RGB, N_((char*)"none (bypass)") },
    { "DT_ADAPTATION_LAST", DT_ADAPTATION_LAST, (char*)"" },
    { NULL, 0 },
  };
  static dt_introspection_type_enum_tuple_t f27[] = { 
    { "CHANNELMIXERRGB_V_1", CHANNELMIXERRGB_V_1, N_((char*)"version 1 (2020)") },
    { "CHANNELMIXERRGB_V_2", CHANNELMIXERRGB_V_2, N_((char*)"version 2 (2021)") },
    { "CHANNELMIXERRGB_V_3", CHANNELMIXERRGB_V_3, N_((char*)"version 3 (Apr 2021)") },
    { NULL, 0 },
  };
  static dt_introspection_field_t *f28[] = {
    &introspection_linear[1],
    &introspection_linear[3],
    &introspection_linear[5],
    &introspection_linear[7],
    &introspection_linear[9],
    &introspection_linear[11],
    &introspection_linear[12],
    &introspection_linear[13],
    &introspection_linear[14],
    &introspection_linear[15],
    &introspection_linear[16],
    &introspection_linear[17],
    &introspection_linear[18],
    &introspection_linear[19],
    &introspection_linear[20],
    &introspection_linear[21],
    &introspection_linear[22],
    &introspection_linear[23],
    &introspection_linear[24],
    &introspection_linear[25],
    &introspection_linear[26],
    &introspection_linear[27],
    NULL
  };

  introspection_linear[18].Enum.values = f18;
  introspection_linear[19].Enum.values = f19;
  introspection_linear[20].Enum.values = f20;
  introspection_linear[21].Enum.values = f21;
  introspection_linear[27].Enum.values = f27;
  introspection_linear[28].Struct.fields = f28;

  return 0;
}

void * get_p(const void * param, const char * name)
{
  dt_iop_channelmixer_rgb_params_t * p = (dt_iop_channelmixer_rgb_params_t*)param;
  if(!strcmp(name, "red[0]"))
    return &(p->red[0]);
  else if(!strcmp(name, "red"))
    return &(p->red);
  else if(!strcmp(name, "green[0]"))
    return &(p->green[0]);
  else if(!strcmp(name, "green"))
    return &(p->green);
  else if(!strcmp(name, "blue[0]"))
    return &(p->blue[0]);
  else if(!strcmp(name, "blue"))
    return &(p->blue);
  else if(!strcmp(name, "saturation[0]"))
    return &(p->saturation[0]);
  else if(!strcmp(name, "saturation"))
    return &(p->saturation);
  else if(!strcmp(name, "lightness[0]"))
    return &(p->lightness[0]);
  else if(!strcmp(name, "lightness"))
    return &(p->lightness);
  else if(!strcmp(name, "grey[0]"))
    return &(p->grey[0]);
  else if(!strcmp(name, "grey"))
    return &(p->grey);
  else if(!strcmp(name, "normalize_R"))
    return &(p->normalize_R);
  else if(!strcmp(name, "normalize_G"))
    return &(p->normalize_G);
  else if(!strcmp(name, "normalize_B"))
    return &(p->normalize_B);
  else if(!strcmp(name, "normalize_sat"))
    return &(p->normalize_sat);
  else if(!strcmp(name, "normalize_light"))
    return &(p->normalize_light);
  else if(!strcmp(name, "normalize_grey"))
    return &(p->normalize_grey);
  else if(!strcmp(name, "illuminant"))
    return &(p->illuminant);
  else if(!strcmp(name, "illum_fluo"))
    return &(p->illum_fluo);
  else if(!strcmp(name, "illum_led"))
    return &(p->illum_led);
  else if(!strcmp(name, "adaptation"))
    return &(p->adaptation);
  else if(!strcmp(name, "x"))
    return &(p->x);
  else if(!strcmp(name, "y"))
    return &(p->y);
  else if(!strcmp(name, "temperature"))
    return &(p->temperature);
  else if(!strcmp(name, "gamut"))
    return &(p->gamut);
  else if(!strcmp(name, "clip"))
    return &(p->clip);
  else if(!strcmp(name, "version"))
    return &(p->version);
  else
    return NULL;
}

dt_introspection_field_t * get_f(const char * name)
{
  if(!g_ascii_strcasecmp(name, "red[0]"))
    return &(introspection_linear[0]);
  else if(!g_ascii_strcasecmp(name, "red"))
    return &(introspection_linear[1]);
  else if(!g_ascii_strcasecmp(name, "green[0]"))
    return &(introspection_linear[2]);
  else if(!g_ascii_strcasecmp(name, "green"))
    return &(introspection_linear[3]);
  else if(!g_ascii_strcasecmp(name, "blue[0]"))
    return &(introspection_linear[4]);
  else if(!g_ascii_strcasecmp(name, "blue"))
    return &(introspection_linear[5]);
  else if(!g_ascii_strcasecmp(name, "saturation[0]"))
    return &(introspection_linear[6]);
  else if(!g_ascii_strcasecmp(name, "saturation"))
    return &(introspection_linear[7]);
  else if(!g_ascii_strcasecmp(name, "lightness[0]"))
    return &(introspection_linear[8]);
  else if(!g_ascii_strcasecmp(name, "lightness"))
    return &(introspection_linear[9]);
  else if(!g_ascii_strcasecmp(name, "grey[0]"))
    return &(introspection_linear[10]);
  else if(!g_ascii_strcasecmp(name, "grey"))
    return &(introspection_linear[11]);
  else if(!g_ascii_strcasecmp(name, "normalize_R"))
    return &(introspection_linear[12]);
  else if(!g_ascii_strcasecmp(name, "normalize_G"))
    return &(introspection_linear[13]);
  else if(!g_ascii_strcasecmp(name, "normalize_B"))
    return &(introspection_linear[14]);
  else if(!g_ascii_strcasecmp(name, "normalize_sat"))
    return &(introspection_linear[15]);
  else if(!g_ascii_strcasecmp(name, "normalize_light"))
    return &(introspection_linear[16]);
  else if(!g_ascii_strcasecmp(name, "normalize_grey"))
    return &(introspection_linear[17]);
  else if(!g_ascii_strcasecmp(name, "illuminant"))
    return &(introspection_linear[18]);
  else if(!g_ascii_strcasecmp(name, "illum_fluo"))
    return &(introspection_linear[19]);
  else if(!g_ascii_strcasecmp(name, "illum_led"))
    return &(introspection_linear[20]);
  else if(!g_ascii_strcasecmp(name, "adaptation"))
    return &(introspection_linear[21]);
  else if(!g_ascii_strcasecmp(name, "x"))
    return &(introspection_linear[22]);
  else if(!g_ascii_strcasecmp(name, "y"))
    return &(introspection_linear[23]);
  else if(!g_ascii_strcasecmp(name, "temperature"))
    return &(introspection_linear[24]);
  else if(!g_ascii_strcasecmp(name, "gamut"))
    return &(introspection_linear[25]);
  else if(!g_ascii_strcasecmp(name, "clip"))
    return &(introspection_linear[26]);
  else if(!g_ascii_strcasecmp(name, "version"))
    return &(introspection_linear[27]);
  else
    return NULL;
}

#ifdef __cplusplus
}
#endif

