/*
 * this code is auto generated. do not edit. change the sources in tools/introspection/ instead.
 *
 * the parse tree:
 *
 * typedef
 *   struct dt_iop_borders_params_t
 *     float
 *     color [3]
 *       description: border color
 *       default: 1.0
 *       range: -G_MAXFLOAT .. G_MAXFLOAT
 * 
 *     float
 *     aspect 
 *       description: aspect ratio
 *       default: DT_IOP_BORDERS_ASPECT_CONSTANT_VALUE
 *       range: 1.0 .. 3.0
 * 
 *     char
 *     aspect_text [20]
 *       default: 0
 *       range: CHAR_MIN .. CHAR_MAX
 * 
 *     enum dt_iop_orientation_t
 *       DT_IOP_BORDERS_ASPECT_ORIENTATION_AUTO : auto
 *       DT_IOP_BORDERS_ASPECT_ORIENTATION_PORTRAIT : portrait
 *       DT_IOP_BORDERS_ASPECT_ORIENTATION_LANDSCAPE : landscape
 *     aspect_orient 
 *       description: orientation
 *       default: 0
 * 
 *     float
 *     size 
 *       description: border size
 *       default: 0.1
 *       range: 0.0 .. 0.5
 * 
 *     float
 *     pos_h 
 *       description: horizontal offset
 *       default: 0.5
 *       range: 0.0 .. 1.0
 * 
 *     char
 *     pos_h_text [20]
 *       default: 0
 *       range: CHAR_MIN .. CHAR_MAX
 * 
 *     float
 *     pos_v 
 *       description: vertical offset
 *       default: 0.5
 *       range: 0.0 .. 1.0
 * 
 *     char
 *     pos_v_text [20]
 *       default: 0
 *       range: CHAR_MIN .. CHAR_MAX
 * 
 *     float
 *     frame_size 
 *       description: frame line size
 *       default: 0.0
 *       range: 0.0 .. 1.0
 * 
 *     float
 *     frame_offset 
 *       description: frame line offset
 *       default: 0.5
 *       range: 0.0 .. 1.0
 * 
 *     float
 *     frame_color [3]
 *       description: frame line color
 *       default: 0.0
 *       range: -G_MAXFLOAT .. G_MAXFLOAT
 * 
 *     boolean
 *     max_border_size 
 *       default: TRUE
 * 
 *     enum dt_iop_basis_t
 *       DT_IOP_BORDERS_BASIS_AUTO : auto
 *       DT_IOP_BORDERS_BASIS_WIDTH : width
 *       DT_IOP_BORDERS_BASIS_HEIGHT : height
 *       DT_IOP_BORDERS_BASIS_SHORTER : shorter
 *       DT_IOP_BORDERS_BASIS_LONGER : longer
 *     basis 
 *       description: basis
 *       default: 0
 * 
 * dt_iop_borders_params_t
 *
 */

#include "/build/darktable/src/darktable-5.6.0/src/iop/borders.c"

#ifdef __cplusplus
extern "C"
{
#endif

#include <string.h>
#include "common/introspection.h"

static dt_introspection_field_t introspection_linear[20+1]
#if !( defined(__cplusplus) && !defined(__clang__) && defined(__GNUC__) && ((__GNUC__*100 + __GNUC_MINOR__) <= 406) )
= {
{
  .Float = {
    { DT_INTROSPECTION_TYPE_FLOAT, (char*)"float", (char*)"color[0]", (char*)"color[0]", N_((char*)"border color"), sizeof(((dt_iop_borders_params_t*)NULL)->color[0]), G_STRUCT_OFFSET(dt_iop_borders_params_t, color), NULL },
    /*Min*/ -G_MAXFLOAT, /*Max*/ G_MAXFLOAT, /*Default*/ 1.0
  }
},
{
  .Array = {
    { DT_INTROSPECTION_TYPE_ARRAY, (char*)"float[]", (char*)"color", (char*)"color", N_((char*)"border color"), sizeof(((dt_iop_borders_params_t*)NULL)->color), G_STRUCT_OFFSET(dt_iop_borders_params_t, color), NULL },
    /*count*/ G_N_ELEMENTS(((dt_iop_borders_params_t*)NULL)->color), /*type*/ DT_INTROSPECTION_TYPE_FLOAT, /*field*/ &introspection_linear[0]
  }
},
{
  .Float = {
    { DT_INTROSPECTION_TYPE_FLOAT, (char*)"float", (char*)"aspect", (char*)"aspect", N_((char*)"aspect ratio"), sizeof(((dt_iop_borders_params_t*)NULL)->aspect), G_STRUCT_OFFSET(dt_iop_borders_params_t, aspect), NULL },
    /*Min*/ 1.0, /*Max*/ 3.0, /*Default*/ DT_IOP_BORDERS_ASPECT_CONSTANT_VALUE
  }
},
{
  .Char = {
    { DT_INTROSPECTION_TYPE_CHAR, (char*)"char", (char*)"aspect_text[0]", (char*)"aspect_text[0]", (char*)"", sizeof(((dt_iop_borders_params_t*)NULL)->aspect_text[0]), G_STRUCT_OFFSET(dt_iop_borders_params_t, aspect_text), NULL },
    /*Min*/ CHAR_MIN, /*Max*/ CHAR_MAX, /*Default*/ 0
  }
},
{
  .Array = {
    { DT_INTROSPECTION_TYPE_ARRAY, (char*)"char[]", (char*)"aspect_text", (char*)"aspect_text", (char*)"", sizeof(((dt_iop_borders_params_t*)NULL)->aspect_text), G_STRUCT_OFFSET(dt_iop_borders_params_t, aspect_text), NULL },
    /*count*/ G_N_ELEMENTS(((dt_iop_borders_params_t*)NULL)->aspect_text), /*type*/ DT_INTROSPECTION_TYPE_CHAR, /*field*/ &introspection_linear[3]
  }
},
{
  .Enum = {
    { DT_INTROSPECTION_TYPE_ENUM, (char*)"dt_iop_orientation_t", (char*)"aspect_orient", (char*)"aspect_orient", N_((char*)"orientation"), sizeof(((dt_iop_borders_params_t*)NULL)->aspect_orient), G_STRUCT_OFFSET(dt_iop_borders_params_t, aspect_orient), NULL },
    /*entries*/ 3, /*values*/ NULL, /*Default*/ 0
  }
},
{
  .Float = {
    { DT_INTROSPECTION_TYPE_FLOAT, (char*)"float", (char*)"size", (char*)"size", N_((char*)"border size"), sizeof(((dt_iop_borders_params_t*)NULL)->size), G_STRUCT_OFFSET(dt_iop_borders_params_t, size), NULL },
    /*Min*/ 0.0, /*Max*/ 0.5, /*Default*/ 0.1
  }
},
{
  .Float = {
    { DT_INTROSPECTION_TYPE_FLOAT, (char*)"float", (char*)"pos_h", (char*)"pos_h", N_((char*)"horizontal offset"), sizeof(((dt_iop_borders_params_t*)NULL)->pos_h), G_STRUCT_OFFSET(dt_iop_borders_params_t, pos_h), NULL },
    /*Min*/ 0.0, /*Max*/ 1.0, /*Default*/ 0.5
  }
},
{
  .Char = {
    { DT_INTROSPECTION_TYPE_CHAR, (char*)"char", (char*)"pos_h_text[0]", (char*)"pos_h_text[0]", (char*)"", sizeof(((dt_iop_borders_params_t*)NULL)->pos_h_text[0]), G_STRUCT_OFFSET(dt_iop_borders_params_t, pos_h_text), NULL },
    /*Min*/ CHAR_MIN, /*Max*/ CHAR_MAX, /*Default*/ 0
  }
},
{
  .Array = {
    { DT_INTROSPECTION_TYPE_ARRAY, (char*)"char[]", (char*)"pos_h_text", (char*)"pos_h_text", (char*)"", sizeof(((dt_iop_borders_params_t*)NULL)->pos_h_text), G_STRUCT_OFFSET(dt_iop_borders_params_t, pos_h_text), NULL },
    /*count*/ G_N_ELEMENTS(((dt_iop_borders_params_t*)NULL)->pos_h_text), /*type*/ DT_INTROSPECTION_TYPE_CHAR, /*field*/ &introspection_linear[8]
  }
},
{
  .Float = {
    { DT_INTROSPECTION_TYPE_FLOAT, (char*)"float", (char*)"pos_v", (char*)"pos_v", N_((char*)"vertical offset"), sizeof(((dt_iop_borders_params_t*)NULL)->pos_v), G_STRUCT_OFFSET(dt_iop_borders_params_t, pos_v), NULL },
    /*Min*/ 0.0, /*Max*/ 1.0, /*Default*/ 0.5
  }
},
{
  .Char = {
    { DT_INTROSPECTION_TYPE_CHAR, (char*)"char", (char*)"pos_v_text[0]", (char*)"pos_v_text[0]", (char*)"", sizeof(((dt_iop_borders_params_t*)NULL)->pos_v_text[0]), G_STRUCT_OFFSET(dt_iop_borders_params_t, pos_v_text), NULL },
    /*Min*/ CHAR_MIN, /*Max*/ CHAR_MAX, /*Default*/ 0
  }
},
{
  .Array = {
    { DT_INTROSPECTION_TYPE_ARRAY, (char*)"char[]", (char*)"pos_v_text", (char*)"pos_v_text", (char*)"", sizeof(((dt_iop_borders_params_t*)NULL)->pos_v_text), G_STRUCT_OFFSET(dt_iop_borders_params_t, pos_v_text), NULL },
    /*count*/ G_N_ELEMENTS(((dt_iop_borders_params_t*)NULL)->pos_v_text), /*type*/ DT_INTROSPECTION_TYPE_CHAR, /*field*/ &introspection_linear[11]
  }
},
{
  .Float = {
    { DT_INTROSPECTION_TYPE_FLOAT, (char*)"float", (char*)"frame_size", (char*)"frame_size", N_((char*)"frame line size"), sizeof(((dt_iop_borders_params_t*)NULL)->frame_size), G_STRUCT_OFFSET(dt_iop_borders_params_t, frame_size), NULL },
    /*Min*/ 0.0, /*Max*/ 1.0, /*Default*/ 0.0
  }
},
{
  .Float = {
    { DT_INTROSPECTION_TYPE_FLOAT, (char*)"float", (char*)"frame_offset", (char*)"frame_offset", N_((char*)"frame line offset"), sizeof(((dt_iop_borders_params_t*)NULL)->frame_offset), G_STRUCT_OFFSET(dt_iop_borders_params_t, frame_offset), NULL },
    /*Min*/ 0.0, /*Max*/ 1.0, /*Default*/ 0.5
  }
},
{
  .Float = {
    { DT_INTROSPECTION_TYPE_FLOAT, (char*)"float", (char*)"frame_color[0]", (char*)"frame_color[0]", N_((char*)"frame line color"), sizeof(((dt_iop_borders_params_t*)NULL)->frame_color[0]), G_STRUCT_OFFSET(dt_iop_borders_params_t, frame_color), NULL },
    /*Min*/ -G_MAXFLOAT, /*Max*/ G_MAXFLOAT, /*Default*/ 0.0
  }
},
{
  .Array = {
    { DT_INTROSPECTION_TYPE_ARRAY, (char*)"float[]", (char*)"frame_color", (char*)"frame_color", N_((char*)"frame line color"), sizeof(((dt_iop_borders_params_t*)NULL)->frame_color), G_STRUCT_OFFSET(dt_iop_borders_params_t, frame_color), NULL },
    /*count*/ G_N_ELEMENTS(((dt_iop_borders_params_t*)NULL)->frame_color), /*type*/ DT_INTROSPECTION_TYPE_FLOAT, /*field*/ &introspection_linear[15]
  }
},
{
  .Bool = {
    { DT_INTROSPECTION_TYPE_BOOL, (char*)"bool", (char*)"max_border_size", (char*)"max_border_size", (char*)"", sizeof(((dt_iop_borders_params_t*)NULL)->max_border_size), G_STRUCT_OFFSET(dt_iop_borders_params_t, max_border_size), NULL },
    /*Default*/ TRUE
  }
},
{
  .Enum = {
    { DT_INTROSPECTION_TYPE_ENUM, (char*)"dt_iop_basis_t", (char*)"basis", (char*)"basis", N_((char*)"basis"), sizeof(((dt_iop_borders_params_t*)NULL)->basis), G_STRUCT_OFFSET(dt_iop_borders_params_t, basis), NULL },
    /*entries*/ 5, /*values*/ NULL, /*Default*/ 0
  }
},
{
  .Struct = {
    { DT_INTROSPECTION_TYPE_STRUCT, (char*)"dt_iop_borders_params_t", (char*)"", (char*)"", (char*)"", sizeof(dt_iop_borders_params_t), 0, NULL },
    /*entries*/ 14, /*fields*/ NULL
  }
},
  { .header = {DT_INTROSPECTION_TYPE_NONE, NULL, NULL, NULL, NULL, 0, 0, NULL} }
}
#endif
;

static dt_introspection_t introspection = {
  8,
  4,
  "dt_iop_borders_params_t",
  sizeof(dt_iop_borders_params_t),
  &introspection_linear[20-1],
  sizeof(dt_iop_module_t),
  G_STRUCT_OFFSET(dt_iop_module_t, default_params)
};

dt_introspection_field_t* get_introspection_linear()
{
  return introspection_linear;
}

dt_introspection_t* get_introspection()
{
  return &introspection;
}

int introspection_init(struct dt_iop_module_so_t *self, int api_version)
{
  // here we check that the generated code matches the api at compile time and also at runtime
  if(introspection.api_version != DT_INTROSPECTION_VERSION || api_version != DT_INTROSPECTION_VERSION)
    return 1;

#if defined(__cplusplus) && !defined(__clang__) && defined(__GNUC__) && ((__GNUC__*100 + __GNUC_MINOR__) <= 406)
  introspection_linear[0].Float = {
    { DT_INTROSPECTION_TYPE_FLOAT, (char*)"float", (char*)"color[0]", (char*)"color[0]", N_((char*)"border color"), sizeof(((dt_iop_borders_params_t*)NULL)->color[0]), G_STRUCT_OFFSET(dt_iop_borders_params_t, color), NULL },
    /*Min*/ -G_MAXFLOAT, /*Max*/ G_MAXFLOAT, /*Default*/ 1.0
  };
  introspection_linear[1].Array = {
    { DT_INTROSPECTION_TYPE_ARRAY, (char*)"float[]", (char*)"color", (char*)"color", N_((char*)"border color"), sizeof(((dt_iop_borders_params_t*)NULL)->color), G_STRUCT_OFFSET(dt_iop_borders_params_t, color), NULL },
    /*count*/ G_N_ELEMENTS(((dt_iop_borders_params_t*)NULL)->color), /*type*/ DT_INTROSPECTION_TYPE_FLOAT, /*field*/ &introspection_linear[0]
  };
  introspection_linear[2].Float = {
    { DT_INTROSPECTION_TYPE_FLOAT, (char*)"float", (char*)"aspect", (char*)"aspect", N_((char*)"aspect ratio"), sizeof(((dt_iop_borders_params_t*)NULL)->aspect), G_STRUCT_OFFSET(dt_iop_borders_params_t, aspect), NULL },
    /*Min*/ 1.0, /*Max*/ 3.0, /*Default*/ DT_IOP_BORDERS_ASPECT_CONSTANT_VALUE
  };
  introspection_linear[3].Char = {
    { DT_INTROSPECTION_TYPE_CHAR, (char*)"char", (char*)"aspect_text[0]", (char*)"aspect_text[0]", (char*)"", sizeof(((dt_iop_borders_params_t*)NULL)->aspect_text[0]), G_STRUCT_OFFSET(dt_iop_borders_params_t, aspect_text), NULL },
    /*Min*/ CHAR_MIN, /*Max*/ CHAR_MAX, /*Default*/ 0
  };
  introspection_linear[4].Array = {
    { DT_INTROSPECTION_TYPE_ARRAY, (char*)"char[]", (char*)"aspect_text", (char*)"aspect_text", (char*)"", sizeof(((dt_iop_borders_params_t*)NULL)->aspect_text), G_STRUCT_OFFSET(dt_iop_borders_params_t, aspect_text), NULL },
    /*count*/ G_N_ELEMENTS(((dt_iop_borders_params_t*)NULL)->aspect_text), /*type*/ DT_INTROSPECTION_TYPE_CHAR, /*field*/ &introspection_linear[3]
  };
  introspection_linear[5].Enum = {
    { DT_INTROSPECTION_TYPE_ENUM, (char*)"dt_iop_orientation_t", (char*)"aspect_orient", (char*)"aspect_orient", N_((char*)"orientation"), sizeof(((dt_iop_borders_params_t*)NULL)->aspect_orient), G_STRUCT_OFFSET(dt_iop_borders_params_t, aspect_orient), NULL },
    /*entries*/ 3, /*values*/ NULL, /*Default*/ 0
  };
  introspection_linear[6].Float = {
    { DT_INTROSPECTION_TYPE_FLOAT, (char*)"float", (char*)"size", (char*)"size", N_((char*)"border size"), sizeof(((dt_iop_borders_params_t*)NULL)->size), G_STRUCT_OFFSET(dt_iop_borders_params_t, size), NULL },
    /*Min*/ 0.0, /*Max*/ 0.5, /*Default*/ 0.1
  };
  introspection_linear[7].Float = {
    { DT_INTROSPECTION_TYPE_FLOAT, (char*)"float", (char*)"pos_h", (char*)"pos_h", N_((char*)"horizontal offset"), sizeof(((dt_iop_borders_params_t*)NULL)->pos_h), G_STRUCT_OFFSET(dt_iop_borders_params_t, pos_h), NULL },
    /*Min*/ 0.0, /*Max*/ 1.0, /*Default*/ 0.5
  };
  introspection_linear[8].Char = {
    { DT_INTROSPECTION_TYPE_CHAR, (char*)"char", (char*)"pos_h_text[0]", (char*)"pos_h_text[0]", (char*)"", sizeof(((dt_iop_borders_params_t*)NULL)->pos_h_text[0]), G_STRUCT_OFFSET(dt_iop_borders_params_t, pos_h_text), NULL },
    /*Min*/ CHAR_MIN, /*Max*/ CHAR_MAX, /*Default*/ 0
  };
  introspection_linear[9].Array = {
    { DT_INTROSPECTION_TYPE_ARRAY, (char*)"char[]", (char*)"pos_h_text", (char*)"pos_h_text", (char*)"", sizeof(((dt_iop_borders_params_t*)NULL)->pos_h_text), G_STRUCT_OFFSET(dt_iop_borders_params_t, pos_h_text), NULL },
    /*count*/ G_N_ELEMENTS(((dt_iop_borders_params_t*)NULL)->pos_h_text), /*type*/ DT_INTROSPECTION_TYPE_CHAR, /*field*/ &introspection_linear[8]
  };
  introspection_linear[10].Float = {
    { DT_INTROSPECTION_TYPE_FLOAT, (char*)"float", (char*)"pos_v", (char*)"pos_v", N_((char*)"vertical offset"), sizeof(((dt_iop_borders_params_t*)NULL)->pos_v), G_STRUCT_OFFSET(dt_iop_borders_params_t, pos_v), NULL },
    /*Min*/ 0.0, /*Max*/ 1.0, /*Default*/ 0.5
  };
  introspection_linear[11].Char = {
    { DT_INTROSPECTION_TYPE_CHAR, (char*)"char", (char*)"pos_v_text[0]", (char*)"pos_v_text[0]", (char*)"", sizeof(((dt_iop_borders_params_t*)NULL)->pos_v_text[0]), G_STRUCT_OFFSET(dt_iop_borders_params_t, pos_v_text), NULL },
    /*Min*/ CHAR_MIN, /*Max*/ CHAR_MAX, /*Default*/ 0
  };
  introspection_linear[12].Array = {
    { DT_INTROSPECTION_TYPE_ARRAY, (char*)"char[]", (char*)"pos_v_text", (char*)"pos_v_text", (char*)"", sizeof(((dt_iop_borders_params_t*)NULL)->pos_v_text), G_STRUCT_OFFSET(dt_iop_borders_params_t, pos_v_text), NULL },
    /*count*/ G_N_ELEMENTS(((dt_iop_borders_params_t*)NULL)->pos_v_text), /*type*/ DT_INTROSPECTION_TYPE_CHAR, /*field*/ &introspection_linear[11]
  };
  introspection_linear[13].Float = {
    { DT_INTROSPECTION_TYPE_FLOAT, (char*)"float", (char*)"frame_size", (char*)"frame_size", N_((char*)"frame line size"), sizeof(((dt_iop_borders_params_t*)NULL)->frame_size), G_STRUCT_OFFSET(dt_iop_borders_params_t, frame_size), NULL },
    /*Min*/ 0.0, /*Max*/ 1.0, /*Default*/ 0.0
  };
  introspection_linear[14].Float = {
    { DT_INTROSPECTION_TYPE_FLOAT, (char*)"float", (char*)"frame_offset", (char*)"frame_offset", N_((char*)"frame line offset"), sizeof(((dt_iop_borders_params_t*)NULL)->frame_offset), G_STRUCT_OFFSET(dt_iop_borders_params_t, frame_offset), NULL },
    /*Min*/ 0.0, /*Max*/ 1.0, /*Default*/ 0.5
  };
  introspection_linear[15].Float = {
    { DT_INTROSPECTION_TYPE_FLOAT, (char*)"float", (char*)"frame_color[0]", (char*)"frame_color[0]", N_((char*)"frame line color"), sizeof(((dt_iop_borders_params_t*)NULL)->frame_color[0]), G_STRUCT_OFFSET(dt_iop_borders_params_t, frame_color), NULL },
    /*Min*/ -G_MAXFLOAT, /*Max*/ G_MAXFLOAT, /*Default*/ 0.0
  };
  introspection_linear[16].Array = {
    { DT_INTROSPECTION_TYPE_ARRAY, (char*)"float[]", (char*)"frame_color", (char*)"frame_color", N_((char*)"frame line color"), sizeof(((dt_iop_borders_params_t*)NULL)->frame_color), G_STRUCT_OFFSET(dt_iop_borders_params_t, frame_color), NULL },
    /*count*/ G_N_ELEMENTS(((dt_iop_borders_params_t*)NULL)->frame_color), /*type*/ DT_INTROSPECTION_TYPE_FLOAT, /*field*/ &introspection_linear[15]
  };
  introspection_linear[17].Bool = {
    { DT_INTROSPECTION_TYPE_BOOL, (char*)"bool", (char*)"max_border_size", (char*)"max_border_size", (char*)"", sizeof(((dt_iop_borders_params_t*)NULL)->max_border_size), G_STRUCT_OFFSET(dt_iop_borders_params_t, max_border_size), NULL },
    /*Default*/ TRUE
  };
  introspection_linear[18].Enum = {
    { DT_INTROSPECTION_TYPE_ENUM, (char*)"dt_iop_basis_t", (char*)"basis", (char*)"basis", N_((char*)"basis"), sizeof(((dt_iop_borders_params_t*)NULL)->basis), G_STRUCT_OFFSET(dt_iop_borders_params_t, basis), NULL },
    /*entries*/ 5, /*values*/ NULL, /*Default*/ 0
  };
  introspection_linear[19].Struct = {
    { DT_INTROSPECTION_TYPE_STRUCT, (char*)"dt_iop_borders_params_t", (char*)"", (char*)"", (char*)"", sizeof(dt_iop_borders_params_t), 0, NULL },
    /*entries*/ 14, /*fields*/ NULL
  };
  introspection_linear[20].header = {DT_INTROSPECTION_TYPE_NONE, NULL, NULL, NULL, NULL, 0, 0, NULL};
#endif

  for(int i = 0; i <= 20; i++)
    introspection_linear[i].header.so = self;

  static dt_introspection_type_enum_tuple_t f5[] = { 
    { "DT_IOP_BORDERS_ASPECT_ORIENTATION_AUTO", DT_IOP_BORDERS_ASPECT_ORIENTATION_AUTO, N_((char*)"auto") },
    { "DT_IOP_BORDERS_ASPECT_ORIENTATION_PORTRAIT", DT_IOP_BORDERS_ASPECT_ORIENTATION_PORTRAIT, N_((char*)"portrait") },
    { "DT_IOP_BORDERS_ASPECT_ORIENTATION_LANDSCAPE", DT_IOP_BORDERS_ASPECT_ORIENTATION_LANDSCAPE, N_((char*)"landscape") },
    { NULL, 0 },
  };
  static dt_introspection_type_enum_tuple_t f18[] = { 
    { "DT_IOP_BORDERS_BASIS_AUTO", DT_IOP_BORDERS_BASIS_AUTO, N_((char*)"auto") },
    { "DT_IOP_BORDERS_BASIS_WIDTH", DT_IOP_BORDERS_BASIS_WIDTH, N_((char*)"width") },
    { "DT_IOP_BORDERS_BASIS_HEIGHT", DT_IOP_BORDERS_BASIS_HEIGHT, N_((char*)"height") },
    { "DT_IOP_BORDERS_BASIS_SHORTER", DT_IOP_BORDERS_BASIS_SHORTER, N_((char*)"shorter") },
    { "DT_IOP_BORDERS_BASIS_LONGER", DT_IOP_BORDERS_BASIS_LONGER, N_((char*)"longer") },
    { NULL, 0 },
  };
  static dt_introspection_field_t *f19[] = {
    &introspection_linear[1],
    &introspection_linear[2],
    &introspection_linear[4],
    &introspection_linear[5],
    &introspection_linear[6],
    &introspection_linear[7],
    &introspection_linear[9],
    &introspection_linear[10],
    &introspection_linear[12],
    &introspection_linear[13],
    &introspection_linear[14],
    &introspection_linear[16],
    &introspection_linear[17],
    &introspection_linear[18],
    NULL
  };

  introspection_linear[5].Enum.values = f5;
  introspection_linear[18].Enum.values = f18;
  introspection_linear[19].Struct.fields = f19;

  return 0;
}

void * get_p(const void * param, const char * name)
{
  dt_iop_borders_params_t * p = (dt_iop_borders_params_t*)param;
  if(!strcmp(name, "color[0]"))
    return &(p->color[0]);
  else if(!strcmp(name, "color"))
    return &(p->color);
  else if(!strcmp(name, "aspect"))
    return &(p->aspect);
  else if(!strcmp(name, "aspect_text[0]"))
    return &(p->aspect_text[0]);
  else if(!strcmp(name, "aspect_text"))
    return &(p->aspect_text);
  else if(!strcmp(name, "aspect_orient"))
    return &(p->aspect_orient);
  else if(!strcmp(name, "size"))
    return &(p->size);
  else if(!strcmp(name, "pos_h"))
    return &(p->pos_h);
  else if(!strcmp(name, "pos_h_text[0]"))
    return &(p->pos_h_text[0]);
  else if(!strcmp(name, "pos_h_text"))
    return &(p->pos_h_text);
  else if(!strcmp(name, "pos_v"))
    return &(p->pos_v);
  else if(!strcmp(name, "pos_v_text[0]"))
    return &(p->pos_v_text[0]);
  else if(!strcmp(name, "pos_v_text"))
    return &(p->pos_v_text);
  else if(!strcmp(name, "frame_size"))
    return &(p->frame_size);
  else if(!strcmp(name, "frame_offset"))
    return &(p->frame_offset);
  else if(!strcmp(name, "frame_color[0]"))
    return &(p->frame_color[0]);
  else if(!strcmp(name, "frame_color"))
    return &(p->frame_color);
  else if(!strcmp(name, "max_border_size"))
    return &(p->max_border_size);
  else if(!strcmp(name, "basis"))
    return &(p->basis);
  else
    return NULL;
}

dt_introspection_field_t * get_f(const char * name)
{
  if(!g_ascii_strcasecmp(name, "color[0]"))
    return &(introspection_linear[0]);
  else if(!g_ascii_strcasecmp(name, "color"))
    return &(introspection_linear[1]);
  else if(!g_ascii_strcasecmp(name, "aspect"))
    return &(introspection_linear[2]);
  else if(!g_ascii_strcasecmp(name, "aspect_text[0]"))
    return &(introspection_linear[3]);
  else if(!g_ascii_strcasecmp(name, "aspect_text"))
    return &(introspection_linear[4]);
  else if(!g_ascii_strcasecmp(name, "aspect_orient"))
    return &(introspection_linear[5]);
  else if(!g_ascii_strcasecmp(name, "size"))
    return &(introspection_linear[6]);
  else if(!g_ascii_strcasecmp(name, "pos_h"))
    return &(introspection_linear[7]);
  else if(!g_ascii_strcasecmp(name, "pos_h_text[0]"))
    return &(introspection_linear[8]);
  else if(!g_ascii_strcasecmp(name, "pos_h_text"))
    return &(introspection_linear[9]);
  else if(!g_ascii_strcasecmp(name, "pos_v"))
    return &(introspection_linear[10]);
  else if(!g_ascii_strcasecmp(name, "pos_v_text[0]"))
    return &(introspection_linear[11]);
  else if(!g_ascii_strcasecmp(name, "pos_v_text"))
    return &(introspection_linear[12]);
  else if(!g_ascii_strcasecmp(name, "frame_size"))
    return &(introspection_linear[13]);
  else if(!g_ascii_strcasecmp(name, "frame_offset"))
    return &(introspection_linear[14]);
  else if(!g_ascii_strcasecmp(name, "frame_color[0]"))
    return &(introspection_linear[15]);
  else if(!g_ascii_strcasecmp(name, "frame_color"))
    return &(introspection_linear[16]);
  else if(!g_ascii_strcasecmp(name, "max_border_size"))
    return &(introspection_linear[17]);
  else if(!g_ascii_strcasecmp(name, "basis"))
    return &(introspection_linear[18]);
  else
    return NULL;
}

#ifdef __cplusplus
}
#endif

