/*
 * this code is auto generated. do not edit. change the sources in tools/introspection/ instead.
 *
 * the parse tree:
 *
 * typedef
 *   struct dt_iop_colorcorrection_params_t
 *     float
 *     hia 
 *       default: 0.0
 *       range: -G_MAXFLOAT .. G_MAXFLOAT
 * 
 *     float
 *     hib 
 *       default: 0.0
 *       range: -G_MAXFLOAT .. G_MAXFLOAT
 * 
 *     float
 *     loa 
 *       default: 0.0
 *       range: -G_MAXFLOAT .. G_MAXFLOAT
 * 
 *     float
 *     lob 
 *       default: 0.0
 *       range: -G_MAXFLOAT .. G_MAXFLOAT
 * 
 *     float
 *     saturation 
 *       default: 1.0
 *       range: -3.0 .. 3.0
 * 
 * dt_iop_colorcorrection_params_t
 *
 */

#include "/build/darktable/src/darktable-5.6.0/src/iop/colorcorrection.c"

#ifdef __cplusplus
extern "C"
{
#endif

#include <string.h>
#include "common/introspection.h"

static dt_introspection_field_t introspection_linear[6+1]
#if !( defined(__cplusplus) && !defined(__clang__) && defined(__GNUC__) && ((__GNUC__*100 + __GNUC_MINOR__) <= 406) )
= {
{
  .Float = {
    { DT_INTROSPECTION_TYPE_FLOAT, (char*)"float", (char*)"hia", (char*)"hia", (char*)"", sizeof(((dt_iop_colorcorrection_params_t*)NULL)->hia), G_STRUCT_OFFSET(dt_iop_colorcorrection_params_t, hia), NULL },
    /*Min*/ -G_MAXFLOAT, /*Max*/ G_MAXFLOAT, /*Default*/ 0.0
  }
},
{
  .Float = {
    { DT_INTROSPECTION_TYPE_FLOAT, (char*)"float", (char*)"hib", (char*)"hib", (char*)"", sizeof(((dt_iop_colorcorrection_params_t*)NULL)->hib), G_STRUCT_OFFSET(dt_iop_colorcorrection_params_t, hib), NULL },
    /*Min*/ -G_MAXFLOAT, /*Max*/ G_MAXFLOAT, /*Default*/ 0.0
  }
},
{
  .Float = {
    { DT_INTROSPECTION_TYPE_FLOAT, (char*)"float", (char*)"loa", (char*)"loa", (char*)"", sizeof(((dt_iop_colorcorrection_params_t*)NULL)->loa), G_STRUCT_OFFSET(dt_iop_colorcorrection_params_t, loa), NULL },
    /*Min*/ -G_MAXFLOAT, /*Max*/ G_MAXFLOAT, /*Default*/ 0.0
  }
},
{
  .Float = {
    { DT_INTROSPECTION_TYPE_FLOAT, (char*)"float", (char*)"lob", (char*)"lob", (char*)"", sizeof(((dt_iop_colorcorrection_params_t*)NULL)->lob), G_STRUCT_OFFSET(dt_iop_colorcorrection_params_t, lob), NULL },
    /*Min*/ -G_MAXFLOAT, /*Max*/ G_MAXFLOAT, /*Default*/ 0.0
  }
},
{
  .Float = {
    { DT_INTROSPECTION_TYPE_FLOAT, (char*)"float", (char*)"saturation", (char*)"saturation", (char*)"", sizeof(((dt_iop_colorcorrection_params_t*)NULL)->saturation), G_STRUCT_OFFSET(dt_iop_colorcorrection_params_t, saturation), NULL },
    /*Min*/ -3.0, /*Max*/ 3.0, /*Default*/ 1.0
  }
},
{
  .Struct = {
    { DT_INTROSPECTION_TYPE_STRUCT, (char*)"dt_iop_colorcorrection_params_t", (char*)"", (char*)"", (char*)"", sizeof(dt_iop_colorcorrection_params_t), 0, NULL },
    /*entries*/ 5, /*fields*/ NULL
  }
},
  { .header = {DT_INTROSPECTION_TYPE_NONE, NULL, NULL, NULL, NULL, 0, 0, NULL} }
}
#endif
;

static dt_introspection_t introspection = {
  8,
  1,
  "dt_iop_colorcorrection_params_t",
  sizeof(dt_iop_colorcorrection_params_t),
  &introspection_linear[6-1],
  sizeof(dt_iop_module_t),
  G_STRUCT_OFFSET(dt_iop_module_t, default_params)
};

dt_introspection_field_t* get_introspection_linear()
{
  return introspection_linear;
}

dt_introspection_t* get_introspection()
{
  return &introspection;
}

int introspection_init(struct dt_iop_module_so_t *self, int api_version)
{
  // here we check that the generated code matches the api at compile time and also at runtime
  if(introspection.api_version != DT_INTROSPECTION_VERSION || api_version != DT_INTROSPECTION_VERSION)
    return 1;

#if defined(__cplusplus) && !defined(__clang__) && defined(__GNUC__) && ((__GNUC__*100 + __GNUC_MINOR__) <= 406)
  introspection_linear[0].Float = {
    { DT_INTROSPECTION_TYPE_FLOAT, (char*)"float", (char*)"hia", (char*)"hia", (char*)"", sizeof(((dt_iop_colorcorrection_params_t*)NULL)->hia), G_STRUCT_OFFSET(dt_iop_colorcorrection_params_t, hia), NULL },
    /*Min*/ -G_MAXFLOAT, /*Max*/ G_MAXFLOAT, /*Default*/ 0.0
  };
  introspection_linear[1].Float = {
    { DT_INTROSPECTION_TYPE_FLOAT, (char*)"float", (char*)"hib", (char*)"hib", (char*)"", sizeof(((dt_iop_colorcorrection_params_t*)NULL)->hib), G_STRUCT_OFFSET(dt_iop_colorcorrection_params_t, hib), NULL },
    /*Min*/ -G_MAXFLOAT, /*Max*/ G_MAXFLOAT, /*Default*/ 0.0
  };
  introspection_linear[2].Float = {
    { DT_INTROSPECTION_TYPE_FLOAT, (char*)"float", (char*)"loa", (char*)"loa", (char*)"", sizeof(((dt_iop_colorcorrection_params_t*)NULL)->loa), G_STRUCT_OFFSET(dt_iop_colorcorrection_params_t, loa), NULL },
    /*Min*/ -G_MAXFLOAT, /*Max*/ G_MAXFLOAT, /*Default*/ 0.0
  };
  introspection_linear[3].Float = {
    { DT_INTROSPECTION_TYPE_FLOAT, (char*)"float", (char*)"lob", (char*)"lob", (char*)"", sizeof(((dt_iop_colorcorrection_params_t*)NULL)->lob), G_STRUCT_OFFSET(dt_iop_colorcorrection_params_t, lob), NULL },
    /*Min*/ -G_MAXFLOAT, /*Max*/ G_MAXFLOAT, /*Default*/ 0.0
  };
  introspection_linear[4].Float = {
    { DT_INTROSPECTION_TYPE_FLOAT, (char*)"float", (char*)"saturation", (char*)"saturation", (char*)"", sizeof(((dt_iop_colorcorrection_params_t*)NULL)->saturation), G_STRUCT_OFFSET(dt_iop_colorcorrection_params_t, saturation), NULL },
    /*Min*/ -3.0, /*Max*/ 3.0, /*Default*/ 1.0
  };
  introspection_linear[5].Struct = {
    { DT_INTROSPECTION_TYPE_STRUCT, (char*)"dt_iop_colorcorrection_params_t", (char*)"", (char*)"", (char*)"", sizeof(dt_iop_colorcorrection_params_t), 0, NULL },
    /*entries*/ 5, /*fields*/ NULL
  };
  introspection_linear[6].header = {DT_INTROSPECTION_TYPE_NONE, NULL, NULL, NULL, NULL, 0, 0, NULL};
#endif

  for(int i = 0; i <= 6; i++)
    introspection_linear[i].header.so = self;

  static dt_introspection_field_t *f5[] = {
    &introspection_linear[0],
    &introspection_linear[1],
    &introspection_linear[2],
    &introspection_linear[3],
    &introspection_linear[4],
    NULL
  };

  introspection_linear[5].Struct.fields = f5;

  return 0;
}

void * get_p(const void * param, const char * name)
{
  dt_iop_colorcorrection_params_t * p = (dt_iop_colorcorrection_params_t*)param;
  if(!strcmp(name, "hia"))
    return &(p->hia);
  else if(!strcmp(name, "hib"))
    return &(p->hib);
  else if(!strcmp(name, "loa"))
    return &(p->loa);
  else if(!strcmp(name, "lob"))
    return &(p->lob);
  else if(!strcmp(name, "saturation"))
    return &(p->saturation);
  else
    return NULL;
}

dt_introspection_field_t * get_f(const char * name)
{
  if(!g_ascii_strcasecmp(name, "hia"))
    return &(introspection_linear[0]);
  else if(!g_ascii_strcasecmp(name, "hib"))
    return &(introspection_linear[1]);
  else if(!g_ascii_strcasecmp(name, "loa"))
    return &(introspection_linear[2]);
  else if(!g_ascii_strcasecmp(name, "lob"))
    return &(introspection_linear[3]);
  else if(!g_ascii_strcasecmp(name, "saturation"))
    return &(introspection_linear[4]);
  else
    return NULL;
}

#ifdef __cplusplus
}
#endif

