/*
 * this code is auto generated. do not edit. change the sources in tools/introspection/ instead.
 *
 * the parse tree:
 *
 * typedef
 *   struct dt_iop_clipping_params_t
 *     float
 *     angle 
 *       default: 0.0
 *       range: -180.0 .. 180.0
 * 
 *     float
 *     cx 
 *       description: left
 *       default: 0.0
 *       range: 0.0 .. 1.0
 * 
 *     float
 *     cy 
 *       description: top
 *       default: 0.0
 *       range: 0.0 .. 1.0
 * 
 *     float
 *     cw 
 *       description: right
 *       default: 0.0
 *       range: 0.0 .. 1.0
 * 
 *     float
 *     ch 
 *       description: bottom
 *       default: 0.0
 *       range: 0.0 .. 1.0
 * 
 *     float
 *     k_h 
 *       default: 0.0
 *       range: -G_MAXFLOAT .. G_MAXFLOAT
 * 
 *     float
 *     k_v 
 *       default: 0.0
 *       range: -G_MAXFLOAT .. G_MAXFLOAT
 * 
 *     float
 *     kxa 
 *       default: 0.2
 *       range: 0.0 .. 1.0
 * 
 *     float
 *     kya 
 *       default: 0.2
 *       range: 0.0 .. 1.0
 * 
 *     float
 *     kxb 
 *       default: 0.8
 *       range: 0.0 .. 1.0
 * 
 *     float
 *     kyb 
 *       default: 0.2
 *       range: 0.0 .. 1.0
 * 
 *     float
 *     kxc 
 *       default: 0.8
 *       range: 0.0 .. 1.0
 * 
 *     float
 *     kyc 
 *       default: 0.8
 *       range: 0.0 .. 1.0
 * 
 *     float
 *     kxd 
 *       default: 0.2
 *       range: 0.0 .. 1.0
 * 
 *     float
 *     kyd 
 *       default: 0.8
 *       range: 0.0 .. 1.0
 * 
 *     int
 *     k_type 
 *       default: 0
 *       range: G_MININT .. G_MAXINT
 * 
 *     int
 *     k_sym 
 *       default: 0
 *       range: G_MININT .. G_MAXINT
 * 
 *     int
 *     k_apply 
 *       default: 0
 *       range: G_MININT .. G_MAXINT
 * 
 *     boolean
 *     crop_auto 
 *       description: automatic cropping
 *       default: TRUE
 * 
 *     int
 *     ratio_n 
 *       default: -1
 *       range: G_MININT .. G_MAXINT
 * 
 *     int
 *     ratio_d 
 *       default: -1
 *       range: G_MININT .. G_MAXINT
 * 
 * dt_iop_clipping_params_t
 *
 */

#include "/build/darktable/src/darktable-5.6.0/src/iop/clipping.c"

#ifdef __cplusplus
extern "C"
{
#endif

#include <string.h>
#include "common/introspection.h"

static dt_introspection_field_t introspection_linear[22+1]
#if !( defined(__cplusplus) && !defined(__clang__) && defined(__GNUC__) && ((__GNUC__*100 + __GNUC_MINOR__) <= 406) )
= {
{
  .Float = {
    { DT_INTROSPECTION_TYPE_FLOAT, (char*)"float", (char*)"angle", (char*)"angle", (char*)"", sizeof(((dt_iop_clipping_params_t*)NULL)->angle), G_STRUCT_OFFSET(dt_iop_clipping_params_t, angle), NULL },
    /*Min*/ -180.0, /*Max*/ 180.0, /*Default*/ 0.0
  }
},
{
  .Float = {
    { DT_INTROSPECTION_TYPE_FLOAT, (char*)"float", (char*)"cx", (char*)"cx", N_((char*)"left"), sizeof(((dt_iop_clipping_params_t*)NULL)->cx), G_STRUCT_OFFSET(dt_iop_clipping_params_t, cx), NULL },
    /*Min*/ 0.0, /*Max*/ 1.0, /*Default*/ 0.0
  }
},
{
  .Float = {
    { DT_INTROSPECTION_TYPE_FLOAT, (char*)"float", (char*)"cy", (char*)"cy", N_((char*)"top"), sizeof(((dt_iop_clipping_params_t*)NULL)->cy), G_STRUCT_OFFSET(dt_iop_clipping_params_t, cy), NULL },
    /*Min*/ 0.0, /*Max*/ 1.0, /*Default*/ 0.0
  }
},
{
  .Float = {
    { DT_INTROSPECTION_TYPE_FLOAT, (char*)"float", (char*)"cw", (char*)"cw", N_((char*)"right"), sizeof(((dt_iop_clipping_params_t*)NULL)->cw), G_STRUCT_OFFSET(dt_iop_clipping_params_t, cw), NULL },
    /*Min*/ 0.0, /*Max*/ 1.0, /*Default*/ 0.0
  }
},
{
  .Float = {
    { DT_INTROSPECTION_TYPE_FLOAT, (char*)"float", (char*)"ch", (char*)"ch", N_((char*)"bottom"), sizeof(((dt_iop_clipping_params_t*)NULL)->ch), G_STRUCT_OFFSET(dt_iop_clipping_params_t, ch), NULL },
    /*Min*/ 0.0, /*Max*/ 1.0, /*Default*/ 0.0
  }
},
{
  .Float = {
    { DT_INTROSPECTION_TYPE_FLOAT, (char*)"float", (char*)"k_h", (char*)"k_h", (char*)"", sizeof(((dt_iop_clipping_params_t*)NULL)->k_h), G_STRUCT_OFFSET(dt_iop_clipping_params_t, k_h), NULL },
    /*Min*/ -G_MAXFLOAT, /*Max*/ G_MAXFLOAT, /*Default*/ 0.0
  }
},
{
  .Float = {
    { DT_INTROSPECTION_TYPE_FLOAT, (char*)"float", (char*)"k_v", (char*)"k_v", (char*)"", sizeof(((dt_iop_clipping_params_t*)NULL)->k_v), G_STRUCT_OFFSET(dt_iop_clipping_params_t, k_v), NULL },
    /*Min*/ -G_MAXFLOAT, /*Max*/ G_MAXFLOAT, /*Default*/ 0.0
  }
},
{
  .Float = {
    { DT_INTROSPECTION_TYPE_FLOAT, (char*)"float", (char*)"kxa", (char*)"kxa", (char*)"", sizeof(((dt_iop_clipping_params_t*)NULL)->kxa), G_STRUCT_OFFSET(dt_iop_clipping_params_t, kxa), NULL },
    /*Min*/ 0.0, /*Max*/ 1.0, /*Default*/ 0.2
  }
},
{
  .Float = {
    { DT_INTROSPECTION_TYPE_FLOAT, (char*)"float", (char*)"kya", (char*)"kya", (char*)"", sizeof(((dt_iop_clipping_params_t*)NULL)->kya), G_STRUCT_OFFSET(dt_iop_clipping_params_t, kya), NULL },
    /*Min*/ 0.0, /*Max*/ 1.0, /*Default*/ 0.2
  }
},
{
  .Float = {
    { DT_INTROSPECTION_TYPE_FLOAT, (char*)"float", (char*)"kxb", (char*)"kxb", (char*)"", sizeof(((dt_iop_clipping_params_t*)NULL)->kxb), G_STRUCT_OFFSET(dt_iop_clipping_params_t, kxb), NULL },
    /*Min*/ 0.0, /*Max*/ 1.0, /*Default*/ 0.8
  }
},
{
  .Float = {
    { DT_INTROSPECTION_TYPE_FLOAT, (char*)"float", (char*)"kyb", (char*)"kyb", (char*)"", sizeof(((dt_iop_clipping_params_t*)NULL)->kyb), G_STRUCT_OFFSET(dt_iop_clipping_params_t, kyb), NULL },
    /*Min*/ 0.0, /*Max*/ 1.0, /*Default*/ 0.2
  }
},
{
  .Float = {
    { DT_INTROSPECTION_TYPE_FLOAT, (char*)"float", (char*)"kxc", (char*)"kxc", (char*)"", sizeof(((dt_iop_clipping_params_t*)NULL)->kxc), G_STRUCT_OFFSET(dt_iop_clipping_params_t, kxc), NULL },
    /*Min*/ 0.0, /*Max*/ 1.0, /*Default*/ 0.8
  }
},
{
  .Float = {
    { DT_INTROSPECTION_TYPE_FLOAT, (char*)"float", (char*)"kyc", (char*)"kyc", (char*)"", sizeof(((dt_iop_clipping_params_t*)NULL)->kyc), G_STRUCT_OFFSET(dt_iop_clipping_params_t, kyc), NULL },
    /*Min*/ 0.0, /*Max*/ 1.0, /*Default*/ 0.8
  }
},
{
  .Float = {
    { DT_INTROSPECTION_TYPE_FLOAT, (char*)"float", (char*)"kxd", (char*)"kxd", (char*)"", sizeof(((dt_iop_clipping_params_t*)NULL)->kxd), G_STRUCT_OFFSET(dt_iop_clipping_params_t, kxd), NULL },
    /*Min*/ 0.0, /*Max*/ 1.0, /*Default*/ 0.2
  }
},
{
  .Float = {
    { DT_INTROSPECTION_TYPE_FLOAT, (char*)"float", (char*)"kyd", (char*)"kyd", (char*)"", sizeof(((dt_iop_clipping_params_t*)NULL)->kyd), G_STRUCT_OFFSET(dt_iop_clipping_params_t, kyd), NULL },
    /*Min*/ 0.0, /*Max*/ 1.0, /*Default*/ 0.8
  }
},
{
  .Int = {
    { DT_INTROSPECTION_TYPE_INT, (char*)"int", (char*)"k_type", (char*)"k_type", (char*)"", sizeof(((dt_iop_clipping_params_t*)NULL)->k_type), G_STRUCT_OFFSET(dt_iop_clipping_params_t, k_type), NULL },
    /*Min*/ G_MININT, /*Max*/ G_MAXINT, /*Default*/ 0
  }
},
{
  .Int = {
    { DT_INTROSPECTION_TYPE_INT, (char*)"int", (char*)"k_sym", (char*)"k_sym", (char*)"", sizeof(((dt_iop_clipping_params_t*)NULL)->k_sym), G_STRUCT_OFFSET(dt_iop_clipping_params_t, k_sym), NULL },
    /*Min*/ G_MININT, /*Max*/ G_MAXINT, /*Default*/ 0
  }
},
{
  .Int = {
    { DT_INTROSPECTION_TYPE_INT, (char*)"int", (char*)"k_apply", (char*)"k_apply", (char*)"", sizeof(((dt_iop_clipping_params_t*)NULL)->k_apply), G_STRUCT_OFFSET(dt_iop_clipping_params_t, k_apply), NULL },
    /*Min*/ G_MININT, /*Max*/ G_MAXINT, /*Default*/ 0
  }
},
{
  .Bool = {
    { DT_INTROSPECTION_TYPE_BOOL, (char*)"bool", (char*)"crop_auto", (char*)"crop_auto", N_((char*)"automatic cropping"), sizeof(((dt_iop_clipping_params_t*)NULL)->crop_auto), G_STRUCT_OFFSET(dt_iop_clipping_params_t, crop_auto), NULL },
    /*Default*/ TRUE
  }
},
{
  .Int = {
    { DT_INTROSPECTION_TYPE_INT, (char*)"int", (char*)"ratio_n", (char*)"ratio_n", (char*)"", sizeof(((dt_iop_clipping_params_t*)NULL)->ratio_n), G_STRUCT_OFFSET(dt_iop_clipping_params_t, ratio_n), NULL },
    /*Min*/ G_MININT, /*Max*/ G_MAXINT, /*Default*/ -1
  }
},
{
  .Int = {
    { DT_INTROSPECTION_TYPE_INT, (char*)"int", (char*)"ratio_d", (char*)"ratio_d", (char*)"", sizeof(((dt_iop_clipping_params_t*)NULL)->ratio_d), G_STRUCT_OFFSET(dt_iop_clipping_params_t, ratio_d), NULL },
    /*Min*/ G_MININT, /*Max*/ G_MAXINT, /*Default*/ -1
  }
},
{
  .Struct = {
    { DT_INTROSPECTION_TYPE_STRUCT, (char*)"dt_iop_clipping_params_t", (char*)"", (char*)"", (char*)"", sizeof(dt_iop_clipping_params_t), 0, NULL },
    /*entries*/ 21, /*fields*/ NULL
  }
},
  { .header = {DT_INTROSPECTION_TYPE_NONE, NULL, NULL, NULL, NULL, 0, 0, NULL} }
}
#endif
;

static dt_introspection_t introspection = {
  8,
  5,
  "dt_iop_clipping_params_t",
  sizeof(dt_iop_clipping_params_t),
  &introspection_linear[22-1],
  sizeof(dt_iop_module_t),
  G_STRUCT_OFFSET(dt_iop_module_t, default_params)
};

dt_introspection_field_t* get_introspection_linear()
{
  return introspection_linear;
}

dt_introspection_t* get_introspection()
{
  return &introspection;
}

int introspection_init(struct dt_iop_module_so_t *self, int api_version)
{
  // here we check that the generated code matches the api at compile time and also at runtime
  if(introspection.api_version != DT_INTROSPECTION_VERSION || api_version != DT_INTROSPECTION_VERSION)
    return 1;

#if defined(__cplusplus) && !defined(__clang__) && defined(__GNUC__) && ((__GNUC__*100 + __GNUC_MINOR__) <= 406)
  introspection_linear[0].Float = {
    { DT_INTROSPECTION_TYPE_FLOAT, (char*)"float", (char*)"angle", (char*)"angle", (char*)"", sizeof(((dt_iop_clipping_params_t*)NULL)->angle), G_STRUCT_OFFSET(dt_iop_clipping_params_t, angle), NULL },
    /*Min*/ -180.0, /*Max*/ 180.0, /*Default*/ 0.0
  };
  introspection_linear[1].Float = {
    { DT_INTROSPECTION_TYPE_FLOAT, (char*)"float", (char*)"cx", (char*)"cx", N_((char*)"left"), sizeof(((dt_iop_clipping_params_t*)NULL)->cx), G_STRUCT_OFFSET(dt_iop_clipping_params_t, cx), NULL },
    /*Min*/ 0.0, /*Max*/ 1.0, /*Default*/ 0.0
  };
  introspection_linear[2].Float = {
    { DT_INTROSPECTION_TYPE_FLOAT, (char*)"float", (char*)"cy", (char*)"cy", N_((char*)"top"), sizeof(((dt_iop_clipping_params_t*)NULL)->cy), G_STRUCT_OFFSET(dt_iop_clipping_params_t, cy), NULL },
    /*Min*/ 0.0, /*Max*/ 1.0, /*Default*/ 0.0
  };
  introspection_linear[3].Float = {
    { DT_INTROSPECTION_TYPE_FLOAT, (char*)"float", (char*)"cw", (char*)"cw", N_((char*)"right"), sizeof(((dt_iop_clipping_params_t*)NULL)->cw), G_STRUCT_OFFSET(dt_iop_clipping_params_t, cw), NULL },
    /*Min*/ 0.0, /*Max*/ 1.0, /*Default*/ 0.0
  };
  introspection_linear[4].Float = {
    { DT_INTROSPECTION_TYPE_FLOAT, (char*)"float", (char*)"ch", (char*)"ch", N_((char*)"bottom"), sizeof(((dt_iop_clipping_params_t*)NULL)->ch), G_STRUCT_OFFSET(dt_iop_clipping_params_t, ch), NULL },
    /*Min*/ 0.0, /*Max*/ 1.0, /*Default*/ 0.0
  };
  introspection_linear[5].Float = {
    { DT_INTROSPECTION_TYPE_FLOAT, (char*)"float", (char*)"k_h", (char*)"k_h", (char*)"", sizeof(((dt_iop_clipping_params_t*)NULL)->k_h), G_STRUCT_OFFSET(dt_iop_clipping_params_t, k_h), NULL },
    /*Min*/ -G_MAXFLOAT, /*Max*/ G_MAXFLOAT, /*Default*/ 0.0
  };
  introspection_linear[6].Float = {
    { DT_INTROSPECTION_TYPE_FLOAT, (char*)"float", (char*)"k_v", (char*)"k_v", (char*)"", sizeof(((dt_iop_clipping_params_t*)NULL)->k_v), G_STRUCT_OFFSET(dt_iop_clipping_params_t, k_v), NULL },
    /*Min*/ -G_MAXFLOAT, /*Max*/ G_MAXFLOAT, /*Default*/ 0.0
  };
  introspection_linear[7].Float = {
    { DT_INTROSPECTION_TYPE_FLOAT, (char*)"float", (char*)"kxa", (char*)"kxa", (char*)"", sizeof(((dt_iop_clipping_params_t*)NULL)->kxa), G_STRUCT_OFFSET(dt_iop_clipping_params_t, kxa), NULL },
    /*Min*/ 0.0, /*Max*/ 1.0, /*Default*/ 0.2
  };
  introspection_linear[8].Float = {
    { DT_INTROSPECTION_TYPE_FLOAT, (char*)"float", (char*)"kya", (char*)"kya", (char*)"", sizeof(((dt_iop_clipping_params_t*)NULL)->kya), G_STRUCT_OFFSET(dt_iop_clipping_params_t, kya), NULL },
    /*Min*/ 0.0, /*Max*/ 1.0, /*Default*/ 0.2
  };
  introspection_linear[9].Float = {
    { DT_INTROSPECTION_TYPE_FLOAT, (char*)"float", (char*)"kxb", (char*)"kxb", (char*)"", sizeof(((dt_iop_clipping_params_t*)NULL)->kxb), G_STRUCT_OFFSET(dt_iop_clipping_params_t, kxb), NULL },
    /*Min*/ 0.0, /*Max*/ 1.0, /*Default*/ 0.8
  };
  introspection_linear[10].Float = {
    { DT_INTROSPECTION_TYPE_FLOAT, (char*)"float", (char*)"kyb", (char*)"kyb", (char*)"", sizeof(((dt_iop_clipping_params_t*)NULL)->kyb), G_STRUCT_OFFSET(dt_iop_clipping_params_t, kyb), NULL },
    /*Min*/ 0.0, /*Max*/ 1.0, /*Default*/ 0.2
  };
  introspection_linear[11].Float = {
    { DT_INTROSPECTION_TYPE_FLOAT, (char*)"float", (char*)"kxc", (char*)"kxc", (char*)"", sizeof(((dt_iop_clipping_params_t*)NULL)->kxc), G_STRUCT_OFFSET(dt_iop_clipping_params_t, kxc), NULL },
    /*Min*/ 0.0, /*Max*/ 1.0, /*Default*/ 0.8
  };
  introspection_linear[12].Float = {
    { DT_INTROSPECTION_TYPE_FLOAT, (char*)"float", (char*)"kyc", (char*)"kyc", (char*)"", sizeof(((dt_iop_clipping_params_t*)NULL)->kyc), G_STRUCT_OFFSET(dt_iop_clipping_params_t, kyc), NULL },
    /*Min*/ 0.0, /*Max*/ 1.0, /*Default*/ 0.8
  };
  introspection_linear[13].Float = {
    { DT_INTROSPECTION_TYPE_FLOAT, (char*)"float", (char*)"kxd", (char*)"kxd", (char*)"", sizeof(((dt_iop_clipping_params_t*)NULL)->kxd), G_STRUCT_OFFSET(dt_iop_clipping_params_t, kxd), NULL },
    /*Min*/ 0.0, /*Max*/ 1.0, /*Default*/ 0.2
  };
  introspection_linear[14].Float = {
    { DT_INTROSPECTION_TYPE_FLOAT, (char*)"float", (char*)"kyd", (char*)"kyd", (char*)"", sizeof(((dt_iop_clipping_params_t*)NULL)->kyd), G_STRUCT_OFFSET(dt_iop_clipping_params_t, kyd), NULL },
    /*Min*/ 0.0, /*Max*/ 1.0, /*Default*/ 0.8
  };
  introspection_linear[15].Int = {
    { DT_INTROSPECTION_TYPE_INT, (char*)"int", (char*)"k_type", (char*)"k_type", (char*)"", sizeof(((dt_iop_clipping_params_t*)NULL)->k_type), G_STRUCT_OFFSET(dt_iop_clipping_params_t, k_type), NULL },
    /*Min*/ G_MININT, /*Max*/ G_MAXINT, /*Default*/ 0
  };
  introspection_linear[16].Int = {
    { DT_INTROSPECTION_TYPE_INT, (char*)"int", (char*)"k_sym", (char*)"k_sym", (char*)"", sizeof(((dt_iop_clipping_params_t*)NULL)->k_sym), G_STRUCT_OFFSET(dt_iop_clipping_params_t, k_sym), NULL },
    /*Min*/ G_MININT, /*Max*/ G_MAXINT, /*Default*/ 0
  };
  introspection_linear[17].Int = {
    { DT_INTROSPECTION_TYPE_INT, (char*)"int", (char*)"k_apply", (char*)"k_apply", (char*)"", sizeof(((dt_iop_clipping_params_t*)NULL)->k_apply), G_STRUCT_OFFSET(dt_iop_clipping_params_t, k_apply), NULL },
    /*Min*/ G_MININT, /*Max*/ G_MAXINT, /*Default*/ 0
  };
  introspection_linear[18].Bool = {
    { DT_INTROSPECTION_TYPE_BOOL, (char*)"bool", (char*)"crop_auto", (char*)"crop_auto", N_((char*)"automatic cropping"), sizeof(((dt_iop_clipping_params_t*)NULL)->crop_auto), G_STRUCT_OFFSET(dt_iop_clipping_params_t, crop_auto), NULL },
    /*Default*/ TRUE
  };
  introspection_linear[19].Int = {
    { DT_INTROSPECTION_TYPE_INT, (char*)"int", (char*)"ratio_n", (char*)"ratio_n", (char*)"", sizeof(((dt_iop_clipping_params_t*)NULL)->ratio_n), G_STRUCT_OFFSET(dt_iop_clipping_params_t, ratio_n), NULL },
    /*Min*/ G_MININT, /*Max*/ G_MAXINT, /*Default*/ -1
  };
  introspection_linear[20].Int = {
    { DT_INTROSPECTION_TYPE_INT, (char*)"int", (char*)"ratio_d", (char*)"ratio_d", (char*)"", sizeof(((dt_iop_clipping_params_t*)NULL)->ratio_d), G_STRUCT_OFFSET(dt_iop_clipping_params_t, ratio_d), NULL },
    /*Min*/ G_MININT, /*Max*/ G_MAXINT, /*Default*/ -1
  };
  introspection_linear[21].Struct = {
    { DT_INTROSPECTION_TYPE_STRUCT, (char*)"dt_iop_clipping_params_t", (char*)"", (char*)"", (char*)"", sizeof(dt_iop_clipping_params_t), 0, NULL },
    /*entries*/ 21, /*fields*/ NULL
  };
  introspection_linear[22].header = {DT_INTROSPECTION_TYPE_NONE, NULL, NULL, NULL, NULL, 0, 0, NULL};
#endif

  for(int i = 0; i <= 22; i++)
    introspection_linear[i].header.so = self;

  static dt_introspection_field_t *f21[] = {
    &introspection_linear[0],
    &introspection_linear[1],
    &introspection_linear[2],
    &introspection_linear[3],
    &introspection_linear[4],
    &introspection_linear[5],
    &introspection_linear[6],
    &introspection_linear[7],
    &introspection_linear[8],
    &introspection_linear[9],
    &introspection_linear[10],
    &introspection_linear[11],
    &introspection_linear[12],
    &introspection_linear[13],
    &introspection_linear[14],
    &introspection_linear[15],
    &introspection_linear[16],
    &introspection_linear[17],
    &introspection_linear[18],
    &introspection_linear[19],
    &introspection_linear[20],
    NULL
  };

  introspection_linear[21].Struct.fields = f21;

  return 0;
}

void * get_p(const void * param, const char * name)
{
  dt_iop_clipping_params_t * p = (dt_iop_clipping_params_t*)param;
  if(!strcmp(name, "angle"))
    return &(p->angle);
  else if(!strcmp(name, "cx"))
    return &(p->cx);
  else if(!strcmp(name, "cy"))
    return &(p->cy);
  else if(!strcmp(name, "cw"))
    return &(p->cw);
  else if(!strcmp(name, "ch"))
    return &(p->ch);
  else if(!strcmp(name, "k_h"))
    return &(p->k_h);
  else if(!strcmp(name, "k_v"))
    return &(p->k_v);
  else if(!strcmp(name, "kxa"))
    return &(p->kxa);
  else if(!strcmp(name, "kya"))
    return &(p->kya);
  else if(!strcmp(name, "kxb"))
    return &(p->kxb);
  else if(!strcmp(name, "kyb"))
    return &(p->kyb);
  else if(!strcmp(name, "kxc"))
    return &(p->kxc);
  else if(!strcmp(name, "kyc"))
    return &(p->kyc);
  else if(!strcmp(name, "kxd"))
    return &(p->kxd);
  else if(!strcmp(name, "kyd"))
    return &(p->kyd);
  else if(!strcmp(name, "k_type"))
    return &(p->k_type);
  else if(!strcmp(name, "k_sym"))
    return &(p->k_sym);
  else if(!strcmp(name, "k_apply"))
    return &(p->k_apply);
  else if(!strcmp(name, "crop_auto"))
    return &(p->crop_auto);
  else if(!strcmp(name, "ratio_n"))
    return &(p->ratio_n);
  else if(!strcmp(name, "ratio_d"))
    return &(p->ratio_d);
  else
    return NULL;
}

dt_introspection_field_t * get_f(const char * name)
{
  if(!g_ascii_strcasecmp(name, "angle"))
    return &(introspection_linear[0]);
  else if(!g_ascii_strcasecmp(name, "cx"))
    return &(introspection_linear[1]);
  else if(!g_ascii_strcasecmp(name, "cy"))
    return &(introspection_linear[2]);
  else if(!g_ascii_strcasecmp(name, "cw"))
    return &(introspection_linear[3]);
  else if(!g_ascii_strcasecmp(name, "ch"))
    return &(introspection_linear[4]);
  else if(!g_ascii_strcasecmp(name, "k_h"))
    return &(introspection_linear[5]);
  else if(!g_ascii_strcasecmp(name, "k_v"))
    return &(introspection_linear[6]);
  else if(!g_ascii_strcasecmp(name, "kxa"))
    return &(introspection_linear[7]);
  else if(!g_ascii_strcasecmp(name, "kya"))
    return &(introspection_linear[8]);
  else if(!g_ascii_strcasecmp(name, "kxb"))
    return &(introspection_linear[9]);
  else if(!g_ascii_strcasecmp(name, "kyb"))
    return &(introspection_linear[10]);
  else if(!g_ascii_strcasecmp(name, "kxc"))
    return &(introspection_linear[11]);
  else if(!g_ascii_strcasecmp(name, "kyc"))
    return &(introspection_linear[12]);
  else if(!g_ascii_strcasecmp(name, "kxd"))
    return &(introspection_linear[13]);
  else if(!g_ascii_strcasecmp(name, "kyd"))
    return &(introspection_linear[14]);
  else if(!g_ascii_strcasecmp(name, "k_type"))
    return &(introspection_linear[15]);
  else if(!g_ascii_strcasecmp(name, "k_sym"))
    return &(introspection_linear[16]);
  else if(!g_ascii_strcasecmp(name, "k_apply"))
    return &(introspection_linear[17]);
  else if(!g_ascii_strcasecmp(name, "crop_auto"))
    return &(introspection_linear[18]);
  else if(!g_ascii_strcasecmp(name, "ratio_n"))
    return &(introspection_linear[19]);
  else if(!g_ascii_strcasecmp(name, "ratio_d"))
    return &(introspection_linear[20]);
  else
    return NULL;
}

#ifdef __cplusplus
}
#endif

