#pragma once

static const char *section1[] = {
"Pascal Obry","Hanno Schwalm","Victor Forsiuk","Mario Zimmermann","Ralf Brown","Andrii Ryzhkov","Dan Torop","Jeronimo Pellegrini","Philipp Lutz","Robert Mongold","Daniele Pighin","Bill Ferguson","Kofa","Martin Straeten","Miloš Komarčević","Diederik ter Rahe","EdgarLux","Ryo Shinozaki","Matjaž Jeran","deekayhd","esq4","David Smith","Hpbirkeland","Marko Vertainen"
,NULL };
gtk_about_dialog_add_credit_section (GTK_ABOUT_DIALOG(dialog), _("Developers"), section1);

static const char *section2[] = {
"Victor Forsiuk","Pascal Obry","Jeronimo Pellegrini","Martin Straeten","Ryo Shinozaki","EdgarLux","Matjaž Jeran","esq4","Hpbirkeland","Marko Vertainen","RIM","Andrii Ryzhkov","Besmir Godolja","Báthory Péter","Johan Schiff","Kang-Wei Hsu","Tianhao Chai","Elia Devito","GameMaster0x","Jan Šmucr","zxs-un"
,NULL };
gtk_about_dialog_add_credit_section (GTK_ABOUT_DIALOG(dialog), _("Translators"), section2);

static const char *section3[] = {
"Miloš Komarčević","Roman Lebedev","Sidhartha","Steven Le Rouzic"
,NULL };
gtk_about_dialog_add_credit_section (GTK_ABOUT_DIALOG(dialog), _("rawspeed contributors"), section3);

static const char *section4[] = {
"Pascal Obry","yuri"
,NULL };
gtk_about_dialog_add_credit_section (GTK_ABOUT_DIALOG(dialog), _("integration contributors"), section4);

static const char *section5[] = {
"Bill Ferguson","Tino Mettler","deekayhd","Andrii Ryzhkov","Ilya Palopezhentsev","Martin Straeten","Björn Lindström","Dan Zwell","dterrahe"
,NULL };
gtk_about_dialog_add_credit_section (GTK_ABOUT_DIALOG(dialog), _("lua-scripts contributors"), section5);

