/*
 * this code is auto generated. do not edit. change the sources in tools/introspection/ instead.
 *
 * the parse tree:
 *
 * typedef
 *   struct dt_iop_invert_params_t
 *     float
 *     color [4]
 *       default: 1.0
 *       range: -G_MAXFLOAT .. G_MAXFLOAT
 * 
 * dt_iop_invert_params_t
 *
 */

#include "/build/darktable/src/darktable-5.6.0/src/iop/invert.c"

#ifdef __cplusplus
extern "C"
{
#endif

#include <string.h>
#include "common/introspection.h"

static dt_introspection_field_t introspection_linear[3+1]
#if !( defined(__cplusplus) && !defined(__clang__) && defined(__GNUC__) && ((__GNUC__*100 + __GNUC_MINOR__) <= 406) )
= {
{
  .Float = {
    { DT_INTROSPECTION_TYPE_FLOAT, (char*)"float", (char*)"color[0]", (char*)"color[0]", (char*)"", sizeof(((dt_iop_invert_params_t*)NULL)->color[0]), G_STRUCT_OFFSET(dt_iop_invert_params_t, color), NULL },
    /*Min*/ -G_MAXFLOAT, /*Max*/ G_MAXFLOAT, /*Default*/ 1.0
  }
},
{
  .Array = {
    { DT_INTROSPECTION_TYPE_ARRAY, (char*)"float[]", (char*)"color", (char*)"color", (char*)"", sizeof(((dt_iop_invert_params_t*)NULL)->color), G_STRUCT_OFFSET(dt_iop_invert_params_t, color), NULL },
    /*count*/ G_N_ELEMENTS(((dt_iop_invert_params_t*)NULL)->color), /*type*/ DT_INTROSPECTION_TYPE_FLOAT, /*field*/ &introspection_linear[0]
  }
},
{
  .Struct = {
    { DT_INTROSPECTION_TYPE_STRUCT, (char*)"dt_iop_invert_params_t", (char*)"", (char*)"", (char*)"", sizeof(dt_iop_invert_params_t), 0, NULL },
    /*entries*/ 1, /*fields*/ NULL
  }
},
  { .header = {DT_INTROSPECTION_TYPE_NONE, NULL, NULL, NULL, NULL, 0, 0, NULL} }
}
#endif
;

static dt_introspection_t introspection = {
  8,
  2,
  "dt_iop_invert_params_t",
  sizeof(dt_iop_invert_params_t),
  &introspection_linear[3-1],
  sizeof(dt_iop_module_t),
  G_STRUCT_OFFSET(dt_iop_module_t, default_params)
};

dt_introspection_field_t* get_introspection_linear()
{
  return introspection_linear;
}

dt_introspection_t* get_introspection()
{
  return &introspection;
}

int introspection_init(struct dt_iop_module_so_t *self, int api_version)
{
  // here we check that the generated code matches the api at compile time and also at runtime
  if(introspection.api_version != DT_INTROSPECTION_VERSION || api_version != DT_INTROSPECTION_VERSION)
    return 1;

#if defined(__cplusplus) && !defined(__clang__) && defined(__GNUC__) && ((__GNUC__*100 + __GNUC_MINOR__) <= 406)
  introspection_linear[0].Float = {
    { DT_INTROSPECTION_TYPE_FLOAT, (char*)"float", (char*)"color[0]", (char*)"color[0]", (char*)"", sizeof(((dt_iop_invert_params_t*)NULL)->color[0]), G_STRUCT_OFFSET(dt_iop_invert_params_t, color), NULL },
    /*Min*/ -G_MAXFLOAT, /*Max*/ G_MAXFLOAT, /*Default*/ 1.0
  };
  introspection_linear[1].Array = {
    { DT_INTROSPECTION_TYPE_ARRAY, (char*)"float[]", (char*)"color", (char*)"color", (char*)"", sizeof(((dt_iop_invert_params_t*)NULL)->color), G_STRUCT_OFFSET(dt_iop_invert_params_t, color), NULL },
    /*count*/ G_N_ELEMENTS(((dt_iop_invert_params_t*)NULL)->color), /*type*/ DT_INTROSPECTION_TYPE_FLOAT, /*field*/ &introspection_linear[0]
  };
  introspection_linear[2].Struct = {
    { DT_INTROSPECTION_TYPE_STRUCT, (char*)"dt_iop_invert_params_t", (char*)"", (char*)"", (char*)"", sizeof(dt_iop_invert_params_t), 0, NULL },
    /*entries*/ 1, /*fields*/ NULL
  };
  introspection_linear[3].header = {DT_INTROSPECTION_TYPE_NONE, NULL, NULL, NULL, NULL, 0, 0, NULL};
#endif

  for(int i = 0; i <= 3; i++)
    introspection_linear[i].header.so = self;

  static dt_introspection_field_t *f2[] = {
    &introspection_linear[1],
    NULL
  };

  introspection_linear[2].Struct.fields = f2;

  return 0;
}

void * get_p(const void * param, const char * name)
{
  dt_iop_invert_params_t * p = (dt_iop_invert_params_t*)param;
  if(!strcmp(name, "color[0]"))
    return &(p->color[0]);
  else if(!strcmp(name, "color"))
    return &(p->color);
  else
    return NULL;
}

dt_introspection_field_t * get_f(const char * name)
{
  if(!g_ascii_strcasecmp(name, "color[0]"))
    return &(introspection_linear[0]);
  else if(!g_ascii_strcasecmp(name, "color"))
    return &(introspection_linear[1]);
  else
    return NULL;
}

#ifdef __cplusplus
}
#endif

