/*
 * this code is auto generated. do not edit. change the sources in tools/introspection/ instead.
 *
 * the parse tree:
 *
 * typedef
 *   struct dt_iop_colorequal_params_t
 *     float
 *     threshold 
 *       description: saturation threshold
 *       default: 0.1
 *       range: 0.0 .. 0.3
 * 
 *     float
 *     smoothing_hue 
 *       description: hue curve
 *       default: 1.0
 *       range: 0.05 .. 2.0
 * 
 *     float
 *     contrast 
 *       description: contrast
 *       default: 0.0
 *       range: -1.0 .. 1.0
 * 
 *     float
 *     white_level 
 *       description: white level
 *       default: 1.0
 *       range: -2.0 .. 16.0
 * 
 *     float
 *     chroma_size 
 *       description: hue analysis radius
 *       default: 1.5
 *       range: 1.0 .. 10.0
 * 
 *     float
 *     param_size 
 *       description: effect radius
 *       default: 1.0
 *       range: 1.0 .. 128.
 * 
 *     boolean
 *     use_filter 
 *       description: use guided filter
 *       default: TRUE
 * 
 *     float
 *     sat_red 
 *       description: red
 *       default: 1.0
 *       range: 0. .. 2.
 * 
 *     float
 *     sat_orange 
 *       description: orange
 *       default: 1.0
 *       range: 0. .. 2.
 * 
 *     float
 *     sat_yellow 
 *       description: yellow
 *       default: 1.0
 *       range: 0. .. 2.
 * 
 *     float
 *     sat_green 
 *       description: green
 *       default: 1.0
 *       range: 0. .. 2.
 * 
 *     float
 *     sat_cyan 
 *       description: cyan
 *       default: 1.0
 *       range: 0. .. 2.
 * 
 *     float
 *     sat_blue 
 *       description: blue
 *       default: 1.0
 *       range: 0. .. 2.
 * 
 *     float
 *     sat_lavender 
 *       description: lavender
 *       default: 1.0
 *       range: 0. .. 2.
 * 
 *     float
 *     sat_magenta 
 *       description: magenta
 *       default: 1.0
 *       range: 0. .. 2.
 * 
 *     float
 *     hue_red 
 *       description: red
 *       default: 0.0
 *       range: -180. .. 180.
 * 
 *     float
 *     hue_orange 
 *       description: orange
 *       default: 0.0
 *       range: -180. .. 180.
 * 
 *     float
 *     hue_yellow 
 *       description: yellow
 *       default: 0.0
 *       range: -180. .. 180.
 * 
 *     float
 *     hue_green 
 *       description: green
 *       default: 0.0
 *       range: -180. .. 180.
 * 
 *     float
 *     hue_cyan 
 *       description: cyan
 *       default: 0.0
 *       range: -180. .. 180.
 * 
 *     float
 *     hue_blue 
 *       description: blue
 *       default: 0.0
 *       range: -180. .. 180.
 * 
 *     float
 *     hue_lavender 
 *       description: lavender
 *       default: 0.0
 *       range: -180. .. 180.
 * 
 *     float
 *     hue_magenta 
 *       description: magenta
 *       default: 0.0
 *       range: -180. .. 180.
 * 
 *     float
 *     bright_red 
 *       description: red
 *       default: 1.0
 *       range: 0. .. 2.
 * 
 *     float
 *     bright_orange 
 *       description: orange
 *       default: 1.0
 *       range: 0. .. 2.
 * 
 *     float
 *     bright_yellow 
 *       description: yellow
 *       default: 1.0
 *       range: 0. .. 2.
 * 
 *     float
 *     bright_green 
 *       description: green
 *       default: 1.0
 *       range: 0. .. 2.
 * 
 *     float
 *     bright_cyan 
 *       description: cyan
 *       default: 1.0
 *       range: 0. .. 2.
 * 
 *     float
 *     bright_blue 
 *       description: blue
 *       default: 1.0
 *       range: 0. .. 2.
 * 
 *     float
 *     bright_lavender 
 *       description: lavender
 *       default: 1.0
 *       range: 0. .. 2.
 * 
 *     float
 *     bright_magenta 
 *       description: magenta
 *       default: 1.0
 *       range: 0. .. 2.
 * 
 *     float
 *     hue_shift 
 *       description: node placement
 *       default: 0.0
 *       range: -23. .. 23.
 * 
 * dt_iop_colorequal_params_t
 *
 */

#include "/build/darktable/src/darktable-5.6.0/src/iop/colorequal.c"

#ifdef __cplusplus
extern "C"
{
#endif

#include <string.h>
#include "common/introspection.h"

static dt_introspection_field_t introspection_linear[33+1]
#if !( defined(__cplusplus) && !defined(__clang__) && defined(__GNUC__) && ((__GNUC__*100 + __GNUC_MINOR__) <= 406) )
= {
{
  .Float = {
    { DT_INTROSPECTION_TYPE_FLOAT, (char*)"float", (char*)"threshold", (char*)"threshold", N_((char*)"saturation threshold"), sizeof(((dt_iop_colorequal_params_t*)NULL)->threshold), G_STRUCT_OFFSET(dt_iop_colorequal_params_t, threshold), NULL },
    /*Min*/ 0.0, /*Max*/ 0.3, /*Default*/ 0.1
  }
},
{
  .Float = {
    { DT_INTROSPECTION_TYPE_FLOAT, (char*)"float", (char*)"smoothing_hue", (char*)"smoothing_hue", N_((char*)"hue curve"), sizeof(((dt_iop_colorequal_params_t*)NULL)->smoothing_hue), G_STRUCT_OFFSET(dt_iop_colorequal_params_t, smoothing_hue), NULL },
    /*Min*/ 0.05, /*Max*/ 2.0, /*Default*/ 1.0
  }
},
{
  .Float = {
    { DT_INTROSPECTION_TYPE_FLOAT, (char*)"float", (char*)"contrast", (char*)"contrast", N_((char*)"contrast"), sizeof(((dt_iop_colorequal_params_t*)NULL)->contrast), G_STRUCT_OFFSET(dt_iop_colorequal_params_t, contrast), NULL },
    /*Min*/ -1.0, /*Max*/ 1.0, /*Default*/ 0.0
  }
},
{
  .Float = {
    { DT_INTROSPECTION_TYPE_FLOAT, (char*)"float", (char*)"white_level", (char*)"white_level", N_((char*)"white level"), sizeof(((dt_iop_colorequal_params_t*)NULL)->white_level), G_STRUCT_OFFSET(dt_iop_colorequal_params_t, white_level), NULL },
    /*Min*/ -2.0, /*Max*/ 16.0, /*Default*/ 1.0
  }
},
{
  .Float = {
    { DT_INTROSPECTION_TYPE_FLOAT, (char*)"float", (char*)"chroma_size", (char*)"chroma_size", N_((char*)"hue analysis radius"), sizeof(((dt_iop_colorequal_params_t*)NULL)->chroma_size), G_STRUCT_OFFSET(dt_iop_colorequal_params_t, chroma_size), NULL },
    /*Min*/ 1.0, /*Max*/ 10.0, /*Default*/ 1.5
  }
},
{
  .Float = {
    { DT_INTROSPECTION_TYPE_FLOAT, (char*)"float", (char*)"param_size", (char*)"param_size", N_((char*)"effect radius"), sizeof(((dt_iop_colorequal_params_t*)NULL)->param_size), G_STRUCT_OFFSET(dt_iop_colorequal_params_t, param_size), NULL },
    /*Min*/ 1.0, /*Max*/ 128., /*Default*/ 1.0
  }
},
{
  .Bool = {
    { DT_INTROSPECTION_TYPE_BOOL, (char*)"bool", (char*)"use_filter", (char*)"use_filter", N_((char*)"use guided filter"), sizeof(((dt_iop_colorequal_params_t*)NULL)->use_filter), G_STRUCT_OFFSET(dt_iop_colorequal_params_t, use_filter), NULL },
    /*Default*/ TRUE
  }
},
{
  .Float = {
    { DT_INTROSPECTION_TYPE_FLOAT, (char*)"float", (char*)"sat_red", (char*)"sat_red", N_((char*)"red"), sizeof(((dt_iop_colorequal_params_t*)NULL)->sat_red), G_STRUCT_OFFSET(dt_iop_colorequal_params_t, sat_red), NULL },
    /*Min*/ 0., /*Max*/ 2., /*Default*/ 1.0
  }
},
{
  .Float = {
    { DT_INTROSPECTION_TYPE_FLOAT, (char*)"float", (char*)"sat_orange", (char*)"sat_orange", N_((char*)"orange"), sizeof(((dt_iop_colorequal_params_t*)NULL)->sat_orange), G_STRUCT_OFFSET(dt_iop_colorequal_params_t, sat_orange), NULL },
    /*Min*/ 0., /*Max*/ 2., /*Default*/ 1.0
  }
},
{
  .Float = {
    { DT_INTROSPECTION_TYPE_FLOAT, (char*)"float", (char*)"sat_yellow", (char*)"sat_yellow", N_((char*)"yellow"), sizeof(((dt_iop_colorequal_params_t*)NULL)->sat_yellow), G_STRUCT_OFFSET(dt_iop_colorequal_params_t, sat_yellow), NULL },
    /*Min*/ 0., /*Max*/ 2., /*Default*/ 1.0
  }
},
{
  .Float = {
    { DT_INTROSPECTION_TYPE_FLOAT, (char*)"float", (char*)"sat_green", (char*)"sat_green", N_((char*)"green"), sizeof(((dt_iop_colorequal_params_t*)NULL)->sat_green), G_STRUCT_OFFSET(dt_iop_colorequal_params_t, sat_green), NULL },
    /*Min*/ 0., /*Max*/ 2., /*Default*/ 1.0
  }
},
{
  .Float = {
    { DT_INTROSPECTION_TYPE_FLOAT, (char*)"float", (char*)"sat_cyan", (char*)"sat_cyan", N_((char*)"cyan"), sizeof(((dt_iop_colorequal_params_t*)NULL)->sat_cyan), G_STRUCT_OFFSET(dt_iop_colorequal_params_t, sat_cyan), NULL },
    /*Min*/ 0., /*Max*/ 2., /*Default*/ 1.0
  }
},
{
  .Float = {
    { DT_INTROSPECTION_TYPE_FLOAT, (char*)"float", (char*)"sat_blue", (char*)"sat_blue", N_((char*)"blue"), sizeof(((dt_iop_colorequal_params_t*)NULL)->sat_blue), G_STRUCT_OFFSET(dt_iop_colorequal_params_t, sat_blue), NULL },
    /*Min*/ 0., /*Max*/ 2., /*Default*/ 1.0
  }
},
{
  .Float = {
    { DT_INTROSPECTION_TYPE_FLOAT, (char*)"float", (char*)"sat_lavender", (char*)"sat_lavender", N_((char*)"lavender"), sizeof(((dt_iop_colorequal_params_t*)NULL)->sat_lavender), G_STRUCT_OFFSET(dt_iop_colorequal_params_t, sat_lavender), NULL },
    /*Min*/ 0., /*Max*/ 2., /*Default*/ 1.0
  }
},
{
  .Float = {
    { DT_INTROSPECTION_TYPE_FLOAT, (char*)"float", (char*)"sat_magenta", (char*)"sat_magenta", N_((char*)"magenta"), sizeof(((dt_iop_colorequal_params_t*)NULL)->sat_magenta), G_STRUCT_OFFSET(dt_iop_colorequal_params_t, sat_magenta), NULL },
    /*Min*/ 0., /*Max*/ 2., /*Default*/ 1.0
  }
},
{
  .Float = {
    { DT_INTROSPECTION_TYPE_FLOAT, (char*)"float", (char*)"hue_red", (char*)"hue_red", N_((char*)"red"), sizeof(((dt_iop_colorequal_params_t*)NULL)->hue_red), G_STRUCT_OFFSET(dt_iop_colorequal_params_t, hue_red), NULL },
    /*Min*/ -180., /*Max*/ 180., /*Default*/ 0.0
  }
},
{
  .Float = {
    { DT_INTROSPECTION_TYPE_FLOAT, (char*)"float", (char*)"hue_orange", (char*)"hue_orange", N_((char*)"orange"), sizeof(((dt_iop_colorequal_params_t*)NULL)->hue_orange), G_STRUCT_OFFSET(dt_iop_colorequal_params_t, hue_orange), NULL },
    /*Min*/ -180., /*Max*/ 180., /*Default*/ 0.0
  }
},
{
  .Float = {
    { DT_INTROSPECTION_TYPE_FLOAT, (char*)"float", (char*)"hue_yellow", (char*)"hue_yellow", N_((char*)"yellow"), sizeof(((dt_iop_colorequal_params_t*)NULL)->hue_yellow), G_STRUCT_OFFSET(dt_iop_colorequal_params_t, hue_yellow), NULL },
    /*Min*/ -180., /*Max*/ 180., /*Default*/ 0.0
  }
},
{
  .Float = {
    { DT_INTROSPECTION_TYPE_FLOAT, (char*)"float", (char*)"hue_green", (char*)"hue_green", N_((char*)"green"), sizeof(((dt_iop_colorequal_params_t*)NULL)->hue_green), G_STRUCT_OFFSET(dt_iop_colorequal_params_t, hue_green), NULL },
    /*Min*/ -180., /*Max*/ 180., /*Default*/ 0.0
  }
},
{
  .Float = {
    { DT_INTROSPECTION_TYPE_FLOAT, (char*)"float", (char*)"hue_cyan", (char*)"hue_cyan", N_((char*)"cyan"), sizeof(((dt_iop_colorequal_params_t*)NULL)->hue_cyan), G_STRUCT_OFFSET(dt_iop_colorequal_params_t, hue_cyan), NULL },
    /*Min*/ -180., /*Max*/ 180., /*Default*/ 0.0
  }
},
{
  .Float = {
    { DT_INTROSPECTION_TYPE_FLOAT, (char*)"float", (char*)"hue_blue", (char*)"hue_blue", N_((char*)"blue"), sizeof(((dt_iop_colorequal_params_t*)NULL)->hue_blue), G_STRUCT_OFFSET(dt_iop_colorequal_params_t, hue_blue), NULL },
    /*Min*/ -180., /*Max*/ 180., /*Default*/ 0.0
  }
},
{
  .Float = {
    { DT_INTROSPECTION_TYPE_FLOAT, (char*)"float", (char*)"hue_lavender", (char*)"hue_lavender", N_((char*)"lavender"), sizeof(((dt_iop_colorequal_params_t*)NULL)->hue_lavender), G_STRUCT_OFFSET(dt_iop_colorequal_params_t, hue_lavender), NULL },
    /*Min*/ -180., /*Max*/ 180., /*Default*/ 0.0
  }
},
{
  .Float = {
    { DT_INTROSPECTION_TYPE_FLOAT, (char*)"float", (char*)"hue_magenta", (char*)"hue_magenta", N_((char*)"magenta"), sizeof(((dt_iop_colorequal_params_t*)NULL)->hue_magenta), G_STRUCT_OFFSET(dt_iop_colorequal_params_t, hue_magenta), NULL },
    /*Min*/ -180., /*Max*/ 180., /*Default*/ 0.0
  }
},
{
  .Float = {
    { DT_INTROSPECTION_TYPE_FLOAT, (char*)"float", (char*)"bright_red", (char*)"bright_red", N_((char*)"red"), sizeof(((dt_iop_colorequal_params_t*)NULL)->bright_red), G_STRUCT_OFFSET(dt_iop_colorequal_params_t, bright_red), NULL },
    /*Min*/ 0., /*Max*/ 2., /*Default*/ 1.0
  }
},
{
  .Float = {
    { DT_INTROSPECTION_TYPE_FLOAT, (char*)"float", (char*)"bright_orange", (char*)"bright_orange", N_((char*)"orange"), sizeof(((dt_iop_colorequal_params_t*)NULL)->bright_orange), G_STRUCT_OFFSET(dt_iop_colorequal_params_t, bright_orange), NULL },
    /*Min*/ 0., /*Max*/ 2., /*Default*/ 1.0
  }
},
{
  .Float = {
    { DT_INTROSPECTION_TYPE_FLOAT, (char*)"float", (char*)"bright_yellow", (char*)"bright_yellow", N_((char*)"yellow"), sizeof(((dt_iop_colorequal_params_t*)NULL)->bright_yellow), G_STRUCT_OFFSET(dt_iop_colorequal_params_t, bright_yellow), NULL },
    /*Min*/ 0., /*Max*/ 2., /*Default*/ 1.0
  }
},
{
  .Float = {
    { DT_INTROSPECTION_TYPE_FLOAT, (char*)"float", (char*)"bright_green", (char*)"bright_green", N_((char*)"green"), sizeof(((dt_iop_colorequal_params_t*)NULL)->bright_green), G_STRUCT_OFFSET(dt_iop_colorequal_params_t, bright_green), NULL },
    /*Min*/ 0., /*Max*/ 2., /*Default*/ 1.0
  }
},
{
  .Float = {
    { DT_INTROSPECTION_TYPE_FLOAT, (char*)"float", (char*)"bright_cyan", (char*)"bright_cyan", N_((char*)"cyan"), sizeof(((dt_iop_colorequal_params_t*)NULL)->bright_cyan), G_STRUCT_OFFSET(dt_iop_colorequal_params_t, bright_cyan), NULL },
    /*Min*/ 0., /*Max*/ 2., /*Default*/ 1.0
  }
},
{
  .Float = {
    { DT_INTROSPECTION_TYPE_FLOAT, (char*)"float", (char*)"bright_blue", (char*)"bright_blue", N_((char*)"blue"), sizeof(((dt_iop_colorequal_params_t*)NULL)->bright_blue), G_STRUCT_OFFSET(dt_iop_colorequal_params_t, bright_blue), NULL },
    /*Min*/ 0., /*Max*/ 2., /*Default*/ 1.0
  }
},
{
  .Float = {
    { DT_INTROSPECTION_TYPE_FLOAT, (char*)"float", (char*)"bright_lavender", (char*)"bright_lavender", N_((char*)"lavender"), sizeof(((dt_iop_colorequal_params_t*)NULL)->bright_lavender), G_STRUCT_OFFSET(dt_iop_colorequal_params_t, bright_lavender), NULL },
    /*Min*/ 0., /*Max*/ 2., /*Default*/ 1.0
  }
},
{
  .Float = {
    { DT_INTROSPECTION_TYPE_FLOAT, (char*)"float", (char*)"bright_magenta", (char*)"bright_magenta", N_((char*)"magenta"), sizeof(((dt_iop_colorequal_params_t*)NULL)->bright_magenta), G_STRUCT_OFFSET(dt_iop_colorequal_params_t, bright_magenta), NULL },
    /*Min*/ 0., /*Max*/ 2., /*Default*/ 1.0
  }
},
{
  .Float = {
    { DT_INTROSPECTION_TYPE_FLOAT, (char*)"float", (char*)"hue_shift", (char*)"hue_shift", N_((char*)"node placement"), sizeof(((dt_iop_colorequal_params_t*)NULL)->hue_shift), G_STRUCT_OFFSET(dt_iop_colorequal_params_t, hue_shift), NULL },
    /*Min*/ -23., /*Max*/ 23., /*Default*/ 0.0
  }
},
{
  .Struct = {
    { DT_INTROSPECTION_TYPE_STRUCT, (char*)"dt_iop_colorequal_params_t", (char*)"", (char*)"", (char*)"", sizeof(dt_iop_colorequal_params_t), 0, NULL },
    /*entries*/ 32, /*fields*/ NULL
  }
},
  { .header = {DT_INTROSPECTION_TYPE_NONE, NULL, NULL, NULL, NULL, 0, 0, NULL} }
}
#endif
;

static dt_introspection_t introspection = {
  8,
  4,
  "dt_iop_colorequal_params_t",
  sizeof(dt_iop_colorequal_params_t),
  &introspection_linear[33-1],
  sizeof(dt_iop_module_t),
  G_STRUCT_OFFSET(dt_iop_module_t, default_params)
};

dt_introspection_field_t* get_introspection_linear()
{
  return introspection_linear;
}

dt_introspection_t* get_introspection()
{
  return &introspection;
}

int introspection_init(struct dt_iop_module_so_t *self, int api_version)
{
  // here we check that the generated code matches the api at compile time and also at runtime
  if(introspection.api_version != DT_INTROSPECTION_VERSION || api_version != DT_INTROSPECTION_VERSION)
    return 1;

#if defined(__cplusplus) && !defined(__clang__) && defined(__GNUC__) && ((__GNUC__*100 + __GNUC_MINOR__) <= 406)
  introspection_linear[0].Float = {
    { DT_INTROSPECTION_TYPE_FLOAT, (char*)"float", (char*)"threshold", (char*)"threshold", N_((char*)"saturation threshold"), sizeof(((dt_iop_colorequal_params_t*)NULL)->threshold), G_STRUCT_OFFSET(dt_iop_colorequal_params_t, threshold), NULL },
    /*Min*/ 0.0, /*Max*/ 0.3, /*Default*/ 0.1
  };
  introspection_linear[1].Float = {
    { DT_INTROSPECTION_TYPE_FLOAT, (char*)"float", (char*)"smoothing_hue", (char*)"smoothing_hue", N_((char*)"hue curve"), sizeof(((dt_iop_colorequal_params_t*)NULL)->smoothing_hue), G_STRUCT_OFFSET(dt_iop_colorequal_params_t, smoothing_hue), NULL },
    /*Min*/ 0.05, /*Max*/ 2.0, /*Default*/ 1.0
  };
  introspection_linear[2].Float = {
    { DT_INTROSPECTION_TYPE_FLOAT, (char*)"float", (char*)"contrast", (char*)"contrast", N_((char*)"contrast"), sizeof(((dt_iop_colorequal_params_t*)NULL)->contrast), G_STRUCT_OFFSET(dt_iop_colorequal_params_t, contrast), NULL },
    /*Min*/ -1.0, /*Max*/ 1.0, /*Default*/ 0.0
  };
  introspection_linear[3].Float = {
    { DT_INTROSPECTION_TYPE_FLOAT, (char*)"float", (char*)"white_level", (char*)"white_level", N_((char*)"white level"), sizeof(((dt_iop_colorequal_params_t*)NULL)->white_level), G_STRUCT_OFFSET(dt_iop_colorequal_params_t, white_level), NULL },
    /*Min*/ -2.0, /*Max*/ 16.0, /*Default*/ 1.0
  };
  introspection_linear[4].Float = {
    { DT_INTROSPECTION_TYPE_FLOAT, (char*)"float", (char*)"chroma_size", (char*)"chroma_size", N_((char*)"hue analysis radius"), sizeof(((dt_iop_colorequal_params_t*)NULL)->chroma_size), G_STRUCT_OFFSET(dt_iop_colorequal_params_t, chroma_size), NULL },
    /*Min*/ 1.0, /*Max*/ 10.0, /*Default*/ 1.5
  };
  introspection_linear[5].Float = {
    { DT_INTROSPECTION_TYPE_FLOAT, (char*)"float", (char*)"param_size", (char*)"param_size", N_((char*)"effect radius"), sizeof(((dt_iop_colorequal_params_t*)NULL)->param_size), G_STRUCT_OFFSET(dt_iop_colorequal_params_t, param_size), NULL },
    /*Min*/ 1.0, /*Max*/ 128., /*Default*/ 1.0
  };
  introspection_linear[6].Bool = {
    { DT_INTROSPECTION_TYPE_BOOL, (char*)"bool", (char*)"use_filter", (char*)"use_filter", N_((char*)"use guided filter"), sizeof(((dt_iop_colorequal_params_t*)NULL)->use_filter), G_STRUCT_OFFSET(dt_iop_colorequal_params_t, use_filter), NULL },
    /*Default*/ TRUE
  };
  introspection_linear[7].Float = {
    { DT_INTROSPECTION_TYPE_FLOAT, (char*)"float", (char*)"sat_red", (char*)"sat_red", N_((char*)"red"), sizeof(((dt_iop_colorequal_params_t*)NULL)->sat_red), G_STRUCT_OFFSET(dt_iop_colorequal_params_t, sat_red), NULL },
    /*Min*/ 0., /*Max*/ 2., /*Default*/ 1.0
  };
  introspection_linear[8].Float = {
    { DT_INTROSPECTION_TYPE_FLOAT, (char*)"float", (char*)"sat_orange", (char*)"sat_orange", N_((char*)"orange"), sizeof(((dt_iop_colorequal_params_t*)NULL)->sat_orange), G_STRUCT_OFFSET(dt_iop_colorequal_params_t, sat_orange), NULL },
    /*Min*/ 0., /*Max*/ 2., /*Default*/ 1.0
  };
  introspection_linear[9].Float = {
    { DT_INTROSPECTION_TYPE_FLOAT, (char*)"float", (char*)"sat_yellow", (char*)"sat_yellow", N_((char*)"yellow"), sizeof(((dt_iop_colorequal_params_t*)NULL)->sat_yellow), G_STRUCT_OFFSET(dt_iop_colorequal_params_t, sat_yellow), NULL },
    /*Min*/ 0., /*Max*/ 2., /*Default*/ 1.0
  };
  introspection_linear[10].Float = {
    { DT_INTROSPECTION_TYPE_FLOAT, (char*)"float", (char*)"sat_green", (char*)"sat_green", N_((char*)"green"), sizeof(((dt_iop_colorequal_params_t*)NULL)->sat_green), G_STRUCT_OFFSET(dt_iop_colorequal_params_t, sat_green), NULL },
    /*Min*/ 0., /*Max*/ 2., /*Default*/ 1.0
  };
  introspection_linear[11].Float = {
    { DT_INTROSPECTION_TYPE_FLOAT, (char*)"float", (char*)"sat_cyan", (char*)"sat_cyan", N_((char*)"cyan"), sizeof(((dt_iop_colorequal_params_t*)NULL)->sat_cyan), G_STRUCT_OFFSET(dt_iop_colorequal_params_t, sat_cyan), NULL },
    /*Min*/ 0., /*Max*/ 2., /*Default*/ 1.0
  };
  introspection_linear[12].Float = {
    { DT_INTROSPECTION_TYPE_FLOAT, (char*)"float", (char*)"sat_blue", (char*)"sat_blue", N_((char*)"blue"), sizeof(((dt_iop_colorequal_params_t*)NULL)->sat_blue), G_STRUCT_OFFSET(dt_iop_colorequal_params_t, sat_blue), NULL },
    /*Min*/ 0., /*Max*/ 2., /*Default*/ 1.0
  };
  introspection_linear[13].Float = {
    { DT_INTROSPECTION_TYPE_FLOAT, (char*)"float", (char*)"sat_lavender", (char*)"sat_lavender", N_((char*)"lavender"), sizeof(((dt_iop_colorequal_params_t*)NULL)->sat_lavender), G_STRUCT_OFFSET(dt_iop_colorequal_params_t, sat_lavender), NULL },
    /*Min*/ 0., /*Max*/ 2., /*Default*/ 1.0
  };
  introspection_linear[14].Float = {
    { DT_INTROSPECTION_TYPE_FLOAT, (char*)"float", (char*)"sat_magenta", (char*)"sat_magenta", N_((char*)"magenta"), sizeof(((dt_iop_colorequal_params_t*)NULL)->sat_magenta), G_STRUCT_OFFSET(dt_iop_colorequal_params_t, sat_magenta), NULL },
    /*Min*/ 0., /*Max*/ 2., /*Default*/ 1.0
  };
  introspection_linear[15].Float = {
    { DT_INTROSPECTION_TYPE_FLOAT, (char*)"float", (char*)"hue_red", (char*)"hue_red", N_((char*)"red"), sizeof(((dt_iop_colorequal_params_t*)NULL)->hue_red), G_STRUCT_OFFSET(dt_iop_colorequal_params_t, hue_red), NULL },
    /*Min*/ -180., /*Max*/ 180., /*Default*/ 0.0
  };
  introspection_linear[16].Float = {
    { DT_INTROSPECTION_TYPE_FLOAT, (char*)"float", (char*)"hue_orange", (char*)"hue_orange", N_((char*)"orange"), sizeof(((dt_iop_colorequal_params_t*)NULL)->hue_orange), G_STRUCT_OFFSET(dt_iop_colorequal_params_t, hue_orange), NULL },
    /*Min*/ -180., /*Max*/ 180., /*Default*/ 0.0
  };
  introspection_linear[17].Float = {
    { DT_INTROSPECTION_TYPE_FLOAT, (char*)"float", (char*)"hue_yellow", (char*)"hue_yellow", N_((char*)"yellow"), sizeof(((dt_iop_colorequal_params_t*)NULL)->hue_yellow), G_STRUCT_OFFSET(dt_iop_colorequal_params_t, hue_yellow), NULL },
    /*Min*/ -180., /*Max*/ 180., /*Default*/ 0.0
  };
  introspection_linear[18].Float = {
    { DT_INTROSPECTION_TYPE_FLOAT, (char*)"float", (char*)"hue_green", (char*)"hue_green", N_((char*)"green"), sizeof(((dt_iop_colorequal_params_t*)NULL)->hue_green), G_STRUCT_OFFSET(dt_iop_colorequal_params_t, hue_green), NULL },
    /*Min*/ -180., /*Max*/ 180., /*Default*/ 0.0
  };
  introspection_linear[19].Float = {
    { DT_INTROSPECTION_TYPE_FLOAT, (char*)"float", (char*)"hue_cyan", (char*)"hue_cyan", N_((char*)"cyan"), sizeof(((dt_iop_colorequal_params_t*)NULL)->hue_cyan), G_STRUCT_OFFSET(dt_iop_colorequal_params_t, hue_cyan), NULL },
    /*Min*/ -180., /*Max*/ 180., /*Default*/ 0.0
  };
  introspection_linear[20].Float = {
    { DT_INTROSPECTION_TYPE_FLOAT, (char*)"float", (char*)"hue_blue", (char*)"hue_blue", N_((char*)"blue"), sizeof(((dt_iop_colorequal_params_t*)NULL)->hue_blue), G_STRUCT_OFFSET(dt_iop_colorequal_params_t, hue_blue), NULL },
    /*Min*/ -180., /*Max*/ 180., /*Default*/ 0.0
  };
  introspection_linear[21].Float = {
    { DT_INTROSPECTION_TYPE_FLOAT, (char*)"float", (char*)"hue_lavender", (char*)"hue_lavender", N_((char*)"lavender"), sizeof(((dt_iop_colorequal_params_t*)NULL)->hue_lavender), G_STRUCT_OFFSET(dt_iop_colorequal_params_t, hue_lavender), NULL },
    /*Min*/ -180., /*Max*/ 180., /*Default*/ 0.0
  };
  introspection_linear[22].Float = {
    { DT_INTROSPECTION_TYPE_FLOAT, (char*)"float", (char*)"hue_magenta", (char*)"hue_magenta", N_((char*)"magenta"), sizeof(((dt_iop_colorequal_params_t*)NULL)->hue_magenta), G_STRUCT_OFFSET(dt_iop_colorequal_params_t, hue_magenta), NULL },
    /*Min*/ -180., /*Max*/ 180., /*Default*/ 0.0
  };
  introspection_linear[23].Float = {
    { DT_INTROSPECTION_TYPE_FLOAT, (char*)"float", (char*)"bright_red", (char*)"bright_red", N_((char*)"red"), sizeof(((dt_iop_colorequal_params_t*)NULL)->bright_red), G_STRUCT_OFFSET(dt_iop_colorequal_params_t, bright_red), NULL },
    /*Min*/ 0., /*Max*/ 2., /*Default*/ 1.0
  };
  introspection_linear[24].Float = {
    { DT_INTROSPECTION_TYPE_FLOAT, (char*)"float", (char*)"bright_orange", (char*)"bright_orange", N_((char*)"orange"), sizeof(((dt_iop_colorequal_params_t*)NULL)->bright_orange), G_STRUCT_OFFSET(dt_iop_colorequal_params_t, bright_orange), NULL },
    /*Min*/ 0., /*Max*/ 2., /*Default*/ 1.0
  };
  introspection_linear[25].Float = {
    { DT_INTROSPECTION_TYPE_FLOAT, (char*)"float", (char*)"bright_yellow", (char*)"bright_yellow", N_((char*)"yellow"), sizeof(((dt_iop_colorequal_params_t*)NULL)->bright_yellow), G_STRUCT_OFFSET(dt_iop_colorequal_params_t, bright_yellow), NULL },
    /*Min*/ 0., /*Max*/ 2., /*Default*/ 1.0
  };
  introspection_linear[26].Float = {
    { DT_INTROSPECTION_TYPE_FLOAT, (char*)"float", (char*)"bright_green", (char*)"bright_green", N_((char*)"green"), sizeof(((dt_iop_colorequal_params_t*)NULL)->bright_green), G_STRUCT_OFFSET(dt_iop_colorequal_params_t, bright_green), NULL },
    /*Min*/ 0., /*Max*/ 2., /*Default*/ 1.0
  };
  introspection_linear[27].Float = {
    { DT_INTROSPECTION_TYPE_FLOAT, (char*)"float", (char*)"bright_cyan", (char*)"bright_cyan", N_((char*)"cyan"), sizeof(((dt_iop_colorequal_params_t*)NULL)->bright_cyan), G_STRUCT_OFFSET(dt_iop_colorequal_params_t, bright_cyan), NULL },
    /*Min*/ 0., /*Max*/ 2., /*Default*/ 1.0
  };
  introspection_linear[28].Float = {
    { DT_INTROSPECTION_TYPE_FLOAT, (char*)"float", (char*)"bright_blue", (char*)"bright_blue", N_((char*)"blue"), sizeof(((dt_iop_colorequal_params_t*)NULL)->bright_blue), G_STRUCT_OFFSET(dt_iop_colorequal_params_t, bright_blue), NULL },
    /*Min*/ 0., /*Max*/ 2., /*Default*/ 1.0
  };
  introspection_linear[29].Float = {
    { DT_INTROSPECTION_TYPE_FLOAT, (char*)"float", (char*)"bright_lavender", (char*)"bright_lavender", N_((char*)"lavender"), sizeof(((dt_iop_colorequal_params_t*)NULL)->bright_lavender), G_STRUCT_OFFSET(dt_iop_colorequal_params_t, bright_lavender), NULL },
    /*Min*/ 0., /*Max*/ 2., /*Default*/ 1.0
  };
  introspection_linear[30].Float = {
    { DT_INTROSPECTION_TYPE_FLOAT, (char*)"float", (char*)"bright_magenta", (char*)"bright_magenta", N_((char*)"magenta"), sizeof(((dt_iop_colorequal_params_t*)NULL)->bright_magenta), G_STRUCT_OFFSET(dt_iop_colorequal_params_t, bright_magenta), NULL },
    /*Min*/ 0., /*Max*/ 2., /*Default*/ 1.0
  };
  introspection_linear[31].Float = {
    { DT_INTROSPECTION_TYPE_FLOAT, (char*)"float", (char*)"hue_shift", (char*)"hue_shift", N_((char*)"node placement"), sizeof(((dt_iop_colorequal_params_t*)NULL)->hue_shift), G_STRUCT_OFFSET(dt_iop_colorequal_params_t, hue_shift), NULL },
    /*Min*/ -23., /*Max*/ 23., /*Default*/ 0.0
  };
  introspection_linear[32].Struct = {
    { DT_INTROSPECTION_TYPE_STRUCT, (char*)"dt_iop_colorequal_params_t", (char*)"", (char*)"", (char*)"", sizeof(dt_iop_colorequal_params_t), 0, NULL },
    /*entries*/ 32, /*fields*/ NULL
  };
  introspection_linear[33].header = {DT_INTROSPECTION_TYPE_NONE, NULL, NULL, NULL, NULL, 0, 0, NULL};
#endif

  for(int i = 0; i <= 33; i++)
    introspection_linear[i].header.so = self;

  static dt_introspection_field_t *f32[] = {
    &introspection_linear[0],
    &introspection_linear[1],
    &introspection_linear[2],
    &introspection_linear[3],
    &introspection_linear[4],
    &introspection_linear[5],
    &introspection_linear[6],
    &introspection_linear[7],
    &introspection_linear[8],
    &introspection_linear[9],
    &introspection_linear[10],
    &introspection_linear[11],
    &introspection_linear[12],
    &introspection_linear[13],
    &introspection_linear[14],
    &introspection_linear[15],
    &introspection_linear[16],
    &introspection_linear[17],
    &introspection_linear[18],
    &introspection_linear[19],
    &introspection_linear[20],
    &introspection_linear[21],
    &introspection_linear[22],
    &introspection_linear[23],
    &introspection_linear[24],
    &introspection_linear[25],
    &introspection_linear[26],
    &introspection_linear[27],
    &introspection_linear[28],
    &introspection_linear[29],
    &introspection_linear[30],
    &introspection_linear[31],
    NULL
  };

  introspection_linear[32].Struct.fields = f32;

  return 0;
}

void * get_p(const void * param, const char * name)
{
  dt_iop_colorequal_params_t * p = (dt_iop_colorequal_params_t*)param;
  if(!strcmp(name, "threshold"))
    return &(p->threshold);
  else if(!strcmp(name, "smoothing_hue"))
    return &(p->smoothing_hue);
  else if(!strcmp(name, "contrast"))
    return &(p->contrast);
  else if(!strcmp(name, "white_level"))
    return &(p->white_level);
  else if(!strcmp(name, "chroma_size"))
    return &(p->chroma_size);
  else if(!strcmp(name, "param_size"))
    return &(p->param_size);
  else if(!strcmp(name, "use_filter"))
    return &(p->use_filter);
  else if(!strcmp(name, "sat_red"))
    return &(p->sat_red);
  else if(!strcmp(name, "sat_orange"))
    return &(p->sat_orange);
  else if(!strcmp(name, "sat_yellow"))
    return &(p->sat_yellow);
  else if(!strcmp(name, "sat_green"))
    return &(p->sat_green);
  else if(!strcmp(name, "sat_cyan"))
    return &(p->sat_cyan);
  else if(!strcmp(name, "sat_blue"))
    return &(p->sat_blue);
  else if(!strcmp(name, "sat_lavender"))
    return &(p->sat_lavender);
  else if(!strcmp(name, "sat_magenta"))
    return &(p->sat_magenta);
  else if(!strcmp(name, "hue_red"))
    return &(p->hue_red);
  else if(!strcmp(name, "hue_orange"))
    return &(p->hue_orange);
  else if(!strcmp(name, "hue_yellow"))
    return &(p->hue_yellow);
  else if(!strcmp(name, "hue_green"))
    return &(p->hue_green);
  else if(!strcmp(name, "hue_cyan"))
    return &(p->hue_cyan);
  else if(!strcmp(name, "hue_blue"))
    return &(p->hue_blue);
  else if(!strcmp(name, "hue_lavender"))
    return &(p->hue_lavender);
  else if(!strcmp(name, "hue_magenta"))
    return &(p->hue_magenta);
  else if(!strcmp(name, "bright_red"))
    return &(p->bright_red);
  else if(!strcmp(name, "bright_orange"))
    return &(p->bright_orange);
  else if(!strcmp(name, "bright_yellow"))
    return &(p->bright_yellow);
  else if(!strcmp(name, "bright_green"))
    return &(p->bright_green);
  else if(!strcmp(name, "bright_cyan"))
    return &(p->bright_cyan);
  else if(!strcmp(name, "bright_blue"))
    return &(p->bright_blue);
  else if(!strcmp(name, "bright_lavender"))
    return &(p->bright_lavender);
  else if(!strcmp(name, "bright_magenta"))
    return &(p->bright_magenta);
  else if(!strcmp(name, "hue_shift"))
    return &(p->hue_shift);
  else
    return NULL;
}

dt_introspection_field_t * get_f(const char * name)
{
  if(!g_ascii_strcasecmp(name, "threshold"))
    return &(introspection_linear[0]);
  else if(!g_ascii_strcasecmp(name, "smoothing_hue"))
    return &(introspection_linear[1]);
  else if(!g_ascii_strcasecmp(name, "contrast"))
    return &(introspection_linear[2]);
  else if(!g_ascii_strcasecmp(name, "white_level"))
    return &(introspection_linear[3]);
  else if(!g_ascii_strcasecmp(name, "chroma_size"))
    return &(introspection_linear[4]);
  else if(!g_ascii_strcasecmp(name, "param_size"))
    return &(introspection_linear[5]);
  else if(!g_ascii_strcasecmp(name, "use_filter"))
    return &(introspection_linear[6]);
  else if(!g_ascii_strcasecmp(name, "sat_red"))
    return &(introspection_linear[7]);
  else if(!g_ascii_strcasecmp(name, "sat_orange"))
    return &(introspection_linear[8]);
  else if(!g_ascii_strcasecmp(name, "sat_yellow"))
    return &(introspection_linear[9]);
  else if(!g_ascii_strcasecmp(name, "sat_green"))
    return &(introspection_linear[10]);
  else if(!g_ascii_strcasecmp(name, "sat_cyan"))
    return &(introspection_linear[11]);
  else if(!g_ascii_strcasecmp(name, "sat_blue"))
    return &(introspection_linear[12]);
  else if(!g_ascii_strcasecmp(name, "sat_lavender"))
    return &(introspection_linear[13]);
  else if(!g_ascii_strcasecmp(name, "sat_magenta"))
    return &(introspection_linear[14]);
  else if(!g_ascii_strcasecmp(name, "hue_red"))
    return &(introspection_linear[15]);
  else if(!g_ascii_strcasecmp(name, "hue_orange"))
    return &(introspection_linear[16]);
  else if(!g_ascii_strcasecmp(name, "hue_yellow"))
    return &(introspection_linear[17]);
  else if(!g_ascii_strcasecmp(name, "hue_green"))
    return &(introspection_linear[18]);
  else if(!g_ascii_strcasecmp(name, "hue_cyan"))
    return &(introspection_linear[19]);
  else if(!g_ascii_strcasecmp(name, "hue_blue"))
    return &(introspection_linear[20]);
  else if(!g_ascii_strcasecmp(name, "hue_lavender"))
    return &(introspection_linear[21]);
  else if(!g_ascii_strcasecmp(name, "hue_magenta"))
    return &(introspection_linear[22]);
  else if(!g_ascii_strcasecmp(name, "bright_red"))
    return &(introspection_linear[23]);
  else if(!g_ascii_strcasecmp(name, "bright_orange"))
    return &(introspection_linear[24]);
  else if(!g_ascii_strcasecmp(name, "bright_yellow"))
    return &(introspection_linear[25]);
  else if(!g_ascii_strcasecmp(name, "bright_green"))
    return &(introspection_linear[26]);
  else if(!g_ascii_strcasecmp(name, "bright_cyan"))
    return &(introspection_linear[27]);
  else if(!g_ascii_strcasecmp(name, "bright_blue"))
    return &(introspection_linear[28]);
  else if(!g_ascii_strcasecmp(name, "bright_lavender"))
    return &(introspection_linear[29]);
  else if(!g_ascii_strcasecmp(name, "bright_magenta"))
    return &(introspection_linear[30]);
  else if(!g_ascii_strcasecmp(name, "hue_shift"))
    return &(introspection_linear[31]);
  else
    return NULL;
}

#ifdef __cplusplus
}
#endif

